# define T41_INCLUDE
# define T33_INCLUDE
# define T42_INCLUDE
# define T_INDEX64
# define _M_AMD64
# define MUTEX

#if defined(_WIN32) || defined(_WIN64)
#  include <windows.h>
#endif

#include <cstring>
#include <iostream>
#include <pthread.h>


///////////////////////////////////////////////////////////////////////////////
//
//  SMP stuff - just copied from Crafty source

#if defined(SMP)

#if defined(_WIN32) || defined(_WIN64)

#include <process.h>

#  define pthread_attr_t  HANDLE
#  define pthread_t       HANDLE
#  define thread_t        HANDLE
#  define tfork(t,f,p)    (pthread_t)_beginthreadex(0,0,(unsigned int (__stdcall *)(void *))(f),(void *)(p),0,0);

#if (defined (_M_ALPHA) && !defined(NT_INTEREX))
   
#  ifdef __cplusplus
      extern "C" __int64 __asm (char *, ...);
      extern "C" void __MB (void);
#     pragma intrinsic (__asm)
#     pragma intrinsic (__MB)
#  endif

  typedef volatile int lock_t[1];

#  define LockInit(v)      ((v)[0] = 0)
#  define LockFree(v)      ((v)[0] = 0)
#  define UnLock(v)        (__MB(), (v)[0] = 0)

   __inline void Lock (volatile int *hPtr) {
   __asm ("lp: ldl_l v0,(a0);"
          "    xor v0,1,v0;"
          "    beq v0,lp;"
          "    stl_c v0,(a0);"
          "    beq v0,lp;"
          "    mb;",
                  hPtr);
   }

#elif (defined (_M_IX86) && !defined(NT_INTEREX))

   typedef volatile int lock_t[1];

#  define LockInit(v)      ((v)[0] = 0)
#  define LockFree(v)      ((v)[0] = 0)
#  define UnLock(v)        ((v)[0] = 0)


#  if (_MSC_VER > 1200)

#     ifdef __cplusplus
        extern "C" long _InterlockedExchange (long*, long);
        extern "C" void __cdecl _ReadWriteBarrier(void);
#     else
        extern long _InterlockedExchange (long*, long);
        extern void _ReadWriteBarrier(void);
#     endif
#     pragma intrinsic (_InterlockedExchange)
#     pragma intrinsic (_ReadWriteBarrier)

   __forceinline void Lock (volatile int *hPtr)
    {
    int iValue;
    volatile int *hPtrTmp;

    hPtrTmp = hPtr;     // Workaround for vc7 beta1 bug
    iValue = 1;
    _ReadWriteBarrier();
    for (;;)
        {
        iValue = _InterlockedExchange ((long*) hPtrTmp, iValue);
        if (0 == iValue)
            return;
        while (*hPtrTmp)
            ;   // Do nothing
        }
    _ReadWriteBarrier();
    }F

#  else

__inline void Lock (volatile int *hPtr)
    {
    __asm
      {
        mov     ecx, hPtr
   la:  mov     eax, 1
        xchg    eax, [ecx]
        test    eax, eax
        jz      end
   lb:  mov     eax, [ecx]
        test    eax, eax
        jz      la
        jmp     lb
   end:
      }
    }

#  endif

#elif ((defined (_M_IA64) || defined(_M_AMD64)) && !defined(NT_INTEREX))

#  include <windows.h>

extern "C" void __cdecl _ReadWriteBarrier(void);
#  pragma intrinsic (_InterlockedExchange)
#  pragma intrinsic (_ReadWriteBarrier)

   typedef volatile LONG lock_t[1];

#  define LockInit(v)      ((v)[0] = 0)
#  define LockFree(v)      ((v)[0] = 0)
#  define UnLock(v)        ((v)[0] = 0)


__forceinline void Lock (volatile LONG *hPtr)
    {
    int iValue;

    _ReadWriteBarrier();
    for (;;)
        {
        iValue = _InterlockedExchange ((LPLONG) hPtr, 1);
        if (0 == iValue)
            return;
        while (*hPtr)
            ;   // Do nothing
        }
    _ReadWriteBarrier();
    }


#else /* NT non-Alpha/Intel, without assembler Lock() */

#  define lock_t           volatile int
#  define LockInit(v)      ((v) = 0)
#  define LockFree(v)      ((v) = 0)
#  define Lock(v)          do {                                         \
                             while(InterlockedExchange((LPLONG)&(v),1) != 0);  \
                           } while (0)
#  define UnLock(v)        ((v) = 0)

#endif /* architecture check */

#else  /* not NT, assume SMP using POSIX threads (LINUX, etc)  */

#include <pthread.h>

#if defined(MUTEX)

#  define Lock(v)          pthread_mutex_lock(&v)
#  define LockInit(v)      pthread_mutex_init(&v,0)
#  define LockFree(v)      pthread_mutex_destroy(&v)
#  define UnLock(v)        pthread_mutex_unlock(&v)
#  define lock_t           pthread_mutex_t

#elif defined(ALPHA)

#  include <machine/builtins.h>

#  define lock_t           volatile long
#  define LockInit(v)      ((v) = 0)
#  define LockFree(v)      ((v) = 0)
#  define Lock(v)          __LOCK_LONG(&(v))
#  define UnLock(v)        __UNLOCK_LONG(&(v))

#else /* POSIX, but not using MUTEXes */

#  define exchange(adr,reg)                                  \
     ({ volatile int _ret;                                   \
     asm volatile ("xchgw %0,%1"                             \
     : "=q" (_ret), "=m" (*(adr))    /* Output %0,%1 */      \
     : "m"  (*(adr)), "0"  (reg));   /* Input (%2),%0 */     \
     _ret;                                                   \
     })
#  define LockInit(p)           (p=0)
#  define LockFree(p)           (p=0)
#  define UnLock(p)             (exchange(&p,0))
#  define Lock(p)               while(exchange(&p,1)) while(p)
#  define lock_t                volatile int

#endif /* MUTEX */

#if defined(CLONE)
#  define tfork(t,f,p)   {                                            \
                         char *m=malloc(0x100010);                    \
                         if (m <= 0) printf("malloc() failed\n");     \
                         (void) clone(f,                            \
                         m+0x100000,                                  \
                         CLONE_VM+CLONE_FILES+17,                     \
                         (void*) p); }
                           
#else
#  define tfork(t,f,p)          pthread_create(&t,NULL,f, (void*)p)
#endif

#endif /* NT or POSIX */

#else
#  define LockInit(p)
#  define LockFree(p)
#  define Lock(p)
#  define UnLock(p)
#endif /*  SMP code */

///////////////////////////////////////////////////////////////////////////////

// UNDONE: on error delete files

#define STOP_ON_ERROR
#define NEW
//#define   KPPKP_16BIT

#if defined(SMP)
static lock_t lockIO;
#endif

#if defined (DEBUG) || defined(EBUG)
#  define   Assert(x)   ((x) ? 0 : * (volatile int*) 0)
#else
#  define   Assert(x)
#endif

#if !defined (DEBUG) && !defined(EBUG)
#define NDEBUG
#endif
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <time.h>

#if defined (_WIN32) || defined(_WIN64)
#include <windows.h>
#endif

// Constants and types

#if defined (_MSC_VER)

#undef  TB_CDECL
#define TB_CDECL    __cdecl
#define TB_FASTCALL __fastcall
#if _MSC_VER >= 1200
#define INLINE      __forceinline
#endif

#else

#define TB_CDECL
#define TB_FASTCALL

#endif

#if !defined (INLINE)
#define INLINE  inline
#endif

typedef int square;
typedef unsigned char   BYTE;
typedef unsigned int    ULONG;

// Index in a table
#if defined (T_INDEX64) && defined (_MSC_VER)
    typedef unsigned __int64    INDEX;
#   define HEX_INDEX_FORMAT "%016I64X"
#   define DEC_INDEX_FORMAT "%I64u"
#   define DEC_8_INDEX_FORMAT "%8I64u"
#elif defined (T_INDEX64)
    typedef unsigned long long INDEX;
//    typedef size_t INDEX;
#   define HEX_INDEX_FORMAT "%016llX"
#   define DEC_INDEX_FORMAT "%llu"
#   define DEC_8_INDEX_FORMAT "%8llu"
#else
    typedef unsigned long   INDEX;
#   define HEX_INDEX_FORMAT "%08X"
#   define DEC_INDEX_FORMAT "%lu"
#   define DEC_8_INDEX_FORMAT "%8lu"
#endif

#define MAX_PIECES  4       /* Maximum # of white or black pieces */
#define MAX_MOVES   256     /* Maximum # of pseudo legal moves in a position */

#define COLOR_DECLARED

typedef int color;

#define x_colorWhite    0
#define x_colorBlack    1
#define x_colorNeutral  2

#define PIECES_DECLARED

typedef int piece;

#define x_pieceNone     0
#define x_piecePawn     1
#define x_pieceKnight   2
#define x_pieceBishop   3
#define x_pieceRook     4
#define x_pieceQueen    5
#define x_pieceKing     6

enum
    {
    A1 = 0x00, B1, C1, D1, E1, F1, G1, H1,
    A2 = 0x10, B2, C2, D2, E2, F2, G2, H2,
    A3 = 0x20, B3, C3, D3, E3, F3, G3, H3,
    A4 = 0x30, B4, C4, D4, E4, F4, G4, H4,
    A5 = 0x40, B5, C5, D5, E5, F5, G5, H5,
    A6 = 0x50, B6, C6, D6, E6, F6, G6, H6,
    A7 = 0x60, B7, C7, D7, E7, F7, G7, H7,
    A8 = 0x70, B8, C8, D8, E8, F8, G8, H8,
    XX = 0x7F
    };

static const int    rgiRowEnPassant[] = { 3, 4 };
static const int    rgiPawnStep[] = { 0x10, -0x10 };
static const int    rgiPawnRow[] = { 1, 6 };    
static const int    rgiUnPawnRow[] = { 3, 4 };  
static const int    rgiPromoteRow[] = { 7, 0 }; 

static const bool   rgfSweep[] = { false, false, false, true, true, true, false };

static const square rgsqNextSquare[] =
    {
    B1, C1, D1, E1, F1, G1, H1, A2, XX, XX, XX, XX, XX, XX, XX, XX, 
    B2, C2, D2, E2, F2, G2, H2, A3, XX, XX, XX, XX, XX, XX, XX, XX, 
    B3, C3, D3, E3, F3, G3, H3, A4, XX, XX, XX, XX, XX, XX, XX, XX, 
    B4, C4, D4, E4, F4, G4, H4, A5, XX, XX, XX, XX, XX, XX, XX, XX, 
    B5, C5, D5, E5, F5, G5, H5, A6, XX, XX, XX, XX, XX, XX, XX, XX, 
    B6, C6, D6, E6, F6, G6, H6, A7, XX, XX, XX, XX, XX, XX, XX, XX, 
    B7, C7, D7, E7, F7, G7, H7, A8, XX, XX, XX, XX, XX, XX, XX, XX,
    B8, C8, D8, E8, F8, G8, H8, XX, XX, XX, XX, XX, XX, XX, XX, XX
    };

static const square rgsqNextSquarePawn[] =
    {
    XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, 
    B2, C2, D2, E2, F2, G2, H2, A3, XX, XX, XX, XX, XX, XX, XX, XX, 
    B3, C3, D3, E3, F3, G3, H3, A4, XX, XX, XX, XX, XX, XX, XX, XX, 
    B4, C4, D4, E4, F4, G4, H4, A5, XX, XX, XX, XX, XX, XX, XX, XX, 
    B5, C5, D5, E5, F5, G5, H5, A6, XX, XX, XX, XX, XX, XX, XX, XX, 
    B6, C6, D6, E6, F6, G6, H6, A7, XX, XX, XX, XX, XX, XX, XX, XX, 
    B7, C7, D7, E7, F7, G7, H7, XX, XX, XX, XX, XX, XX, XX, XX, XX
    };

static const square rgsqNextSquareTriangle[] =
    {
    B1, C1, D1, B2, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, 
    XX, C2, D2, C3, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, 
    XX, XX, D3, D4, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, 
    XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, 
    };

static const square rgsqNextSquareHalf[] =
    {
    B1, C1, D1, A2, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, 
    B2, C2, D2, A3, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, 
    B3, C3, D3, A4, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, 
    B4, C4, D4, A5, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, 
    B5, C5, D5, A6, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, 
    B6, C6, D6, A7, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, 
    B7, C7, D7, A8, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, 
    B8, C8, D8, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, 
    };

static const square rgsqFirstSquare[]   = { XX, A2, A1, A1, A1, A1, A1 };

static const piece rgpiPiecesLinear [] =
    {
    x_piecePawn,
    x_pieceKnight,
    x_pieceBishop,
    x_pieceRook,
    x_pieceQueen
    };

enum
    {
    x_mvCapture = 1,
    x_mvEnPassant = 2,
    x_mvPromoteKnight = 4,
    x_mvPromoteBishop = 8,
    x_mvPromoteRook = 16,
    x_mvPromoteQueen = 32
    };

const int x_mvPromotion = x_mvPromoteKnight |
                          x_mvPromoteBishop |
                          x_mvPromoteRook |
                          x_mvPromoteQueen;

typedef struct
    {
    square  m_sqFrom;
    square  m_sqTo;
    ULONG   m_ulFlags;
    int     m_iIndexFrom;
    int     m_iIndexTo;
    square  m_sqSaveEnPassant;
    piece   m_piCaptured;
    }
    CMove;

// Pointers to functions

class CBoard;
typedef bool (* PfnBool) (CBoard *pbd, color side, color xside);
#define PFNCALCINDEX_DECLARED
typedef INDEX (TB_FASTCALL *PfnCalcIndex) (square *psqW, square *psqB,
                                           square sqEnP, int fInverse);

// Statistics

static INDEX *prgStatistics;

// SMP

#if !defined (CPUS)
#define CPUS    1
#endif

int cCPUS;

// Useful macros

#define FValidSquare(sq)    (!((sq) & 0x88))
#define OtherSide(side)     (color) ((side) ^ 1)
#define Row(sq)             ((sq) >> 4)
#define Column(sq)          ((sq) & 7)
#define SquareMake(row,col) (((row) << 4) + (col))
#define UlMvFlags(sq)       (x_colorNeutral == m_rgColor[sq] ? 0 : x_mvCapture)
#define Linear(sq)          (rgindLinear [sq]) /* ((INDEX) (((sq) & 7) | (((sq) >> 1) & 0x38))) */
#define PchColor(side)      (x_colorWhite == side ? "white" : "black")
#define PchTm(side)         (x_colorWhite == side ? "wtm" : "btm")


///////////////////////////////////////////////////////////////////////////////

#if defined (T_INDEX64) || defined (T33_INCLUDE) || defined (T42_INCLUDE)
#  define   TB_DIRECTORY_SIZE       4096    /* # of cache buckets */
#else
#  define   TB_DIRECTORY_SIZE       256     /* # of cache buckets */
#endif

///////////////////////////////////////////////////////////////////////////////
//
//  Instruction used to copy the file

#if defined (_WIN32) || defined(_WIN64)
#define COPY_COMMAND    "copy "
#else
#define COPY_COMMAND    "cp "
#endif



///////////////////////////////////////////////////////////////////////////////
//
// File I/O primitives

#if defined (_WIN32) || defined(_WIN64)

typedef HANDLE file;

file f_open
    (
    const char  *szFile,
    const char  *szMode
    )
    {
    HANDLE  h;
    h = CreateFile (szFile, GENERIC_READ | GENERIC_WRITE, 0, NULL,
                    ('r' == szMode[0]) ? OPEN_ALWAYS : CREATE_ALWAYS,
                    FILE_ATTRIBUTE_NORMAL, NULL);
    if (INVALID_HANDLE_VALUE == h)
        {
        printf ("*** Unable to create or open file");
        fflush (stdout);
        exit (1);
        }
    return h;
    }

void f_close
    (
    file f
    )
    {
    if (0 == CloseHandle (f))
        {
        printf ("*** Close failed\n");
        fflush (stdout);
        exit (1);
        }
    }

size_t  f_read
    (
    void    *pv,
    size_t  cb,
    file    fp,
    INDEX   indStart
    )
    {
    LONG  lHiPos;
    DWORD dwPtr;
    BOOL  bResult;
    DWORD cbRead;

    if (sizeof (INDEX) == 4)
        lHiPos = 0;
    else
        lHiPos = (LONG) (indStart >> 32);
    dwPtr = SetFilePointer (fp, (LONG) indStart, &lHiPos, FILE_BEGIN);
    if (0xFFFFFFFF == dwPtr && NO_ERROR != GetLastError())
        {
        printf ("*** Seek failed: pos is %I64x code is %lx\n",
                indStart, GetLastError());
        fflush (stdout);
        exit (1);
        return 0;
        }
    bResult = ReadFile (fp, pv, cb, &cbRead, NULL);
    if (0 == bResult)
        {
        printf ("*** Read failed\n");
        fflush (stdout);
        exit (1);
        return 0;
        }
    return cbRead;
    }

void f_write
    (
    void    *pv,
    size_t  cb,
    file    fp,
    INDEX   indStart
    )
    {
    LONG  lHiPos;
    DWORD dwPtr;
    BOOL  bResult;
    DWORD cbWritten;

    if (sizeof (INDEX) == 4)
        lHiPos = 0;
    else
        lHiPos = (LONG) (indStart >> 32);
    dwPtr = SetFilePointer (fp, (LONG) indStart, &lHiPos, FILE_BEGIN);
    if (0xFFFFFFFF == dwPtr && NO_ERROR != GetLastError())
        {
        printf ("*** Seek failed\n");
        fflush (stdout);
        exit (1);
        return;
        }
    bResult = WriteFile (fp, pv, cb, &cbWritten, NULL);
    if (0 == bResult || cbWritten != cb)
        {
        printf ("*** Write failed\n");
        fflush (stdout);
        exit (1);
        return;
        }
    }

#else

typedef FILE *file;

file f_open
    (
    const char  *szFile,
    const char  *szMode
    )
    {
    FILE *fp;

    fp = fopen (szFile, szMode);
    if (NULL == fp)
        {
        printf ("*** Unable to create or open file");
        fflush (stdout);
        exit (1);
        }
    return fp;
    }

void f_close
    (
    file f
    )
    {
    if (0 != fclose (f))
        {
        printf ("*** Close failed\n");
        fflush (stdout);
        exit (1);
        }
    }

size_t  f_read
    (
    void    *pv,
    size_t  cb,
    file    fp,
    INDEX   indStart
    )
    {
    size_t  cbRead;

    if (0 != fseek (fp, (long) indStart, SEEK_SET))
        {
        printf ("*** Seek failed\n");
        fflush (stdout);
        exit (1);
        return 0;
        }
    cbRead = fread (pv, 1, cb, fp);
    if ((size_t) -1 == cbRead || ferror (fp))
        {
        printf ("*** Read failed\n");
        fflush (stdout);
        exit (1);
        return 0;
        }
    return cbRead;
    }

void f_write
    (
    void    *pv,
    INDEX   cb,
    file    fp,
    INDEX   indStart
    )
    {
    if (0 != fseek (fp, (size_t) indStart, SEEK_SET))
        {
        printf ("*** Seek failed\n");
        fflush (stdout);
        exit (1);
        return;
        }
    if (cb != (unsigned long long) fwrite (pv, 1, cb, fp))
        {
        printf ("*** Write failed?\n");
        std::cout << "Wrote " << cb << " bytes" << std::endl;
        fflush (stdout);
//        exit (1);
//        return;
        }
    }

#endif

// Cannot declare it later

static  size_t  cbMemoryUsed;

// Bit vectors

#if defined (SMP)
static lock_t   rgBitLocks [65536]; // Length must be power of 2
pthread_t       threads[CPUS];
#endif

// For 32-bit Windows we cannot allocate 2 contigious huge
// areas of memory, so partition it into smaller pieces

#if defined (_M_IX86) && defined (T_INDEX64)

#define PARTITION_SIZE      (128*1024*1024) /* Power of 2 */
#define ULONG_IN_PARTITION  (PARTITION_SIZE/sizeof(ULONG))

typedef class {
public:
           void  VInit (size_t cBytes);
           void  VZero (INDEX cBits);
    INLINE void  VSetBit (INDEX ind);
    INLINE void  VClearBit (INDEX ind);
    INLINE void  VSetBitUnlocked (INDEX ind);
    INLINE ULONG FGetBit (INDEX ind);
    INLINE BYTE  *PbData ();
    INLINE size_t CbData();
private:
    ULONG   **m_ppulData;
} CBitVector;

BYTE *CBitVector::PbData ()
    {
    return (BYTE*) m_ppulData[0];
    }

size_t CBitVector::CbData ()
    {
    return min (cbMemoryUsed/2, PARTITION_SIZE);
    }

void CBitVector::VInit
    (
    size_t  cBytes
    )
    {
    size_t  i, cb, cPartitions;
    
    cPartitions = (cBytes+PARTITION_SIZE-1)/PARTITION_SIZE;
    m_ppulData = (ULONG**) malloc(sizeof(ULONG*)*cPartitions);
    if (NULL == m_ppulData)
        {
        printf ("Unable to allocate memory");
        fflush (stdout);
        exit (1);
        }
    for (i = 0; i < cPartitions; i ++)
        {
        cb = min (PARTITION_SIZE, cBytes);
        m_ppulData[i] = (ULONG*) malloc(cb);
        if (NULL == m_ppulData[i])
            {
            printf ("Unable to allocate %luMb of memory", cb/(1024*1024));
            fflush (stdout);
            exit (1);
            }
        cBytes -= PARTITION_SIZE;
        }
    }

INLINE void CBitVector::VSetBit
    (
    INDEX   ind
    )
    {
    ULONG   bit = ((ULONG) ind) & 31;
    ULONG   offs = ((ULONG) (ind >> 5)) & (ULONG_IN_PARTITION-1);
    ULONG   partition = ((ULONG) (ind >> 5)) / ULONG_IN_PARTITION;
#if defined (SMP)
    ULONG   iLock = offs & (sizeof (rgBitLocks) / sizeof (lock_t) - 1);
#endif
    
    Lock (rgBitLocks [iLock]);
    m_ppulData [partition][offs] |= (1 << bit);
    UnLock (rgBitLocks [iLock]);
    }

INLINE void CBitVector::VSetBitUnlocked
    (
    INDEX   ind
    )
    {
    ULONG   bit = ((ULONG) ind) & 31;
    ULONG   offs = ((ULONG) (ind >> 5)) & (ULONG_IN_PARTITION-1);
    ULONG   partition = ((ULONG) (ind >> 5)) / ULONG_IN_PARTITION;

    m_ppulData [partition][offs] |= (1 << bit);
    }

INLINE void CBitVector::VClearBit
    (
    INDEX   ind
    )
    {
    ULONG   bit = ((ULONG) ind) & 31;
    ULONG   offs = ((ULONG) (ind >> 5)) & (ULONG_IN_PARTITION-1);
    ULONG   partition = ((ULONG) (ind >> 5)) / ULONG_IN_PARTITION;
#if defined (SMP)
    ULONG   iLock = offs & (sizeof (rgBitLocks) / sizeof (lock_t) - 1);
#endif
    
    Lock (rgBitLocks [iLock]);
    m_ppulData [partition][offs] &= ~(1 << bit);
    UnLock (rgBitLocks [iLock]);
    }

INLINE ULONG CBitVector::FGetBit
    (
    INDEX   ind
    )
    {
    ULONG   bit = ((ULONG) ind) & 31;
    ULONG   offs = ((ULONG) (ind >> 5)) & (ULONG_IN_PARTITION-1);
    ULONG   partition = ((ULONG) (ind >> 5)) / ULONG_IN_PARTITION;
    
    return m_ppulData [partition][offs] & (1 << bit);
    }

void CBitVector::VZero (INDEX cBits)
    {
    size_t cBytes;
    size_t i;

    cBytes = (cBits+sizeof(ULONG)*8-1)/(sizeof(ULONG)*8)*4;
    i = 0;
    while (cBytes > PARTITION_SIZE)
        {
        memset (m_ppulData[i], 0, PARTITION_SIZE);
        cBytes -= PARTITION_SIZE;
        i ++;
        }
    memset (m_ppulData[i], 0, cBytes);
    }

#else

typedef class {
public:
           void  VInit (size_t cBytes);
           void  VZero (INDEX cBits);
    INLINE void  VSetBit (INDEX ind);
    INLINE void  VClearBit (INDEX ind);
    INLINE void  VSetBitUnlocked (INDEX ind);
    INLINE ULONG FGetBit (INDEX ind);
    INLINE BYTE  *PbData ();
    INLINE size_t CbData();
private:
    ULONG   *m_pulData;
} CBitVector;

size_t min (size_t a, size_t b){
	return (a>b)?b:a;
}

BYTE *CBitVector::PbData ()
    {
    return (BYTE*) m_pulData;
    }

size_t CBitVector::CbData ()
    {
    return min (cbMemoryUsed/2, 256*1024*1024);
    }

void CBitVector::VInit
    (
    size_t  cBytes
    )
    {
    m_pulData = (ULONG*) malloc(cBytes);
    if (NULL == m_pulData)
        {
        printf ("Unable to allocate %luMb of memory", cBytes/(1024*1024));
        fflush (stdout);
        exit (1);
        }
    }

INLINE void CBitVector::VSetBit
    (
    INDEX   ind
    )
    {
    ULONG   bit = ((ULONG) ind) & 31;
    ULONG   offs = (ULONG) (ind >> 5);
#if defined (SMP)
    ULONG   iLock = offs & (sizeof (rgBitLocks) / sizeof (lock_t) - 1);
#endif
    
    Lock (rgBitLocks [iLock]);
    m_pulData [offs] |= (1 << bit);
    UnLock (rgBitLocks [iLock]);
    }

INLINE void CBitVector::VSetBitUnlocked
    (
    INDEX   ind
    )
    {
    m_pulData [ind >> 5] |= (1 << (ind & 31));
    }

INLINE void CBitVector::VClearBit
    (
    INDEX   ind
    )
    {
    ULONG   bit = ((ULONG) ind) & 31;
    ULONG   offs = (ULONG) (ind >> 5);
#if defined (SMP)
    ULONG   iLock = offs & (sizeof (rgBitLocks) / sizeof (lock_t) - 1);
#endif
    
    Lock (rgBitLocks [iLock]);
    m_pulData [offs] &= ~(1 << bit);
    UnLock (rgBitLocks [iLock]);
    }

INLINE ULONG CBitVector::FGetBit
    (
    INDEX   ind
    )
    {
    ULONG   bit = ((ULONG) ind) & 31;
    ULONG   offs = (ULONG) (ind >> 5);
    
    return m_pulData [offs] & (1 << bit);
    }

void CBitVector::VZero (INDEX cBits)
    {
    memset (m_pulData, 0, (cBits+sizeof(ULONG)*8-1)/(sizeof(ULONG)*8)*4);
    }

#endif


// 2 vectors - one is new accumulated, and one is argument

static CBitVector bvArg;
static CBitVector bvNew;
static size_t   cbv;        // Size of bit vector in ULONGs
static INDEX    indBVSize;  // Size of bit vector in # of positions

// Structure that is passed to SqFind* functions

typedef struct
    {
    square  *m_pSq;
    piece   *m_pBoard;
    }
    CSqInfo;

// Structure for storing results of minor probing (bit vector approach only)

typedef struct
    {
    INDEX   m_ind;
    int     m_tbt;
    }
    CMinorResult;

// Board representation

class CBoard
    {
public:
    // Data
    square          m_rgsqPieceList[2][MAX_PIECES];
    int             m_rgPieceCount[2];
    int             m_rgiCount[10];
    color           m_rgColor [120];
    piece           m_rgBoard [120];
    square          m_sqEnPassant;
    piece           m_rgPieces[2][MAX_PIECES];
    bool            m_rgfHasPawns[2];
    bool            m_fBothSidesHavePawns;
    bool            m_fReverse;
    const square    *m_rgpsqNext[2][MAX_PIECES];
    PfnCalcIndex    m_rgpfnCalcIndex[2];

    // For determining file size
    INDEX           m_cb;

    // Bit vector stuff
#if !defined(NDEBUG)
    unsigned        m_uGuard;
#endif
    CMinorResult    *m_pmr;         // Accumulated minor results
    ULONG           m_cmr;          // # of those results
    ULONG           m_cmrMax;       // Maximum # of those results
    file            m_fp;           // File where to write results of probing minors
    BYTE            *m_pbBuf;       // Buffer for writing minor results to disk
    ULONG           m_cbBuf;        // Length of that buffer
    int             m_cWinMax;      // Longest win after minor probe
    int             m_cLostMax;     // Longest loss after minor probe

    // Search stuff
    unsigned        m_cSearch;

    // SMP stuff
#if defined (SMP)
    volatile bool       m_fHasWork;
    volatile color      m_side;
    volatile PfnBool    m_pfn;
#endif
    CMove           m_rgmv [MAX_MOVES];

    // Methods
    INLINE  INDEX   IndCalc (color side);
    INLINE  INDEX   IndCalc (int iTb, color side, square *psqW, square *psqB, square sqEnP, int fInverse);
    INLINE  void    VUpdateOnCapture (CMove *pmv, square sqTo, color xside, piece pi);
    INLINE  void    VRestoreOnCapture (CMove *pmv, square sqTo, color xside, piece pi);
    INLINE  void    VSetEnPassant (square sqFrom, square sqTo, color side);
            void    VMakeMove (CMove *pmv);
            void    VUnmakeMove (CMove *pmv);
    INLINE  void    VMinimalMakeQuietMove (CMove *pmv, color side);
    INLINE  void    VMinimalMakeQuietUnMove (CMove *pmv, color side);
    INLINE  void    VUnmakeMinimalQuietMove (CMove *pmv, color side);
    INLINE  bool    FAtacked (square sqToCheck, color side);
    INLINE  CMove   *PmvGenPseudoLegalDirection (CMove *pmv, color side, square sqFrom, int iIndexFrom, int iStep);
            CMove   *PmvGenPseudoLegalMoves (CMove *pmv, color side);
    INLINE  CMove   *PmvGenPseudoLegalQuietDirection (CMove *pmv, color side, square sqFrom, int iIndexFrom, int iStep);
    INLINE  CMove   *PmvGenPseudoLegalQuietMovesForPiece (CMove *pmv, int iIndexFrom, color side, color xside);
    INLINE  CMove   *PmvGenPseudoLegalQuietUnMovesForPiece (CMove *pmv, int iIndexFrom, color side, color xside);
    INLINE  void    VPutPiece (color side, square sq, piece pi, int index);
            void    VPrintFEN (color side);
            void    VPrintPosition ();
            void    VInitIterations (const char *pchTable);
            void    VQuery ();
    };

//-----------------------------------------------------------------------------
// Calculate index.
// Functions are in separate file, include it here.

static const unsigned char rgindLinear[] =
    {
     0,  1,  2,  3,  4,  5,  6,  7, XX, XX, XX, XX, XX, XX, XX, XX,
     8,  9, 10, 11, 12, 13, 14, 15, XX, XX, XX, XX, XX, XX, XX, XX,
    16, 17, 18, 19, 20, 21, 22, 23, XX, XX, XX, XX, XX, XX, XX, XX,
    24, 25, 26, 27, 28, 29, 30, 31, XX, XX, XX, XX, XX, XX, XX, XX,
    32, 33, 34, 35, 36, 37, 38, 39, XX, XX, XX, XX, XX, XX, XX, XX,
    40, 41, 42, 43, 44, 45, 46, 47, XX, XX, XX, XX, XX, XX, XX, XX,
    48, 49, 50, 51, 52, 53, 54, 55, XX, XX, XX, XX, XX, XX, XX, XX,
    56, 57, 58, 59, 60, 61, 62, 63, XX, XX, XX, XX, XX, XX, XX, XX
    };

// Find piece in a list
// UNDONE:  Make a specialized version for generator;
//          It can use the fact that pieces in a list are sorted.

// NOTE: Code assuming that we'll produce MAX 5-man tables,
// so shall try only 3 elements max

#if defined (T41_INCLUDE) || defined (T42_INCLUDE)

INLINE square SqFindFirst
    (
    square  *pv,
    piece   pi
    )
    {
    CSqInfo *psqInfo;
    square  sq;

    psqInfo = (CSqInfo*) pv;
    sq = psqInfo->m_pSq[1];
    if (psqInfo->m_pBoard[sq] != pi)
        {
        sq = psqInfo->m_pSq[2];
        if (psqInfo->m_pBoard[sq] != pi)
            {
            sq = psqInfo->m_pSq[3];
            Assert (psqInfo->m_pBoard[sq] == pi);
            }
        }
    Assert (FValidSquare (sq));
    return Linear (sq);
    }

INLINE square SqFindSecond
    (
    square  *pv,
    piece   pi
    )
    {
    CSqInfo *psqInfo;
    square  *psq;

    psqInfo = (CSqInfo*) pv;
    psq = psqInfo->m_pSq;
    while (psqInfo->m_pBoard[psq[1]] != pi)
        psq ++;
    while (psqInfo->m_pBoard[psq[2]] != pi)
        psq ++;
    Assert (FValidSquare (psq[2]));
    return Linear (psq[2]);
    }

#else

INLINE square SqFindFirst
    (
    square  *pv,
    piece   pi
    )
    {
    CSqInfo *psqInfo;
    square  sq;

    psqInfo = (CSqInfo*) pv;
    sq = psqInfo->m_pSq[1];
    if (psqInfo->m_pBoard[sq] != pi)
        {
        sq = psqInfo->m_pSq[2];
        Assert (psqInfo->m_pBoard[sq] == pi);
        }
    Assert (FValidSquare (sq));
    return Linear (sq);
    }

#define SqFindSecond(pv, pi)    Linear (((CSqInfo*)pv)->m_pSq[2])

#endif

#define SqFindKing(pv)              Linear (((CSqInfo*)pv)->m_pSq[0])
#define SqFindOne(pv, pi)           Linear (((CSqInfo*)pv)->m_pSq[1])
#define SqFindThird(pv, pi)         Linear (((CSqInfo*)pv)->m_pSq[3])
#define SqFind2(pv,pi1,sq1,pi2,sq2)\
            {\
            CSqInfo *psqInfo;\
            square  *psq;\
            piece   *pboard;\
            \
            psqInfo = (CSqInfo*) pv;\
            psq = psqInfo->m_pSq;\
            pboard = psqInfo->m_pBoard;\
            if (pi1==pboard[psq[1]])\
                {\
                sq1 = psq[1];\
                sq2 = psq[2];\
                }\
            else\
                {\
                sq2 = psq[1];\
                sq1 = psq[2];\
                }\
            sq1 = Linear(sq1);\
            sq2 = Linear(sq2);\
            }

#define ILLEGAL_POSSIBLE

#include "tbindex.cpp"

// For bitvector stuff only

#define bev_unknown     bev_miminx
#define L_bev_unknown   L_bev_miminx

// Global data

static  INDEX   cbCalc;
static  INDEX   rgCb[2];
static  file    rgFp[2];
static  BYTE    *rgpbFile[2];

static  ULONG   cbMemoryCache;
static  BYTE    *pbMemory;
static  BYTE    *pbCache;

static  BYTE    *pbBuf;
static  ULONG   cbBuf;
static  ULONG   *pulGuardBuf;

static bool fMapFiles  = false;
static bool fBitVectors = false;
static bool fReverse = false;
static bool fResume = false;
static bool fSaveStatus = false;
static char *pszSavePref;
static char *pszResumePref;

static CBoard   rgBoards[CPUS];

// Globals of main pass

static  bool    fChanged;
static  tb_t    tbtLo;
static  tb_t    tbtHi;
static  int     L_tbtLo;
static  int     L_tbtHi;

// TB search globals

static color colSearch;
static int   tbtSearch;

// Replace "result in N" by "Result in N+1"

#define TbtReplaceResult(tbt)\
    (tb_t) ((bev_mimin == tbt || bev_miminx == tbt) ? bev_limaxx :\
            (bev_limaxx == tbt) ? bev_miminx :\
            (((tbt) > 0) - (tbt)))

#define L_TbtReplaceResult(tbt) (((tbt) > 0) - (tbt))

/* (((tbt) <= 0) ? (-tbt) : (1-tbt)) */

// Does new score better than current best?

#define FBetter(best, score)\
    (bev_miminx == best ? (score > 0) :\
     bev_limaxx == best ? (score >= 0 || bev_miminx == score) :\
     bev_miminx == score ? (best <= 0) :\
     bev_limaxx == score ? (best < 0 && bev_miminx != score) :\
     (best < score))

#define L_FBetter(best, score)  (best < score)

// Get 16-bit score in a machine-independent way

#define L_Get(pb)   ((((signed char *) (pb)) [1] << 8) + ((BYTE *) (pb)) [0])

// Is square atacked?

static ULONG    rgulAtacks [(H8 - A1) * 2 + 1];
static int      rgiStep [(H8 - A1) * 2 + 1];

INLINE bool CBoard::FAtacked
    (
    square  sqToCheck,  // IN | Square to check (typically king's square)
    color   side        // IN | Side to atack
    )
    {
    int i;

    i = m_rgPieceCount[side]-1;
    do
        {
        square  sq;
        piece   pi;

        sq = m_rgsqPieceList[side][i];
        pi = m_rgBoard[sq];
        if ((x_piecePawn == pi) && (x_colorBlack == side))
            pi = x_pieceNone;
        if (rgulAtacks [H8 + sq - sqToCheck] & (1 << pi))
            {
            if (rgfSweep[pi])
                {
                int iStep = rgiStep [H8 + sq - sqToCheck];

                for (square sqTemp = sq + iStep;; sqTemp += iStep)
                    {
                    if (x_pieceNone != m_rgBoard[sqTemp])
                        {
                        if (sqTemp != sqToCheck)
                            goto next;
                        break;
                        }
                    }
                }
            return true;
            }
next:   i --;
        }
    while (i >= 0);
    return false;
    }

// Initialize one direction in atacks detection tables

static void VInitializeDirection
    (
    int     iStep,
    piece   pi,
    int     count
    )
    {
    for (int iOffset = iStep; count != 0; iOffset += iStep, count --)
        {
        rgulAtacks[H8 - iOffset] |= (1 << pi);
        rgiStep[H8 - iOffset] = iStep;
        }
    }
    
// Initialize atacks detection tables

static void VInitializeAttacks (void)
    {
    // White pawns
    VInitializeDirection (0x0F, x_piecePawn, 1);
    VInitializeDirection (0x11, x_piecePawn, 1);

    // Black pawns
    VInitializeDirection (-0x0F, x_pieceNone, 1);
    VInitializeDirection (-0x11, x_pieceNone, 1);

    // Knights
    VInitializeDirection (0x12, x_pieceKnight, 1);
    VInitializeDirection (0x21, x_pieceKnight, 1);
    VInitializeDirection (0x0E, x_pieceKnight, 1);
    VInitializeDirection (0x1F, x_pieceKnight, 1);
    VInitializeDirection (-0x12, x_pieceKnight, 1);
    VInitializeDirection (-0x21, x_pieceKnight, 1);
    VInitializeDirection (-0x0E, x_pieceKnight, 1);
    VInitializeDirection (-0x1F, x_pieceKnight, 1);

    // Bishops
    VInitializeDirection (0x11, x_pieceBishop, 7);
    VInitializeDirection (0x0F, x_pieceBishop, 7);
    VInitializeDirection (-0x11, x_pieceBishop, 7);
    VInitializeDirection (-0x0F, x_pieceBishop, 7);

    // Rooks
    VInitializeDirection (0x01, x_pieceRook, 7);
    VInitializeDirection (0x10, x_pieceRook, 7);
    VInitializeDirection (-0x01, x_pieceRook, 7);
    VInitializeDirection (-0x10, x_pieceRook, 7);

    // Queens
    VInitializeDirection (0x01, x_pieceQueen, 7);
    VInitializeDirection (0x10, x_pieceQueen, 7);
    VInitializeDirection (0x11, x_pieceQueen, 7);
    VInitializeDirection (0x0F, x_pieceQueen, 7);
    VInitializeDirection (-0x01, x_pieceQueen, 7);
    VInitializeDirection (-0x10, x_pieceQueen, 7);
    VInitializeDirection (-0x11, x_pieceQueen, 7);
    VInitializeDirection (-0x0F, x_pieceQueen, 7);

    // Kings
    VInitializeDirection (0x01, x_pieceKing, 1);
    VInitializeDirection (0x10, x_pieceKing, 1);
    VInitializeDirection (0x11, x_pieceKing, 1);
    VInitializeDirection (0x0F, x_pieceKing, 1);
    VInitializeDirection (-0x01, x_pieceKing, 1);
    VInitializeDirection (-0x10, x_pieceKing, 1);
    VInitializeDirection (-0x11, x_pieceKing, 1);
    VInitializeDirection (-0x0F, x_pieceKing, 1);
    }

// Update PieceList when capture happened

INLINE void CBoard::VUpdateOnCapture
    (
    CMove   *pmv,
    square  sqTo,
    color   xside,
    piece   pi
    )
    {
    int i;

    m_rgiCount[5*xside+pi-x_piecePawn] --;
    for (i = 1; m_rgsqPieceList[xside][i] != sqTo; i ++)
        ;
    pmv->m_iIndexTo = i;
    for (i ++; i < m_rgPieceCount[xside]; i ++)
        m_rgsqPieceList[xside][i-1] = m_rgsqPieceList[xside][i];
    m_rgPieceCount[xside] --;
    }

// Set en passant square

INLINE void CBoard::VSetEnPassant
    (
    square  sqFrom,
    square  sqTo,
    color   side
    )
    {
    m_sqEnPassant = XX;
    if (x_piecePawn == m_rgBoard[sqTo])
        {
        if (0!=Column(sqTo) && x_piecePawn==m_rgBoard[sqTo-1] && side!=m_rgColor[sqTo-1] ||
            7!=Column(sqTo) && x_piecePawn==m_rgBoard[sqTo+1] && side!=m_rgColor[sqTo+1])
            {
            if (1 == Row(sqFrom) && 3 == Row(sqTo))
                m_sqEnPassant = sqFrom + 0x10;
            else if (6 == Row(sqFrom) && 4 == Row(sqTo))
                m_sqEnPassant = sqFrom - 0x10;
            }
        }
    }

// Make one move

void CBoard::VMakeMove
    (
    CMove   *pmv
    )
    {
    square  sqFrom = pmv->m_sqFrom;
    square  sqTo = pmv->m_sqTo;
    color   side = m_rgColor[sqFrom];

    Assert (m_rgColor[sqFrom] == side);
    Assert (x_pieceNone != m_rgBoard[sqFrom] && x_pieceKing != m_rgBoard[sqTo]);
    Assert (m_rgColor[sqTo] != side);

    // Save info for UnamkeMove
    pmv->m_sqSaveEnPassant = m_sqEnPassant;
    pmv->m_piCaptured = m_rgBoard[sqTo];
    m_rgsqPieceList[side][pmv->m_iIndexFrom] = sqTo;

    // Modify board structures
    if (pmv->m_ulFlags & x_mvCapture)
        VUpdateOnCapture (pmv, sqTo, OtherSide (side), m_rgBoard[sqTo]);
    else if (pmv->m_ulFlags & x_mvEnPassant)
        {
        square  sqEnP = SquareMake (Row (sqFrom), Column (sqTo));

        VUpdateOnCapture (pmv, sqEnP, OtherSide (side), x_piecePawn);
        m_rgColor[sqEnP] = x_colorNeutral;
        m_rgBoard[sqEnP] = x_pieceNone;
        }
    m_rgColor[sqTo] = side;
    m_rgBoard[sqTo] = m_rgBoard[sqFrom];
    m_rgColor[sqFrom] = x_colorNeutral;
    m_rgBoard[sqFrom] = x_pieceNone;

    // Fix necessary data if promotion
    if (pmv->m_ulFlags & x_mvPromotion)
        {
        m_rgBoard[sqTo] = pmv->m_ulFlags & x_mvPromoteKnight ? x_pieceKnight :
                        pmv->m_ulFlags & x_mvPromoteBishop ? x_pieceBishop :
                        pmv->m_ulFlags & x_mvPromoteRook ? x_pieceRook : x_pieceQueen;
        m_rgiCount[5*side+m_rgBoard[sqTo]-x_piecePawn] ++;
        m_rgiCount[5*side+x_piecePawn-x_piecePawn] --;
        }

    // Set en passant
    VSetEnPassant (sqFrom, sqTo, side);
    }

// Restore PieceList, Color and Board after capture

INLINE void CBoard::VRestoreOnCapture
    (
    CMove   *pmv,
    square  sqTo,
    color   xside,
    piece   pi
    )
    {
    int     i;

    for (i = (m_rgPieceCount[xside] ++); i > pmv->m_iIndexTo; i --)
        m_rgsqPieceList[xside][i] = m_rgsqPieceList[xside][i-1];
    m_rgsqPieceList[xside][i] = sqTo;
    m_rgColor[sqTo] = xside;
    m_rgBoard[sqTo] = pi;
    m_rgiCount[5*xside+pi-x_piecePawn] ++;
    }

// Unmake one move

void CBoard::VUnmakeMove
    (
    CMove   *pmv
    )
    {
    square  sqFrom = pmv->m_sqFrom;
    square  sqTo = pmv->m_sqTo;
    color   side = m_rgColor[sqTo];

    m_sqEnPassant = pmv->m_sqSaveEnPassant;

    // Restore 'from' square
    m_rgColor[sqFrom] = side;
    m_rgBoard[sqFrom] = m_rgBoard[sqTo];
    m_rgsqPieceList[side][pmv->m_iIndexFrom] = sqFrom;

    // Now restore 'to' and possible en passant square
    if (pmv->m_ulFlags & x_mvCapture)
        VRestoreOnCapture (pmv, sqTo, OtherSide (side), pmv->m_piCaptured);
    else
        {
        if (pmv->m_ulFlags & x_mvEnPassant)
            VRestoreOnCapture (pmv, SquareMake (Row (sqFrom), Column (sqTo)), OtherSide (side), x_piecePawn);
        m_rgColor[sqTo] = x_colorNeutral;
        m_rgBoard[sqTo] = x_pieceNone;
        }

    // Fix board data, if promotion
    if (pmv->m_ulFlags & x_mvPromotion)
        {
        m_rgiCount[5*side+m_rgBoard[sqFrom]-x_piecePawn] --;
        m_rgiCount[5*side+x_piecePawn-x_piecePawn] ++;
        m_rgBoard[sqFrom] = x_piecePawn;
        }

    Assert (m_rgColor[sqFrom] == side);
    Assert (x_pieceNone != m_rgBoard[sqFrom]);
    Assert (m_rgColor[sqTo] != side);
    }

// Add pseudo legal move to the list

#define VAddMove(pmv,sqFrom,sqTo,iIndexFrom,ulFlags) (pmv)->m_sqFrom=(sqFrom),(pmv)->m_sqTo=(sqTo),(pmv)->m_ulFlags=(ulFlags),(pmv)->m_iIndexFrom=(iIndexFrom)
#define VAddQuietMove(pmv,sqFrom,sqTo,iIndexFrom) (pmv)->m_sqFrom=(sqFrom),(pmv)->m_sqTo=(sqTo),(pmv)->m_ulFlags=0,(pmv)->m_iIndexFrom=(iIndexFrom)

// Generate all pseudo legal moves in one direction

INLINE CMove* CBoard::PmvGenPseudoLegalDirection
    (
    CMove   *pmv,
    color   side,
    square  sqFrom,
    int     iIndexFrom,
    int     iStep
    )
    {
    for (square sq = sqFrom + iStep; FValidSquare(sq); sq += iStep)
        {
        color col = m_rgColor[sq];

        if (x_colorNeutral == col)
            {
            VAddMove (pmv, sqFrom, sq, iIndexFrom, 0);  // Non-capture
            pmv ++;
            }
        else
            {
            if (side != col)
                {
                VAddMove (pmv, sqFrom, sq, iIndexFrom, x_mvCapture);
                pmv ++;
                }
            break;
            }
        }
    return pmv;
    }

// Generate all quiet pseudo legal moves in one direction

INLINE CMove* CBoard::PmvGenPseudoLegalQuietDirection
    (
    CMove   *pmv,
    color   side,
    square  sqFrom,
    int     iIndexFrom,
    int     iStep
    )
    {
    for (square sq = sqFrom + iStep; FValidSquare(sq) && x_colorNeutral == m_rgColor[sq]; sq += iStep)
        {
        VAddQuietMove (pmv, sqFrom, sq, iIndexFrom);
        pmv ++;
        }
    return pmv;
    }

// Add pawn move to the list

INLINE CMove* PmvAddPawnMove
    (
    CMove   *pmv,
    color   side,
    square  sqFrom,
    square  sqTo,
    int     iIndexFrom,
    ULONG   ulFlags
    )
    {
    if (Row(sqTo) == rgiPromoteRow[side])
        {
        VAddMove (pmv, sqFrom, sqTo, iIndexFrom, ulFlags | x_mvPromoteKnight);
        VAddMove (pmv + 1, sqFrom, sqTo, iIndexFrom, ulFlags | x_mvPromoteBishop);
        VAddMove (pmv + 2, sqFrom, sqTo, iIndexFrom, ulFlags | x_mvPromoteRook);
        ulFlags |= x_mvPromoteQueen;
        pmv += 3;
        }
    VAddMove (pmv, sqFrom, sqTo, iIndexFrom, ulFlags);
    return pmv + 1;
    }

// Generate all pseudo legal unmoves for one piece

INLINE CMove* CBoard::PmvGenPseudoLegalQuietUnMovesForPiece
    (
    CMove   *pmv,
    int     iIndexFrom,
    color   side,
    color   xside
    )
    {
    square  sq = m_rgsqPieceList[side][iIndexFrom];
    piece   pi = m_rgBoard[sq];

    switch (pi)
        {
    case x_piecePawn:
        {
        if (Row (sq) != rgiPawnRow[side])
            {
            int iPawnStep = rgiPawnStep[xside];

            if (x_colorNeutral == m_rgColor[sq + iPawnStep])
                {
                VAddQuietMove (pmv, sq, sq + iPawnStep, iIndexFrom);
                pmv ++;
                if (Row (sq) == rgiUnPawnRow[side] && x_colorNeutral == m_rgColor[sq + 2*iPawnStep])
                    {
                    VAddQuietMove (pmv, sq, sq + 2*iPawnStep, iIndexFrom);
                    pmv ++;
                    }
                }
            }
        }
        break;
    case x_pieceKnight:
        if (FValidSquare (sq + 0x12) && m_rgColor[sq + 0x12] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq + 0x12, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq + 0x21) && m_rgColor[sq + 0x21] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq + 0x21, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq + 0x0E) && m_rgColor[sq + 0x0E] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq + 0x0E, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq + 0x1F) && m_rgColor[sq + 0x1F] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq + 0x1F, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq - 0x12) && m_rgColor[sq - 0x12] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq - 0x12, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq - 0x21) && m_rgColor[sq - 0x21] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq - 0x21, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq - 0x0E) && m_rgColor[sq - 0x0E] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq - 0x0E, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq - 0x1F) && m_rgColor[sq - 0x1F] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq - 0x1F, iIndexFrom);
            pmv ++;
            }
        break;
    case x_pieceBishop:
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, 0x11);
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, 0x0F);
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, -0x11);
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, -0x0F);
        break;
    case x_pieceQueen:
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, 0x11);
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, 0x0F);
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, -0x11);
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, -0x0F);
        // Fall through
    case x_pieceRook:
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, 0x01);
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, 0x10);
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, -0x01);
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, -0x10);
        break;
    case x_pieceKing:
        if (FValidSquare (sq + 0x01) && m_rgColor[sq + 0x01] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq + 0x01, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq + 0x10) && m_rgColor[sq + 0x10] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq + 0x10, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq + 0x11) && m_rgColor[sq + 0x11] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq + 0x11, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq + 0x0F) && m_rgColor[sq + 0x0F] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq + 0x0F, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq - 0x01) && m_rgColor[sq - 0x01] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq - 0x01, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq - 0x10) && m_rgColor[sq - 0x10] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq - 0x10, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq - 0x11) && m_rgColor[sq - 0x11] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq - 0x11, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq - 0x0F) && m_rgColor[sq - 0x0F] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq - 0x0F, iIndexFrom);
            pmv ++;
            }
        break;
    default:
        Assert (0);
        }
    return pmv;
    }

// Generate all pseudo legal moves for one piece

INLINE CMove* CBoard::PmvGenPseudoLegalQuietMovesForPiece
    (
    CMove   *pmv,
    int     iIndexFrom,
    color   side,
    color   xside
    )
    {
    square  sq = m_rgsqPieceList[side][iIndexFrom];
    piece   pi = m_rgBoard[sq];

    switch (pi)
        {
    case x_piecePawn:
        {
        if (Row (sq) != rgiPawnRow[xside])
            {
            int iPawnStep = rgiPawnStep[side];

            if (x_colorNeutral == m_rgColor[sq + iPawnStep])
                {
                VAddQuietMove (pmv, sq, sq + iPawnStep, iIndexFrom);
                pmv ++;
                if (Row (sq) == rgiPawnRow[side] && x_colorNeutral == m_rgColor[sq + 2*iPawnStep])
                    {
                    VAddQuietMove (pmv, sq, sq + 2*iPawnStep, iIndexFrom);
                    pmv ++;
                    }
                }
            }
        }
        break;
    case x_pieceKnight:
        if (FValidSquare (sq + 0x12) && m_rgColor[sq + 0x12] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq + 0x12, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq + 0x21) && m_rgColor[sq + 0x21] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq + 0x21, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq + 0x0E) && m_rgColor[sq + 0x0E] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq + 0x0E, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq + 0x1F) && m_rgColor[sq + 0x1F] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq + 0x1F, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq - 0x12) && m_rgColor[sq - 0x12] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq - 0x12, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq - 0x21) && m_rgColor[sq - 0x21] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq - 0x21, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq - 0x0E) && m_rgColor[sq - 0x0E] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq - 0x0E, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq - 0x1F) && m_rgColor[sq - 0x1F] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq - 0x1F, iIndexFrom);
            pmv ++;
            }
        break;
    case x_pieceBishop:
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, 0x11);
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, 0x0F);
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, -0x11);
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, -0x0F);
        break;
    case x_pieceQueen:
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, 0x11);
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, 0x0F);
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, -0x11);
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, -0x0F);
        // Fall through
    case x_pieceRook:
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, 0x01);
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, 0x10);
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, -0x01);
        pmv = PmvGenPseudoLegalQuietDirection (pmv, side, sq, iIndexFrom, -0x10);
        break;
    case x_pieceKing:
        if (FValidSquare (sq + 0x01) && m_rgColor[sq + 0x01] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq + 0x01, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq + 0x10) && m_rgColor[sq + 0x10] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq + 0x10, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq + 0x11) && m_rgColor[sq + 0x11] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq + 0x11, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq + 0x0F) && m_rgColor[sq + 0x0F] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq + 0x0F, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq - 0x01) && m_rgColor[sq - 0x01] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq - 0x01, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq - 0x10) && m_rgColor[sq - 0x10] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq - 0x10, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq - 0x11) && m_rgColor[sq - 0x11] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq - 0x11, iIndexFrom);
            pmv ++;
            }
        if (FValidSquare (sq - 0x0F) && m_rgColor[sq - 0x0F] == x_colorNeutral)
            {
            VAddQuietMove (pmv, sq, sq - 0x0F, iIndexFrom);
            pmv ++;
            }
        break;
    default:
        Assert (0);
        }
    return pmv;
    }

// Generate all pseudo legal moves

CMove* CBoard::PmvGenPseudoLegalMoves
    (
    CMove   *pmv,
    color   side
    )
    {
    int     cPieces = m_rgPieceCount[side];
    color   xside = OtherSide (side);
    int     iPawnStep = rgiPawnStep[side];

    for (int iIndexFrom = 0; iIndexFrom < cPieces; iIndexFrom++)
        {
        square  sq = m_rgsqPieceList[side][iIndexFrom];
        piece   pi = m_rgBoard[sq];

        switch (pi)
            {
        case x_piecePawn:
            if (x_colorNeutral == m_rgColor[sq + iPawnStep])
                {
                pmv = PmvAddPawnMove (pmv, side, sq, sq + iPawnStep, iIndexFrom, 0);
                if (Row (sq) == rgiPawnRow[side] && x_colorNeutral == m_rgColor[sq + 2*iPawnStep])
                    {
                    VAddMove (pmv, sq, sq + 2*iPawnStep, iIndexFrom, 0);
                    pmv ++;
                    }
                }
            if (FValidSquare (sq + iPawnStep + 1))
                {
                if (m_rgColor[sq + iPawnStep + 1] == xside)
                    pmv = PmvAddPawnMove (pmv, side, sq, sq + iPawnStep + 1, iIndexFrom, x_mvCapture);
                else if (sq + iPawnStep + 1 == m_sqEnPassant)
                    {
                    VAddMove (pmv, sq, sq + iPawnStep + 1, iIndexFrom, x_mvEnPassant);
                    pmv ++;
                    }
                }
            if (FValidSquare (sq + iPawnStep - 1))
                {
                if (m_rgColor[sq + iPawnStep - 1] == xside)
                    pmv = PmvAddPawnMove (pmv, side, sq, sq + iPawnStep - 1, iIndexFrom, x_mvCapture);
                else if (sq + iPawnStep - 1 == m_sqEnPassant)
                    {
                    VAddMove (pmv, sq, sq + iPawnStep - 1, iIndexFrom, x_mvEnPassant);
                    pmv ++;
                    }
                }
            break;
        case x_pieceKnight:
            if (FValidSquare (sq + 0x12) && m_rgColor[sq + 0x12] != side)
                {
                VAddMove (pmv, sq, sq + 0x12, iIndexFrom, UlMvFlags (sq + 0x12));
                pmv ++;
                }
            if (FValidSquare (sq + 0x21) && m_rgColor[sq + 0x21] != side)
                {
                VAddMove (pmv, sq, sq + 0x21, iIndexFrom, UlMvFlags (sq + 0x21));
                pmv ++;
                }
            if (FValidSquare (sq + 0x0E) && m_rgColor[sq + 0x0E] != side)
                {
                VAddMove (pmv, sq, sq + 0x0E, iIndexFrom, UlMvFlags (sq + 0x0E));
                pmv ++;
                }
            if (FValidSquare (sq + 0x1F) && m_rgColor[sq + 0x1F] != side)
                {
                VAddMove (pmv, sq, sq + 0x1F, iIndexFrom, UlMvFlags (sq + 0x1F));
                pmv ++;
                }
            if (FValidSquare (sq - 0x12) && m_rgColor[sq - 0x12] != side)
                {
                VAddMove (pmv, sq, sq - 0x12, iIndexFrom, UlMvFlags (sq - 0x12));
                pmv ++;
                }
            if (FValidSquare (sq - 0x21) && m_rgColor[sq - 0x21] != side)
                {
                VAddMove (pmv, sq, sq - 0x21, iIndexFrom, UlMvFlags (sq - 0x21));
                pmv ++;
                }
            if (FValidSquare (sq - 0x0E) && m_rgColor[sq - 0x0E] != side)
                {
                VAddMove (pmv, sq, sq - 0x0E, iIndexFrom, UlMvFlags (sq - 0x0E));
                pmv ++;
                }
            if (FValidSquare (sq - 0x1F) && m_rgColor[sq - 0x1F] != side)
                {
                VAddMove (pmv, sq, sq - 0x1F, iIndexFrom, UlMvFlags (sq - 0x1F));
                pmv ++;
                }
            break;
        case x_pieceBishop:
            pmv = PmvGenPseudoLegalDirection (pmv, side, sq, iIndexFrom, 0x11);
            pmv = PmvGenPseudoLegalDirection (pmv, side, sq, iIndexFrom, 0x0F);
            pmv = PmvGenPseudoLegalDirection (pmv, side, sq, iIndexFrom, -0x11);
            pmv = PmvGenPseudoLegalDirection (pmv, side, sq, iIndexFrom, -0x0F);
            break;
        case x_pieceQueen:
            pmv = PmvGenPseudoLegalDirection (pmv, side, sq, iIndexFrom, 0x11);
            pmv = PmvGenPseudoLegalDirection (pmv, side, sq, iIndexFrom, 0x0F);
            pmv = PmvGenPseudoLegalDirection (pmv, side, sq, iIndexFrom, -0x11);
            pmv = PmvGenPseudoLegalDirection (pmv, side, sq, iIndexFrom, -0x0F);
            // Fall through
        case x_pieceRook:
            pmv = PmvGenPseudoLegalDirection (pmv, side, sq, iIndexFrom, 0x01);
            pmv = PmvGenPseudoLegalDirection (pmv, side, sq, iIndexFrom, 0x10);
            pmv = PmvGenPseudoLegalDirection (pmv, side, sq, iIndexFrom, -0x01);
            pmv = PmvGenPseudoLegalDirection (pmv, side, sq, iIndexFrom, -0x10);
            break;
        case x_pieceKing:
            if (FValidSquare (sq + 0x01) && m_rgColor[sq + 0x01] != side)
                {
                VAddMove (pmv, sq, sq + 0x01, iIndexFrom, UlMvFlags (sq + 0x01));
                pmv ++;
                }
            if (FValidSquare (sq + 0x10) && m_rgColor[sq + 0x10] != side)
                {
                VAddMove (pmv, sq, sq + 0x10, iIndexFrom, UlMvFlags (sq + 0x10));
                pmv ++;
                }
            if (FValidSquare (sq + 0x11) && m_rgColor[sq + 0x11] != side)
                {
                VAddMove (pmv, sq, sq + 0x11, iIndexFrom, UlMvFlags (sq + 0x11));
                pmv ++;
                }
            if (FValidSquare (sq + 0x0F) && m_rgColor[sq + 0x0F] != side)
                {
                VAddMove (pmv, sq, sq + 0x0F, iIndexFrom, UlMvFlags (sq + 0x0F));
                pmv ++;
                }
            if (FValidSquare (sq - 0x01) && m_rgColor[sq - 0x01] != side)
                {
                VAddMove (pmv, sq, sq - 0x01, iIndexFrom, UlMvFlags (sq - 0x01));
                pmv ++;
                }
            if (FValidSquare (sq - 0x10) && m_rgColor[sq - 0x10] != side)
                {
                VAddMove (pmv, sq, sq - 0x10, iIndexFrom, UlMvFlags (sq - 0x10));
                pmv ++;
                }
            if (FValidSquare (sq - 0x11) && m_rgColor[sq - 0x11] != side)
                {
                VAddMove (pmv, sq, sq - 0x11, iIndexFrom, UlMvFlags (sq - 0x11));
                pmv ++;
                }
            if (FValidSquare (sq - 0x0F) && m_rgColor[sq - 0x0F] != side)
                {
                VAddMove (pmv, sq, sq - 0x0F, iIndexFrom, UlMvFlags (sq - 0x0F));
                pmv ++;
                }
            break;
        default:
            Assert (0);
            }
        }
    return pmv;
    }

// Put piece to the board

INLINE void CBoard::VPutPiece
    (
    color   side,
    square  sq,
    piece   pi,
    int     index
    )
    {
    m_rgColor[sq] = side;
    m_rgBoard[sq] = pi;
    m_rgsqPieceList[side][index] = sq;
    }
    
// Check guard values

static void VCheckGuards ()
    {
    if (0x12345678 != *pulGuardBuf)
        {
        printf ("*** Guard overwritten\n");
        fflush (stdout);
        exit (1);
        }
    }

// Iterate through all the positions

template <int iWhite, int iBlack>
class CIterate
    {
public:
    static INLINE void VIterateThroughPositions
        (
        CBoard  *pbd,
        color   side,
        PfnBool pfn,
        color   colPrev,
        piece   piPrev,
        square  sqPrev
        )
        {
        color   col;
        int     ind;
        square  sq;
        piece   pi;
        const square *psqNext;

        col = (iWhite > 0) ? x_colorWhite : x_colorBlack;
        ind = (iWhite > 0) ? iWhite : iBlack;
        psqNext = pbd->m_rgpsqNext[col][ind];
        pi = pbd->m_rgPieces[col][ind];
        if (pi == piPrev && col == colPrev)
            {
            sq = psqNext[sqPrev];
            if (XX == sq)
                return;
            }
        else
            sq = rgsqFirstSquare[pi];
        do
            {
            if (x_pieceNone == pbd->m_rgBoard[sq])
                {
                pbd->VPutPiece (col, sq, pi, ind);
                CIterate<(iWhite>0)?iWhite-1:iWhite, (iWhite>0)?iBlack:iBlack-1>::
                    VIterateThroughPositions (pbd, side, pfn, col, pi, sq);
                pbd->m_rgColor[sq] = x_colorNeutral;
                pbd->m_rgBoard[sq] = x_pieceNone;
                }
            sq = psqNext[sq];
            }
        while (XX != sq);
        }
    };

template<> class CIterate<0,0>
    {
public:
    static INLINE void VIterateThroughPositions
        (
        CBoard  *pbd,
        color   side,
        PfnBool pfn,
        color   colPrev,
        piece   piPrev,
        square  sqPrev
        )
        {
        color xside;

        xside = OtherSide (side);
        if ((*pfn) (pbd, side, xside))  // Legal position?
            {
            // UNDONE: speedup that check (bitboard?)
            for (int i = pbd->m_rgPieceCount[xside]-1; i > 0; i --)
                {
                square sq = pbd->m_rgsqPieceList[xside][i];
                if ((rgiRowEnPassant[xside] == Row(sq)) &&
                    (x_piecePawn == pbd->m_rgBoard[sq]) &&
                    (x_pieceNone == pbd->m_rgBoard[sq-rgiPawnStep[xside]]) &&
                    (x_pieceNone == pbd->m_rgBoard[sq-rgiPawnStep[xside]*2]) &&
                    ((0 != Column(sq) && x_piecePawn == pbd->m_rgBoard[sq-1] && side == pbd->m_rgColor[sq-1]) ||
                     (7 != Column(sq) && x_piecePawn == pbd->m_rgBoard[sq+1] && side == pbd->m_rgColor[sq+1])))
                    {
                    pbd->m_sqEnPassant = sq-rgiPawnStep[xside];
                    (*pfn) (pbd, side, xside);
                    pbd->m_sqEnPassant = XX;
                    }
                }
            }
        }
    };

// Iterator table

typedef void (* PfnIterate) (CBoard *pbd, color side, PfnBool pfn, color colPrev, piece piPrev, square sqPrev);

#define IT(cW,cB)   CIterate<cW-1,cB-1>::VIterateThroughPositions

static const PfnIterate rgpfnIterate[3][4] =
    {
        { IT(2,1), IT(3,1), IT(4,1), NULL },
        { IT(2,2), IT(3,2), IT(4,2), NULL },
        { NULL,    IT(3,3), NULL,    NULL },
    };

INLINE void VIterateThroughPositions
    (
    square  sqwk,
    square  sqbk,
    CBoard  *pbd,
    color   side,
    PfnBool pfn
    )
    {
    pbd->VPutPiece (x_colorWhite, sqwk, x_pieceKing, 0);
    pbd->VPutPiece (x_colorBlack, sqbk, x_pieceKing, 0);
    pbd->m_sqEnPassant = XX;
    rgpfnIterate[pbd->m_rgPieceCount[x_colorBlack]-1][pbd->m_rgPieceCount[x_colorWhite]-2] (pbd, side, pfn, x_colorBlack, x_pieceKing, sqbk);
    pbd->m_rgColor[sqwk] = x_colorNeutral;
    pbd->m_rgBoard[sqwk] = x_pieceNone;
    pbd->m_rgColor[sqbk] = x_colorNeutral;
    pbd->m_rgBoard[sqbk] = x_pieceNone;
    }

// Iterate through all the positions - SMP code

static  bool    fVerify = false;
static  color   colVerify;

typedef square rgsqKings[2];

static  rgsqKings   *prgsqKings;
static  volatile int iKings;
static  volatile int cKings;

static void VPrintDone
    (
    int i,
    int cKings
    )
    {
    if (fVerify)
        {
        Lock (lockIO);
        printf ("%d%% done\r", (int) (100.0 * ((double) i) / (double) cKings));
        fflush (stdout);
        UnLock (lockIO);
        }
    }

#if defined (SMP)

lock_t  lockKings;

static void VDoWork (CBoard *pbd)
    {
    for (;;)
        {
        int i;

        Lock (lockKings);
        if (iKings == cKings)
            {
            UnLock (lockKings);
            break;
            }
        i = iKings;
        iKings ++;
        UnLock (lockKings);

        VPrintDone (i, cKings);
        if (x_colorWhite == pbd->m_side)
            VIterateThroughPositions (prgsqKings[i][x_colorWhite], prgsqKings[i][x_colorBlack],
                                      pbd, pbd->m_side, pbd->m_pfn);
        else
            VIterateThroughPositions (prgsqKings[i][x_colorBlack], prgsqKings[i][x_colorWhite],
                                      pbd, pbd->m_side, pbd->m_pfn);
        }
    }

static void* TB_CDECL VWorker (void *pv)
    {
    CBoard  *pbd = (CBoard*) pv;

    for (;;)
        {
        while (!pbd->m_fHasWork)
            usleep(100);
        VDoWork (pbd);
        pbd->m_fHasWork = false;
        Lock   (lockKings); // Memory barrier
        UnLock (lockKings);
        }
    }

#endif

// Iterate through all the positions - upper-level function

static square rgsqKings462[462][2] =
{
    { 0, 2 },
    { 0, 3 },
    { 0, 4 },
    { 0, 5 },
    { 0, 6 },
    { 0, 7 },
    { 0, 18 },
    { 0, 19 },
    { 0, 20 },
    { 0, 21 },
    { 0, 22 },
    { 0, 23 },
    { 0, 34 },
    { 0, 35 },
    { 0, 36 },
    { 0, 37 },
    { 0, 38 },
    { 0, 39 },
    { 0, 51 },
    { 0, 52 },
    { 0, 53 },
    { 0, 54 },
    { 0, 55 },
    { 0, 68 },
    { 0, 69 },
    { 0, 70 },
    { 0, 71 },
    { 0, 85 },
    { 0, 86 },
    { 0, 87 },
    { 0, 102 },
    { 0, 103 },
    { 0, 119 },
    { 1, 3 },
    { 1, 4 },
    { 1, 5 },
    { 1, 6 },
    { 1, 7 },
    { 1, 19 },
    { 1, 20 },
    { 1, 21 },
    { 1, 22 },
    { 1, 23 },
    { 1, 32 },
    { 1, 33 },
    { 1, 34 },
    { 1, 35 },
    { 1, 36 },
    { 1, 37 },
    { 1, 38 },
    { 1, 39 },
    { 1, 48 },
    { 1, 49 },
    { 1, 50 },
    { 1, 51 },
    { 1, 52 },
    { 1, 53 },
    { 1, 54 },
    { 1, 55 },
    { 1, 64 },
    { 1, 65 },
    { 1, 66 },
    { 1, 67 },
    { 1, 68 },
    { 1, 69 },
    { 1, 70 },
    { 1, 71 },
    { 1, 80 },
    { 1, 81 },
    { 1, 82 },
    { 1, 83 },
    { 1, 84 },
    { 1, 85 },
    { 1, 86 },
    { 1, 87 },
    { 1, 96 },
    { 1, 97 },
    { 1, 98 },
    { 1, 99 },
    { 1, 100 },
    { 1, 101 },
    { 1, 102 },
    { 1, 103 },
    { 1, 112 },
    { 1, 113 },
    { 1, 114 },
    { 1, 115 },
    { 1, 116 },
    { 1, 117 },
    { 1, 118 },
    { 1, 119 },
    { 2, 0 },
    { 2, 4 },
    { 2, 5 },
    { 2, 6 },
    { 2, 7 },
    { 2, 16 },
    { 2, 20 },
    { 2, 21 },
    { 2, 22 },
    { 2, 23 },
    { 2, 32 },
    { 2, 33 },
    { 2, 34 },
    { 2, 35 },
    { 2, 36 },
    { 2, 37 },
    { 2, 38 },
    { 2, 39 },
    { 2, 48 },
    { 2, 49 },
    { 2, 50 },
    { 2, 51 },
    { 2, 52 },
    { 2, 53 },
    { 2, 54 },
    { 2, 55 },
    { 2, 64 },
    { 2, 65 },
    { 2, 66 },
    { 2, 67 },
    { 2, 68 },
    { 2, 69 },
    { 2, 70 },
    { 2, 71 },
    { 2, 80 },
    { 2, 81 },
    { 2, 82 },
    { 2, 83 },
    { 2, 84 },
    { 2, 85 },
    { 2, 86 },
    { 2, 87 },
    { 2, 96 },
    { 2, 97 },
    { 2, 98 },
    { 2, 99 },
    { 2, 100 },
    { 2, 101 },
    { 2, 102 },
    { 2, 103 },
    { 2, 112 },
    { 2, 113 },
    { 2, 114 },
    { 2, 115 },
    { 2, 116 },
    { 2, 117 },
    { 2, 118 },
    { 2, 119 },
    { 3, 0 },
    { 3, 1 },
    { 3, 5 },
    { 3, 6 },
    { 3, 7 },
    { 3, 16 },
    { 3, 17 },
    { 3, 21 },
    { 3, 22 },
    { 3, 23 },
    { 3, 32 },
    { 3, 33 },
    { 3, 34 },
    { 3, 35 },
    { 3, 36 },
    { 3, 37 },
    { 3, 38 },
    { 3, 39 },
    { 3, 48 },
    { 3, 49 },
    { 3, 50 },
    { 3, 51 },
    { 3, 52 },
    { 3, 53 },
    { 3, 54 },
    { 3, 55 },
    { 3, 64 },
    { 3, 65 },
    { 3, 66 },
    { 3, 67 },
    { 3, 68 },
    { 3, 69 },
    { 3, 70 },
    { 3, 71 },
    { 3, 80 },
    { 3, 81 },
    { 3, 82 },
    { 3, 83 },
    { 3, 84 },
    { 3, 85 },
    { 3, 86 },
    { 3, 87 },
    { 3, 96 },
    { 3, 97 },
    { 3, 98 },
    { 3, 99 },
    { 3, 100 },
    { 3, 101 },
    { 3, 102 },
    { 3, 103 },
    { 3, 112 },
    { 3, 113 },
    { 3, 114 },
    { 3, 115 },
    { 3, 116 },
    { 3, 117 },
    { 3, 118 },
    { 3, 119 },
    { 17, 3 },
    { 17, 4 },
    { 17, 5 },
    { 17, 6 },
    { 17, 7 },
    { 17, 19 },
    { 17, 20 },
    { 17, 21 },
    { 17, 22 },
    { 17, 23 },
    { 17, 35 },
    { 17, 36 },
    { 17, 37 },
    { 17, 38 },
    { 17, 39 },
    { 17, 51 },
    { 17, 52 },
    { 17, 53 },
    { 17, 54 },
    { 17, 55 },
    { 17, 68 },
    { 17, 69 },
    { 17, 70 },
    { 17, 71 },
    { 17, 85 },
    { 17, 86 },
    { 17, 87 },
    { 17, 102 },
    { 17, 103 },
    { 17, 119 },
    { 18, 0 },
    { 18, 4 },
    { 18, 5 },
    { 18, 6 },
    { 18, 7 },
    { 18, 16 },
    { 18, 20 },
    { 18, 21 },
    { 18, 22 },
    { 18, 23 },
    { 18, 32 },
    { 18, 36 },
    { 18, 37 },
    { 18, 38 },
    { 18, 39 },
    { 18, 48 },
    { 18, 49 },
    { 18, 50 },
    { 18, 51 },
    { 18, 52 },
    { 18, 53 },
    { 18, 54 },
    { 18, 55 },
    { 18, 64 },
    { 18, 65 },
    { 18, 66 },
    { 18, 67 },
    { 18, 68 },
    { 18, 69 },
    { 18, 70 },
    { 18, 71 },
    { 18, 80 },
    { 18, 81 },
    { 18, 82 },
    { 18, 83 },
    { 18, 84 },
    { 18, 85 },
    { 18, 86 },
    { 18, 87 },
    { 18, 96 },
    { 18, 97 },
    { 18, 98 },
    { 18, 99 },
    { 18, 100 },
    { 18, 101 },
    { 18, 102 },
    { 18, 103 },
    { 18, 112 },
    { 18, 113 },
    { 18, 114 },
    { 18, 115 },
    { 18, 116 },
    { 18, 117 },
    { 18, 118 },
    { 18, 119 },
    { 19, 0 },
    { 19, 1 },
    { 19, 5 },
    { 19, 6 },
    { 19, 7 },
    { 19, 16 },
    { 19, 17 },
    { 19, 21 },
    { 19, 22 },
    { 19, 23 },
    { 19, 32 },
    { 19, 33 },
    { 19, 37 },
    { 19, 38 },
    { 19, 39 },
    { 19, 48 },
    { 19, 49 },
    { 19, 50 },
    { 19, 51 },
    { 19, 52 },
    { 19, 53 },
    { 19, 54 },
    { 19, 55 },
    { 19, 64 },
    { 19, 65 },
    { 19, 66 },
    { 19, 67 },
    { 19, 68 },
    { 19, 69 },
    { 19, 70 },
    { 19, 71 },
    { 19, 80 },
    { 19, 81 },
    { 19, 82 },
    { 19, 83 },
    { 19, 84 },
    { 19, 85 },
    { 19, 86 },
    { 19, 87 },
    { 19, 96 },
    { 19, 97 },
    { 19, 98 },
    { 19, 99 },
    { 19, 100 },
    { 19, 101 },
    { 19, 102 },
    { 19, 103 },
    { 19, 112 },
    { 19, 113 },
    { 19, 114 },
    { 19, 115 },
    { 19, 116 },
    { 19, 117 },
    { 19, 118 },
    { 19, 119 },
    { 34, 0 },
    { 34, 1 },
    { 34, 2 },
    { 34, 3 },
    { 34, 4 },
    { 34, 5 },
    { 34, 6 },
    { 34, 7 },
    { 34, 20 },
    { 34, 21 },
    { 34, 22 },
    { 34, 23 },
    { 34, 36 },
    { 34, 37 },
    { 34, 38 },
    { 34, 39 },
    { 34, 52 },
    { 34, 53 },
    { 34, 54 },
    { 34, 55 },
    { 34, 68 },
    { 34, 69 },
    { 34, 70 },
    { 34, 71 },
    { 34, 85 },
    { 34, 86 },
    { 34, 87 },
    { 34, 102 },
    { 34, 103 },
    { 34, 119 },
    { 35, 0 },
    { 35, 1 },
    { 35, 2 },
    { 35, 3 },
    { 35, 4 },
    { 35, 5 },
    { 35, 6 },
    { 35, 7 },
    { 35, 16 },
    { 35, 17 },
    { 35, 21 },
    { 35, 22 },
    { 35, 23 },
    { 35, 32 },
    { 35, 33 },
    { 35, 37 },
    { 35, 38 },
    { 35, 39 },
    { 35, 48 },
    { 35, 49 },
    { 35, 53 },
    { 35, 54 },
    { 35, 55 },
    { 35, 64 },
    { 35, 65 },
    { 35, 66 },
    { 35, 67 },
    { 35, 68 },
    { 35, 69 },
    { 35, 70 },
    { 35, 71 },
    { 35, 80 },
    { 35, 81 },
    { 35, 82 },
    { 35, 83 },
    { 35, 84 },
    { 35, 85 },
    { 35, 86 },
    { 35, 87 },
    { 35, 96 },
    { 35, 97 },
    { 35, 98 },
    { 35, 99 },
    { 35, 100 },
    { 35, 101 },
    { 35, 102 },
    { 35, 103 },
    { 35, 112 },
    { 35, 113 },
    { 35, 114 },
    { 35, 115 },
    { 35, 116 },
    { 35, 117 },
    { 35, 118 },
    { 35, 119 },
    { 51, 0 },
    { 51, 1 },
    { 51, 2 },
    { 51, 3 },
    { 51, 4 },
    { 51, 5 },
    { 51, 6 },
    { 51, 7 },
    { 51, 17 },
    { 51, 18 },
    { 51, 19 },
    { 51, 20 },
    { 51, 21 },
    { 51, 22 },
    { 51, 23 },
    { 51, 37 },
    { 51, 38 },
    { 51, 39 },
    { 51, 53 },
    { 51, 54 },
    { 51, 55 },
    { 51, 69 },
    { 51, 70 },
    { 51, 71 },
    { 51, 85 },
    { 51, 86 },
    { 51, 87 },
    { 51, 102 },
    { 51, 103 },
    { 51, 119 },
};

static square rgsqKings564[564][2] =
{
    { 0, 2 },
    { 0, 3 },
    { 0, 4 },
    { 0, 5 },
    { 0, 6 },
    { 0, 7 },
    { 0, 18 },
    { 0, 19 },
    { 0, 20 },
    { 0, 21 },
    { 0, 22 },
    { 0, 23 },
    { 0, 32 },
    { 0, 33 },
    { 0, 34 },
    { 0, 35 },
    { 0, 36 },
    { 0, 37 },
    { 0, 38 },
    { 0, 39 },
    { 0, 48 },
    { 0, 49 },
    { 0, 50 },
    { 0, 51 },
    { 0, 52 },
    { 0, 53 },
    { 0, 54 },
    { 0, 55 },
    { 0, 64 },
    { 0, 65 },
    { 0, 66 },
    { 0, 67 },
    { 0, 68 },
    { 0, 69 },
    { 0, 70 },
    { 0, 71 },
    { 0, 80 },
    { 0, 81 },
    { 0, 82 },
    { 0, 83 },
    { 0, 84 },
    { 0, 85 },
    { 0, 86 },
    { 0, 87 },
    { 0, 96 },
    { 0, 97 },
    { 0, 98 },
    { 0, 99 },
    { 0, 100 },
    { 0, 101 },
    { 0, 102 },
    { 0, 103 },
    { 0, 112 },
    { 0, 113 },
    { 0, 114 },
    { 0, 115 },
    { 0, 116 },
    { 0, 117 },
    { 0, 118 },
    { 0, 119 },
    { 1, 3 },
    { 1, 4 },
    { 1, 5 },
    { 1, 6 },
    { 1, 7 },
    { 1, 19 },
    { 1, 20 },
    { 1, 21 },
    { 1, 22 },
    { 1, 23 },
    { 1, 32 },
    { 1, 33 },
    { 1, 34 },
    { 1, 35 },
    { 1, 36 },
    { 1, 37 },
    { 1, 38 },
    { 1, 39 },
    { 1, 48 },
    { 1, 49 },
    { 1, 50 },
    { 1, 51 },
    { 1, 52 },
    { 1, 53 },
    { 1, 54 },
    { 1, 55 },
    { 1, 64 },
    { 1, 65 },
    { 1, 66 },
    { 1, 67 },
    { 1, 68 },
    { 1, 69 },
    { 1, 70 },
    { 1, 71 },
    { 1, 80 },
    { 1, 81 },
    { 1, 82 },
    { 1, 83 },
    { 1, 84 },
    { 1, 85 },
    { 1, 86 },
    { 1, 87 },
    { 1, 96 },
    { 1, 97 },
    { 1, 98 },
    { 1, 99 },
    { 1, 100 },
    { 1, 101 },
    { 1, 102 },
    { 1, 103 },
    { 1, 112 },
    { 1, 113 },
    { 1, 114 },
    { 1, 115 },
    { 1, 116 },
    { 1, 117 },
    { 1, 118 },
    { 1, 119 },
    { 2, 0 },
    { 2, 4 },
    { 2, 5 },
    { 2, 6 },
    { 2, 7 },
    { 2, 16 },
    { 2, 20 },
    { 2, 21 },
    { 2, 22 },
    { 2, 23 },
    { 2, 32 },
    { 2, 33 },
    { 2, 34 },
    { 2, 35 },
    { 2, 36 },
    { 2, 37 },
    { 2, 38 },
    { 2, 39 },
    { 2, 48 },
    { 2, 49 },
    { 2, 50 },
    { 2, 51 },
    { 2, 52 },
    { 2, 53 },
    { 2, 54 },
    { 2, 55 },
    { 2, 64 },
    { 2, 65 },
    { 2, 66 },
    { 2, 67 },
    { 2, 68 },
    { 2, 69 },
    { 2, 70 },
    { 2, 71 },
    { 2, 80 },
    { 2, 81 },
    { 2, 82 },
    { 2, 83 },
    { 2, 84 },
    { 2, 85 },
    { 2, 86 },
    { 2, 87 },
    { 2, 96 },
    { 2, 97 },
    { 2, 98 },
    { 2, 99 },
    { 2, 100 },
    { 2, 101 },
    { 2, 102 },
    { 2, 103 },
    { 2, 112 },
    { 2, 113 },
    { 2, 114 },
    { 2, 115 },
    { 2, 116 },
    { 2, 117 },
    { 2, 118 },
    { 2, 119 },
    { 3, 0 },
    { 3, 1 },
    { 3, 5 },
    { 3, 6 },
    { 3, 7 },
    { 3, 16 },
    { 3, 17 },
    { 3, 21 },
    { 3, 22 },
    { 3, 23 },
    { 3, 32 },
    { 3, 33 },
    { 3, 34 },
    { 3, 35 },
    { 3, 36 },
    { 3, 37 },
    { 3, 38 },
    { 3, 39 },
    { 3, 48 },
    { 3, 49 },
    { 3, 50 },
    { 3, 51 },
    { 3, 52 },
    { 3, 53 },
    { 3, 54 },
    { 3, 55 },
    { 3, 64 },
    { 3, 65 },
    { 3, 66 },
    { 3, 67 },
    { 3, 68 },
    { 3, 69 },
    { 3, 70 },
    { 3, 71 },
    { 3, 80 },
    { 3, 81 },
    { 3, 82 },
    { 3, 83 },
    { 3, 84 },
    { 3, 85 },
    { 3, 86 },
    { 3, 87 },
    { 3, 96 },
    { 3, 97 },
    { 3, 98 },
    { 3, 99 },
    { 3, 100 },
    { 3, 101 },
    { 3, 102 },
    { 3, 103 },
    { 3, 112 },
    { 3, 113 },
    { 3, 114 },
    { 3, 115 },
    { 3, 116 },
    { 3, 117 },
    { 3, 118 },
    { 3, 119 },
    { 17, 3 },
    { 17, 4 },
    { 17, 5 },
    { 17, 6 },
    { 17, 7 },
    { 17, 19 },
    { 17, 20 },
    { 17, 21 },
    { 17, 22 },
    { 17, 23 },
    { 17, 35 },
    { 17, 36 },
    { 17, 37 },
    { 17, 38 },
    { 17, 39 },
    { 17, 48 },
    { 17, 49 },
    { 17, 50 },
    { 17, 51 },
    { 17, 52 },
    { 17, 53 },
    { 17, 54 },
    { 17, 55 },
    { 17, 64 },
    { 17, 65 },
    { 17, 66 },
    { 17, 67 },
    { 17, 68 },
    { 17, 69 },
    { 17, 70 },
    { 17, 71 },
    { 17, 80 },
    { 17, 81 },
    { 17, 82 },
    { 17, 83 },
    { 17, 84 },
    { 17, 85 },
    { 17, 86 },
    { 17, 87 },
    { 17, 96 },
    { 17, 97 },
    { 17, 98 },
    { 17, 99 },
    { 17, 100 },
    { 17, 101 },
    { 17, 102 },
    { 17, 103 },
    { 17, 112 },
    { 17, 113 },
    { 17, 114 },
    { 17, 115 },
    { 17, 116 },
    { 17, 117 },
    { 17, 118 },
    { 17, 119 },
    { 18, 0 },
    { 18, 4 },
    { 18, 5 },
    { 18, 6 },
    { 18, 7 },
    { 18, 16 },
    { 18, 20 },
    { 18, 21 },
    { 18, 22 },
    { 18, 23 },
    { 18, 32 },
    { 18, 36 },
    { 18, 37 },
    { 18, 38 },
    { 18, 39 },
    { 18, 48 },
    { 18, 49 },
    { 18, 50 },
    { 18, 51 },
    { 18, 52 },
    { 18, 53 },
    { 18, 54 },
    { 18, 55 },
    { 18, 64 },
    { 18, 65 },
    { 18, 66 },
    { 18, 67 },
    { 18, 68 },
    { 18, 69 },
    { 18, 70 },
    { 18, 71 },
    { 18, 80 },
    { 18, 81 },
    { 18, 82 },
    { 18, 83 },
    { 18, 84 },
    { 18, 85 },
    { 18, 86 },
    { 18, 87 },
    { 18, 96 },
    { 18, 97 },
    { 18, 98 },
    { 18, 99 },
    { 18, 100 },
    { 18, 101 },
    { 18, 102 },
    { 18, 103 },
    { 18, 112 },
    { 18, 113 },
    { 18, 114 },
    { 18, 115 },
    { 18, 116 },
    { 18, 117 },
    { 18, 118 },
    { 18, 119 },
    { 19, 0 },
    { 19, 1 },
    { 19, 5 },
    { 19, 6 },
    { 19, 7 },
    { 19, 16 },
    { 19, 17 },
    { 19, 21 },
    { 19, 22 },
    { 19, 23 },
    { 19, 32 },
    { 19, 33 },
    { 19, 37 },
    { 19, 38 },
    { 19, 39 },
    { 19, 48 },
    { 19, 49 },
    { 19, 50 },
    { 19, 51 },
    { 19, 52 },
    { 19, 53 },
    { 19, 54 },
    { 19, 55 },
    { 19, 64 },
    { 19, 65 },
    { 19, 66 },
    { 19, 67 },
    { 19, 68 },
    { 19, 69 },
    { 19, 70 },
    { 19, 71 },
    { 19, 80 },
    { 19, 81 },
    { 19, 82 },
    { 19, 83 },
    { 19, 84 },
    { 19, 85 },
    { 19, 86 },
    { 19, 87 },
    { 19, 96 },
    { 19, 97 },
    { 19, 98 },
    { 19, 99 },
    { 19, 100 },
    { 19, 101 },
    { 19, 102 },
    { 19, 103 },
    { 19, 112 },
    { 19, 113 },
    { 19, 114 },
    { 19, 115 },
    { 19, 116 },
    { 19, 117 },
    { 19, 118 },
    { 19, 119 },
    { 34, 0 },
    { 34, 1 },
    { 34, 2 },
    { 34, 3 },
    { 34, 4 },
    { 34, 5 },
    { 34, 6 },
    { 34, 7 },
    { 34, 16 },
    { 34, 20 },
    { 34, 21 },
    { 34, 22 },
    { 34, 23 },
    { 34, 32 },
    { 34, 36 },
    { 34, 37 },
    { 34, 38 },
    { 34, 39 },
    { 34, 48 },
    { 34, 52 },
    { 34, 53 },
    { 34, 54 },
    { 34, 55 },
    { 34, 64 },
    { 34, 65 },
    { 34, 66 },
    { 34, 67 },
    { 34, 68 },
    { 34, 69 },
    { 34, 70 },
    { 34, 71 },
    { 34, 80 },
    { 34, 81 },
    { 34, 82 },
    { 34, 83 },
    { 34, 84 },
    { 34, 85 },
    { 34, 86 },
    { 34, 87 },
    { 34, 96 },
    { 34, 97 },
    { 34, 98 },
    { 34, 99 },
    { 34, 100 },
    { 34, 101 },
    { 34, 102 },
    { 34, 103 },
    { 34, 112 },
    { 34, 113 },
    { 34, 114 },
    { 34, 115 },
    { 34, 116 },
    { 34, 117 },
    { 34, 118 },
    { 34, 119 },
    { 35, 0 },
    { 35, 1 },
    { 35, 2 },
    { 35, 3 },
    { 35, 4 },
    { 35, 5 },
    { 35, 6 },
    { 35, 7 },
    { 35, 16 },
    { 35, 17 },
    { 35, 21 },
    { 35, 22 },
    { 35, 23 },
    { 35, 32 },
    { 35, 33 },
    { 35, 37 },
    { 35, 38 },
    { 35, 39 },
    { 35, 48 },
    { 35, 49 },
    { 35, 53 },
    { 35, 54 },
    { 35, 55 },
    { 35, 64 },
    { 35, 65 },
    { 35, 66 },
    { 35, 67 },
    { 35, 68 },
    { 35, 69 },
    { 35, 70 },
    { 35, 71 },
    { 35, 80 },
    { 35, 81 },
    { 35, 82 },
    { 35, 83 },
    { 35, 84 },
    { 35, 85 },
    { 35, 86 },
    { 35, 87 },
    { 35, 96 },
    { 35, 97 },
    { 35, 98 },
    { 35, 99 },
    { 35, 100 },
    { 35, 101 },
    { 35, 102 },
    { 35, 103 },
    { 35, 112 },
    { 35, 113 },
    { 35, 114 },
    { 35, 115 },
    { 35, 116 },
    { 35, 117 },
    { 35, 118 },
    { 35, 119 },
    { 51, 0 },
    { 51, 1 },
    { 51, 2 },
    { 51, 3 },
    { 51, 4 },
    { 51, 5 },
    { 51, 6 },
    { 51, 7 },
    { 51, 16 },
    { 51, 17 },
    { 51, 18 },
    { 51, 19 },
    { 51, 20 },
    { 51, 21 },
    { 51, 22 },
    { 51, 23 },
    { 51, 32 },
    { 51, 33 },
    { 51, 37 },
    { 51, 38 },
    { 51, 39 },
    { 51, 48 },
    { 51, 49 },
    { 51, 53 },
    { 51, 54 },
    { 51, 55 },
    { 51, 64 },
    { 51, 65 },
    { 51, 69 },
    { 51, 70 },
    { 51, 71 },
    { 51, 80 },
    { 51, 81 },
    { 51, 82 },
    { 51, 83 },
    { 51, 84 },
    { 51, 85 },
    { 51, 86 },
    { 51, 87 },
    { 51, 96 },
    { 51, 97 },
    { 51, 98 },
    { 51, 99 },
    { 51, 100 },
    { 51, 101 },
    { 51, 102 },
    { 51, 103 },
    { 51, 112 },
    { 51, 113 },
    { 51, 114 },
    { 51, 115 },
    { 51, 116 },
    { 51, 117 },
    { 51, 118 },
    { 51, 119 },
};

static square rgsqKings1806[1806][2] =
{
    { 0, 2 },
    { 0, 3 },
    { 0, 4 },
    { 0, 5 },
    { 0, 6 },
    { 0, 7 },
    { 0, 18 },
    { 0, 19 },
    { 0, 20 },
    { 0, 21 },
    { 0, 22 },
    { 0, 23 },
    { 0, 32 },
    { 0, 33 },
    { 0, 34 },
    { 0, 35 },
    { 0, 36 },
    { 0, 37 },
    { 0, 38 },
    { 0, 39 },
    { 0, 48 },
    { 0, 49 },
    { 0, 50 },
    { 0, 51 },
    { 0, 52 },
    { 0, 53 },
    { 0, 54 },
    { 0, 55 },
    { 0, 64 },
    { 0, 65 },
    { 0, 66 },
    { 0, 67 },
    { 0, 68 },
    { 0, 69 },
    { 0, 70 },
    { 0, 71 },
    { 0, 80 },
    { 0, 81 },
    { 0, 82 },
    { 0, 83 },
    { 0, 84 },
    { 0, 85 },
    { 0, 86 },
    { 0, 87 },
    { 0, 96 },
    { 0, 97 },
    { 0, 98 },
    { 0, 99 },
    { 0, 100 },
    { 0, 101 },
    { 0, 102 },
    { 0, 103 },
    { 0, 112 },
    { 0, 113 },
    { 0, 114 },
    { 0, 115 },
    { 0, 116 },
    { 0, 117 },
    { 0, 118 },
    { 0, 119 },
    { 1, 3 },
    { 1, 4 },
    { 1, 5 },
    { 1, 6 },
    { 1, 7 },
    { 1, 19 },
    { 1, 20 },
    { 1, 21 },
    { 1, 22 },
    { 1, 23 },
    { 1, 32 },
    { 1, 33 },
    { 1, 34 },
    { 1, 35 },
    { 1, 36 },
    { 1, 37 },
    { 1, 38 },
    { 1, 39 },
    { 1, 48 },
    { 1, 49 },
    { 1, 50 },
    { 1, 51 },
    { 1, 52 },
    { 1, 53 },
    { 1, 54 },
    { 1, 55 },
    { 1, 64 },
    { 1, 65 },
    { 1, 66 },
    { 1, 67 },
    { 1, 68 },
    { 1, 69 },
    { 1, 70 },
    { 1, 71 },
    { 1, 80 },
    { 1, 81 },
    { 1, 82 },
    { 1, 83 },
    { 1, 84 },
    { 1, 85 },
    { 1, 86 },
    { 1, 87 },
    { 1, 96 },
    { 1, 97 },
    { 1, 98 },
    { 1, 99 },
    { 1, 100 },
    { 1, 101 },
    { 1, 102 },
    { 1, 103 },
    { 1, 112 },
    { 1, 113 },
    { 1, 114 },
    { 1, 115 },
    { 1, 116 },
    { 1, 117 },
    { 1, 118 },
    { 1, 119 },
    { 2, 0 },
    { 2, 4 },
    { 2, 5 },
    { 2, 6 },
    { 2, 7 },
    { 2, 16 },
    { 2, 20 },
    { 2, 21 },
    { 2, 22 },
    { 2, 23 },
    { 2, 32 },
    { 2, 33 },
    { 2, 34 },
    { 2, 35 },
    { 2, 36 },
    { 2, 37 },
    { 2, 38 },
    { 2, 39 },
    { 2, 48 },
    { 2, 49 },
    { 2, 50 },
    { 2, 51 },
    { 2, 52 },
    { 2, 53 },
    { 2, 54 },
    { 2, 55 },
    { 2, 64 },
    { 2, 65 },
    { 2, 66 },
    { 2, 67 },
    { 2, 68 },
    { 2, 69 },
    { 2, 70 },
    { 2, 71 },
    { 2, 80 },
    { 2, 81 },
    { 2, 82 },
    { 2, 83 },
    { 2, 84 },
    { 2, 85 },
    { 2, 86 },
    { 2, 87 },
    { 2, 96 },
    { 2, 97 },
    { 2, 98 },
    { 2, 99 },
    { 2, 100 },
    { 2, 101 },
    { 2, 102 },
    { 2, 103 },
    { 2, 112 },
    { 2, 113 },
    { 2, 114 },
    { 2, 115 },
    { 2, 116 },
    { 2, 117 },
    { 2, 118 },
    { 2, 119 },
    { 3, 0 },
    { 3, 1 },
    { 3, 5 },
    { 3, 6 },
    { 3, 7 },
    { 3, 16 },
    { 3, 17 },
    { 3, 21 },
    { 3, 22 },
    { 3, 23 },
    { 3, 32 },
    { 3, 33 },
    { 3, 34 },
    { 3, 35 },
    { 3, 36 },
    { 3, 37 },
    { 3, 38 },
    { 3, 39 },
    { 3, 48 },
    { 3, 49 },
    { 3, 50 },
    { 3, 51 },
    { 3, 52 },
    { 3, 53 },
    { 3, 54 },
    { 3, 55 },
    { 3, 64 },
    { 3, 65 },
    { 3, 66 },
    { 3, 67 },
    { 3, 68 },
    { 3, 69 },
    { 3, 70 },
    { 3, 71 },
    { 3, 80 },
    { 3, 81 },
    { 3, 82 },
    { 3, 83 },
    { 3, 84 },
    { 3, 85 },
    { 3, 86 },
    { 3, 87 },
    { 3, 96 },
    { 3, 97 },
    { 3, 98 },
    { 3, 99 },
    { 3, 100 },
    { 3, 101 },
    { 3, 102 },
    { 3, 103 },
    { 3, 112 },
    { 3, 113 },
    { 3, 114 },
    { 3, 115 },
    { 3, 116 },
    { 3, 117 },
    { 3, 118 },
    { 3, 119 },
    { 16, 2 },
    { 16, 3 },
    { 16, 4 },
    { 16, 5 },
    { 16, 6 },
    { 16, 7 },
    { 16, 18 },
    { 16, 19 },
    { 16, 20 },
    { 16, 21 },
    { 16, 22 },
    { 16, 23 },
    { 16, 34 },
    { 16, 35 },
    { 16, 36 },
    { 16, 37 },
    { 16, 38 },
    { 16, 39 },
    { 16, 48 },
    { 16, 49 },
    { 16, 50 },
    { 16, 51 },
    { 16, 52 },
    { 16, 53 },
    { 16, 54 },
    { 16, 55 },
    { 16, 64 },
    { 16, 65 },
    { 16, 66 },
    { 16, 67 },
    { 16, 68 },
    { 16, 69 },
    { 16, 70 },
    { 16, 71 },
    { 16, 80 },
    { 16, 81 },
    { 16, 82 },
    { 16, 83 },
    { 16, 84 },
    { 16, 85 },
    { 16, 86 },
    { 16, 87 },
    { 16, 96 },
    { 16, 97 },
    { 16, 98 },
    { 16, 99 },
    { 16, 100 },
    { 16, 101 },
    { 16, 102 },
    { 16, 103 },
    { 16, 112 },
    { 16, 113 },
    { 16, 114 },
    { 16, 115 },
    { 16, 116 },
    { 16, 117 },
    { 16, 118 },
    { 16, 119 },
    { 17, 3 },
    { 17, 4 },
    { 17, 5 },
    { 17, 6 },
    { 17, 7 },
    { 17, 19 },
    { 17, 20 },
    { 17, 21 },
    { 17, 22 },
    { 17, 23 },
    { 17, 35 },
    { 17, 36 },
    { 17, 37 },
    { 17, 38 },
    { 17, 39 },
    { 17, 48 },
    { 17, 49 },
    { 17, 50 },
    { 17, 51 },
    { 17, 52 },
    { 17, 53 },
    { 17, 54 },
    { 17, 55 },
    { 17, 64 },
    { 17, 65 },
    { 17, 66 },
    { 17, 67 },
    { 17, 68 },
    { 17, 69 },
    { 17, 70 },
    { 17, 71 },
    { 17, 80 },
    { 17, 81 },
    { 17, 82 },
    { 17, 83 },
    { 17, 84 },
    { 17, 85 },
    { 17, 86 },
    { 17, 87 },
    { 17, 96 },
    { 17, 97 },
    { 17, 98 },
    { 17, 99 },
    { 17, 100 },
    { 17, 101 },
    { 17, 102 },
    { 17, 103 },
    { 17, 112 },
    { 17, 113 },
    { 17, 114 },
    { 17, 115 },
    { 17, 116 },
    { 17, 117 },
    { 17, 118 },
    { 17, 119 },
    { 18, 0 },
    { 18, 4 },
    { 18, 5 },
    { 18, 6 },
    { 18, 7 },
    { 18, 16 },
    { 18, 20 },
    { 18, 21 },
    { 18, 22 },
    { 18, 23 },
    { 18, 32 },
    { 18, 36 },
    { 18, 37 },
    { 18, 38 },
    { 18, 39 },
    { 18, 48 },
    { 18, 49 },
    { 18, 50 },
    { 18, 51 },
    { 18, 52 },
    { 18, 53 },
    { 18, 54 },
    { 18, 55 },
    { 18, 64 },
    { 18, 65 },
    { 18, 66 },
    { 18, 67 },
    { 18, 68 },
    { 18, 69 },
    { 18, 70 },
    { 18, 71 },
    { 18, 80 },
    { 18, 81 },
    { 18, 82 },
    { 18, 83 },
    { 18, 84 },
    { 18, 85 },
    { 18, 86 },
    { 18, 87 },
    { 18, 96 },
    { 18, 97 },
    { 18, 98 },
    { 18, 99 },
    { 18, 100 },
    { 18, 101 },
    { 18, 102 },
    { 18, 103 },
    { 18, 112 },
    { 18, 113 },
    { 18, 114 },
    { 18, 115 },
    { 18, 116 },
    { 18, 117 },
    { 18, 118 },
    { 18, 119 },
    { 19, 0 },
    { 19, 1 },
    { 19, 5 },
    { 19, 6 },
    { 19, 7 },
    { 19, 16 },
    { 19, 17 },
    { 19, 21 },
    { 19, 22 },
    { 19, 23 },
    { 19, 32 },
    { 19, 33 },
    { 19, 37 },
    { 19, 38 },
    { 19, 39 },
    { 19, 48 },
    { 19, 49 },
    { 19, 50 },
    { 19, 51 },
    { 19, 52 },
    { 19, 53 },
    { 19, 54 },
    { 19, 55 },
    { 19, 64 },
    { 19, 65 },
    { 19, 66 },
    { 19, 67 },
    { 19, 68 },
    { 19, 69 },
    { 19, 70 },
    { 19, 71 },
    { 19, 80 },
    { 19, 81 },
    { 19, 82 },
    { 19, 83 },
    { 19, 84 },
    { 19, 85 },
    { 19, 86 },
    { 19, 87 },
    { 19, 96 },
    { 19, 97 },
    { 19, 98 },
    { 19, 99 },
    { 19, 100 },
    { 19, 101 },
    { 19, 102 },
    { 19, 103 },
    { 19, 112 },
    { 19, 113 },
    { 19, 114 },
    { 19, 115 },
    { 19, 116 },
    { 19, 117 },
    { 19, 118 },
    { 19, 119 },
    { 32, 0 },
    { 32, 1 },
    { 32, 2 },
    { 32, 3 },
    { 32, 4 },
    { 32, 5 },
    { 32, 6 },
    { 32, 7 },
    { 32, 18 },
    { 32, 19 },
    { 32, 20 },
    { 32, 21 },
    { 32, 22 },
    { 32, 23 },
    { 32, 34 },
    { 32, 35 },
    { 32, 36 },
    { 32, 37 },
    { 32, 38 },
    { 32, 39 },
    { 32, 50 },
    { 32, 51 },
    { 32, 52 },
    { 32, 53 },
    { 32, 54 },
    { 32, 55 },
    { 32, 64 },
    { 32, 65 },
    { 32, 66 },
    { 32, 67 },
    { 32, 68 },
    { 32, 69 },
    { 32, 70 },
    { 32, 71 },
    { 32, 80 },
    { 32, 81 },
    { 32, 82 },
    { 32, 83 },
    { 32, 84 },
    { 32, 85 },
    { 32, 86 },
    { 32, 87 },
    { 32, 96 },
    { 32, 97 },
    { 32, 98 },
    { 32, 99 },
    { 32, 100 },
    { 32, 101 },
    { 32, 102 },
    { 32, 103 },
    { 32, 112 },
    { 32, 113 },
    { 32, 114 },
    { 32, 115 },
    { 32, 116 },
    { 32, 117 },
    { 32, 118 },
    { 32, 119 },
    { 33, 0 },
    { 33, 1 },
    { 33, 2 },
    { 33, 3 },
    { 33, 4 },
    { 33, 5 },
    { 33, 6 },
    { 33, 7 },
    { 33, 19 },
    { 33, 20 },
    { 33, 21 },
    { 33, 22 },
    { 33, 23 },
    { 33, 35 },
    { 33, 36 },
    { 33, 37 },
    { 33, 38 },
    { 33, 39 },
    { 33, 51 },
    { 33, 52 },
    { 33, 53 },
    { 33, 54 },
    { 33, 55 },
    { 33, 64 },
    { 33, 65 },
    { 33, 66 },
    { 33, 67 },
    { 33, 68 },
    { 33, 69 },
    { 33, 70 },
    { 33, 71 },
    { 33, 80 },
    { 33, 81 },
    { 33, 82 },
    { 33, 83 },
    { 33, 84 },
    { 33, 85 },
    { 33, 86 },
    { 33, 87 },
    { 33, 96 },
    { 33, 97 },
    { 33, 98 },
    { 33, 99 },
    { 33, 100 },
    { 33, 101 },
    { 33, 102 },
    { 33, 103 },
    { 33, 112 },
    { 33, 113 },
    { 33, 114 },
    { 33, 115 },
    { 33, 116 },
    { 33, 117 },
    { 33, 118 },
    { 33, 119 },
    { 34, 0 },
    { 34, 1 },
    { 34, 2 },
    { 34, 3 },
    { 34, 4 },
    { 34, 5 },
    { 34, 6 },
    { 34, 7 },
    { 34, 16 },
    { 34, 20 },
    { 34, 21 },
    { 34, 22 },
    { 34, 23 },
    { 34, 32 },
    { 34, 36 },
    { 34, 37 },
    { 34, 38 },
    { 34, 39 },
    { 34, 48 },
    { 34, 52 },
    { 34, 53 },
    { 34, 54 },
    { 34, 55 },
    { 34, 64 },
    { 34, 65 },
    { 34, 66 },
    { 34, 67 },
    { 34, 68 },
    { 34, 69 },
    { 34, 70 },
    { 34, 71 },
    { 34, 80 },
    { 34, 81 },
    { 34, 82 },
    { 34, 83 },
    { 34, 84 },
    { 34, 85 },
    { 34, 86 },
    { 34, 87 },
    { 34, 96 },
    { 34, 97 },
    { 34, 98 },
    { 34, 99 },
    { 34, 100 },
    { 34, 101 },
    { 34, 102 },
    { 34, 103 },
    { 34, 112 },
    { 34, 113 },
    { 34, 114 },
    { 34, 115 },
    { 34, 116 },
    { 34, 117 },
    { 34, 118 },
    { 34, 119 },
    { 35, 0 },
    { 35, 1 },
    { 35, 2 },
    { 35, 3 },
    { 35, 4 },
    { 35, 5 },
    { 35, 6 },
    { 35, 7 },
    { 35, 16 },
    { 35, 17 },
    { 35, 21 },
    { 35, 22 },
    { 35, 23 },
    { 35, 32 },
    { 35, 33 },
    { 35, 37 },
    { 35, 38 },
    { 35, 39 },
    { 35, 48 },
    { 35, 49 },
    { 35, 53 },
    { 35, 54 },
    { 35, 55 },
    { 35, 64 },
    { 35, 65 },
    { 35, 66 },
    { 35, 67 },
    { 35, 68 },
    { 35, 69 },
    { 35, 70 },
    { 35, 71 },
    { 35, 80 },
    { 35, 81 },
    { 35, 82 },
    { 35, 83 },
    { 35, 84 },
    { 35, 85 },
    { 35, 86 },
    { 35, 87 },
    { 35, 96 },
    { 35, 97 },
    { 35, 98 },
    { 35, 99 },
    { 35, 100 },
    { 35, 101 },
    { 35, 102 },
    { 35, 103 },
    { 35, 112 },
    { 35, 113 },
    { 35, 114 },
    { 35, 115 },
    { 35, 116 },
    { 35, 117 },
    { 35, 118 },
    { 35, 119 },
    { 48, 0 },
    { 48, 1 },
    { 48, 2 },
    { 48, 3 },
    { 48, 4 },
    { 48, 5 },
    { 48, 6 },
    { 48, 7 },
    { 48, 16 },
    { 48, 17 },
    { 48, 18 },
    { 48, 19 },
    { 48, 20 },
    { 48, 21 },
    { 48, 22 },
    { 48, 23 },
    { 48, 34 },
    { 48, 35 },
    { 48, 36 },
    { 48, 37 },
    { 48, 38 },
    { 48, 39 },
    { 48, 50 },
    { 48, 51 },
    { 48, 52 },
    { 48, 53 },
    { 48, 54 },
    { 48, 55 },
    { 48, 66 },
    { 48, 67 },
    { 48, 68 },
    { 48, 69 },
    { 48, 70 },
    { 48, 71 },
    { 48, 80 },
    { 48, 81 },
    { 48, 82 },
    { 48, 83 },
    { 48, 84 },
    { 48, 85 },
    { 48, 86 },
    { 48, 87 },
    { 48, 96 },
    { 48, 97 },
    { 48, 98 },
    { 48, 99 },
    { 48, 100 },
    { 48, 101 },
    { 48, 102 },
    { 48, 103 },
    { 48, 112 },
    { 48, 113 },
    { 48, 114 },
    { 48, 115 },
    { 48, 116 },
    { 48, 117 },
    { 48, 118 },
    { 48, 119 },
    { 49, 0 },
    { 49, 1 },
    { 49, 2 },
    { 49, 3 },
    { 49, 4 },
    { 49, 5 },
    { 49, 6 },
    { 49, 7 },
    { 49, 16 },
    { 49, 17 },
    { 49, 18 },
    { 49, 19 },
    { 49, 20 },
    { 49, 21 },
    { 49, 22 },
    { 49, 23 },
    { 49, 35 },
    { 49, 36 },
    { 49, 37 },
    { 49, 38 },
    { 49, 39 },
    { 49, 51 },
    { 49, 52 },
    { 49, 53 },
    { 49, 54 },
    { 49, 55 },
    { 49, 67 },
    { 49, 68 },
    { 49, 69 },
    { 49, 70 },
    { 49, 71 },
    { 49, 80 },
    { 49, 81 },
    { 49, 82 },
    { 49, 83 },
    { 49, 84 },
    { 49, 85 },
    { 49, 86 },
    { 49, 87 },
    { 49, 96 },
    { 49, 97 },
    { 49, 98 },
    { 49, 99 },
    { 49, 100 },
    { 49, 101 },
    { 49, 102 },
    { 49, 103 },
    { 49, 112 },
    { 49, 113 },
    { 49, 114 },
    { 49, 115 },
    { 49, 116 },
    { 49, 117 },
    { 49, 118 },
    { 49, 119 },
    { 50, 0 },
    { 50, 1 },
    { 50, 2 },
    { 50, 3 },
    { 50, 4 },
    { 50, 5 },
    { 50, 6 },
    { 50, 7 },
    { 50, 16 },
    { 50, 17 },
    { 50, 18 },
    { 50, 19 },
    { 50, 20 },
    { 50, 21 },
    { 50, 22 },
    { 50, 23 },
    { 50, 32 },
    { 50, 36 },
    { 50, 37 },
    { 50, 38 },
    { 50, 39 },
    { 50, 48 },
    { 50, 52 },
    { 50, 53 },
    { 50, 54 },
    { 50, 55 },
    { 50, 64 },
    { 50, 68 },
    { 50, 69 },
    { 50, 70 },
    { 50, 71 },
    { 50, 80 },
    { 50, 81 },
    { 50, 82 },
    { 50, 83 },
    { 50, 84 },
    { 50, 85 },
    { 50, 86 },
    { 50, 87 },
    { 50, 96 },
    { 50, 97 },
    { 50, 98 },
    { 50, 99 },
    { 50, 100 },
    { 50, 101 },
    { 50, 102 },
    { 50, 103 },
    { 50, 112 },
    { 50, 113 },
    { 50, 114 },
    { 50, 115 },
    { 50, 116 },
    { 50, 117 },
    { 50, 118 },
    { 50, 119 },
    { 51, 0 },
    { 51, 1 },
    { 51, 2 },
    { 51, 3 },
    { 51, 4 },
    { 51, 5 },
    { 51, 6 },
    { 51, 7 },
    { 51, 16 },
    { 51, 17 },
    { 51, 18 },
    { 51, 19 },
    { 51, 20 },
    { 51, 21 },
    { 51, 22 },
    { 51, 23 },
    { 51, 32 },
    { 51, 33 },
    { 51, 37 },
    { 51, 38 },
    { 51, 39 },
    { 51, 48 },
    { 51, 49 },
    { 51, 53 },
    { 51, 54 },
    { 51, 55 },
    { 51, 64 },
    { 51, 65 },
    { 51, 69 },
    { 51, 70 },
    { 51, 71 },
    { 51, 80 },
    { 51, 81 },
    { 51, 82 },
    { 51, 83 },
    { 51, 84 },
    { 51, 85 },
    { 51, 86 },
    { 51, 87 },
    { 51, 96 },
    { 51, 97 },
    { 51, 98 },
    { 51, 99 },
    { 51, 100 },
    { 51, 101 },
    { 51, 102 },
    { 51, 103 },
    { 51, 112 },
    { 51, 113 },
    { 51, 114 },
    { 51, 115 },
    { 51, 116 },
    { 51, 117 },
    { 51, 118 },
    { 51, 119 },
    { 64, 0 },
    { 64, 1 },
    { 64, 2 },
    { 64, 3 },
    { 64, 4 },
    { 64, 5 },
    { 64, 6 },
    { 64, 7 },
    { 64, 16 },
    { 64, 17 },
    { 64, 18 },
    { 64, 19 },
    { 64, 20 },
    { 64, 21 },
    { 64, 22 },
    { 64, 23 },
    { 64, 32 },
    { 64, 33 },
    { 64, 34 },
    { 64, 35 },
    { 64, 36 },
    { 64, 37 },
    { 64, 38 },
    { 64, 39 },
    { 64, 50 },
    { 64, 51 },
    { 64, 52 },
    { 64, 53 },
    { 64, 54 },
    { 64, 55 },
    { 64, 66 },
    { 64, 67 },
    { 64, 68 },
    { 64, 69 },
    { 64, 70 },
    { 64, 71 },
    { 64, 82 },
    { 64, 83 },
    { 64, 84 },
    { 64, 85 },
    { 64, 86 },
    { 64, 87 },
    { 64, 96 },
    { 64, 97 },
    { 64, 98 },
    { 64, 99 },
    { 64, 100 },
    { 64, 101 },
    { 64, 102 },
    { 64, 103 },
    { 64, 112 },
    { 64, 113 },
    { 64, 114 },
    { 64, 115 },
    { 64, 116 },
    { 64, 117 },
    { 64, 118 },
    { 64, 119 },
    { 65, 0 },
    { 65, 1 },
    { 65, 2 },
    { 65, 3 },
    { 65, 4 },
    { 65, 5 },
    { 65, 6 },
    { 65, 7 },
    { 65, 16 },
    { 65, 17 },
    { 65, 18 },
    { 65, 19 },
    { 65, 20 },
    { 65, 21 },
    { 65, 22 },
    { 65, 23 },
    { 65, 32 },
    { 65, 33 },
    { 65, 34 },
    { 65, 35 },
    { 65, 36 },
    { 65, 37 },
    { 65, 38 },
    { 65, 39 },
    { 65, 51 },
    { 65, 52 },
    { 65, 53 },
    { 65, 54 },
    { 65, 55 },
    { 65, 67 },
    { 65, 68 },
    { 65, 69 },
    { 65, 70 },
    { 65, 71 },
    { 65, 83 },
    { 65, 84 },
    { 65, 85 },
    { 65, 86 },
    { 65, 87 },
    { 65, 96 },
    { 65, 97 },
    { 65, 98 },
    { 65, 99 },
    { 65, 100 },
    { 65, 101 },
    { 65, 102 },
    { 65, 103 },
    { 65, 112 },
    { 65, 113 },
    { 65, 114 },
    { 65, 115 },
    { 65, 116 },
    { 65, 117 },
    { 65, 118 },
    { 65, 119 },
    { 66, 0 },
    { 66, 1 },
    { 66, 2 },
    { 66, 3 },
    { 66, 4 },
    { 66, 5 },
    { 66, 6 },
    { 66, 7 },
    { 66, 16 },
    { 66, 17 },
    { 66, 18 },
    { 66, 19 },
    { 66, 20 },
    { 66, 21 },
    { 66, 22 },
    { 66, 23 },
    { 66, 32 },
    { 66, 33 },
    { 66, 34 },
    { 66, 35 },
    { 66, 36 },
    { 66, 37 },
    { 66, 38 },
    { 66, 39 },
    { 66, 48 },
    { 66, 52 },
    { 66, 53 },
    { 66, 54 },
    { 66, 55 },
    { 66, 64 },
    { 66, 68 },
    { 66, 69 },
    { 66, 70 },
    { 66, 71 },
    { 66, 80 },
    { 66, 84 },
    { 66, 85 },
    { 66, 86 },
    { 66, 87 },
    { 66, 96 },
    { 66, 97 },
    { 66, 98 },
    { 66, 99 },
    { 66, 100 },
    { 66, 101 },
    { 66, 102 },
    { 66, 103 },
    { 66, 112 },
    { 66, 113 },
    { 66, 114 },
    { 66, 115 },
    { 66, 116 },
    { 66, 117 },
    { 66, 118 },
    { 66, 119 },
    { 67, 0 },
    { 67, 1 },
    { 67, 2 },
    { 67, 3 },
    { 67, 4 },
    { 67, 5 },
    { 67, 6 },
    { 67, 7 },
    { 67, 16 },
    { 67, 17 },
    { 67, 18 },
    { 67, 19 },
    { 67, 20 },
    { 67, 21 },
    { 67, 22 },
    { 67, 23 },
    { 67, 32 },
    { 67, 33 },
    { 67, 34 },
    { 67, 35 },
    { 67, 36 },
    { 67, 37 },
    { 67, 38 },
    { 67, 39 },
    { 67, 48 },
    { 67, 49 },
    { 67, 53 },
    { 67, 54 },
    { 67, 55 },
    { 67, 64 },
    { 67, 65 },
    { 67, 69 },
    { 67, 70 },
    { 67, 71 },
    { 67, 80 },
    { 67, 81 },
    { 67, 85 },
    { 67, 86 },
    { 67, 87 },
    { 67, 96 },
    { 67, 97 },
    { 67, 98 },
    { 67, 99 },
    { 67, 100 },
    { 67, 101 },
    { 67, 102 },
    { 67, 103 },
    { 67, 112 },
    { 67, 113 },
    { 67, 114 },
    { 67, 115 },
    { 67, 116 },
    { 67, 117 },
    { 67, 118 },
    { 67, 119 },
    { 80, 0 },
    { 80, 1 },
    { 80, 2 },
    { 80, 3 },
    { 80, 4 },
    { 80, 5 },
    { 80, 6 },
    { 80, 7 },
    { 80, 16 },
    { 80, 17 },
    { 80, 18 },
    { 80, 19 },
    { 80, 20 },
    { 80, 21 },
    { 80, 22 },
    { 80, 23 },
    { 80, 32 },
    { 80, 33 },
    { 80, 34 },
    { 80, 35 },
    { 80, 36 },
    { 80, 37 },
    { 80, 38 },
    { 80, 39 },
    { 80, 48 },
    { 80, 49 },
    { 80, 50 },
    { 80, 51 },
    { 80, 52 },
    { 80, 53 },
    { 80, 54 },
    { 80, 55 },
    { 80, 66 },
    { 80, 67 },
    { 80, 68 },
    { 80, 69 },
    { 80, 70 },
    { 80, 71 },
    { 80, 82 },
    { 80, 83 },
    { 80, 84 },
    { 80, 85 },
    { 80, 86 },
    { 80, 87 },
    { 80, 98 },
    { 80, 99 },
    { 80, 100 },
    { 80, 101 },
    { 80, 102 },
    { 80, 103 },
    { 80, 112 },
    { 80, 113 },
    { 80, 114 },
    { 80, 115 },
    { 80, 116 },
    { 80, 117 },
    { 80, 118 },
    { 80, 119 },
    { 81, 0 },
    { 81, 1 },
    { 81, 2 },
    { 81, 3 },
    { 81, 4 },
    { 81, 5 },
    { 81, 6 },
    { 81, 7 },
    { 81, 16 },
    { 81, 17 },
    { 81, 18 },
    { 81, 19 },
    { 81, 20 },
    { 81, 21 },
    { 81, 22 },
    { 81, 23 },
    { 81, 32 },
    { 81, 33 },
    { 81, 34 },
    { 81, 35 },
    { 81, 36 },
    { 81, 37 },
    { 81, 38 },
    { 81, 39 },
    { 81, 48 },
    { 81, 49 },
    { 81, 50 },
    { 81, 51 },
    { 81, 52 },
    { 81, 53 },
    { 81, 54 },
    { 81, 55 },
    { 81, 67 },
    { 81, 68 },
    { 81, 69 },
    { 81, 70 },
    { 81, 71 },
    { 81, 83 },
    { 81, 84 },
    { 81, 85 },
    { 81, 86 },
    { 81, 87 },
    { 81, 99 },
    { 81, 100 },
    { 81, 101 },
    { 81, 102 },
    { 81, 103 },
    { 81, 112 },
    { 81, 113 },
    { 81, 114 },
    { 81, 115 },
    { 81, 116 },
    { 81, 117 },
    { 81, 118 },
    { 81, 119 },
    { 82, 0 },
    { 82, 1 },
    { 82, 2 },
    { 82, 3 },
    { 82, 4 },
    { 82, 5 },
    { 82, 6 },
    { 82, 7 },
    { 82, 16 },
    { 82, 17 },
    { 82, 18 },
    { 82, 19 },
    { 82, 20 },
    { 82, 21 },
    { 82, 22 },
    { 82, 23 },
    { 82, 32 },
    { 82, 33 },
    { 82, 34 },
    { 82, 35 },
    { 82, 36 },
    { 82, 37 },
    { 82, 38 },
    { 82, 39 },
    { 82, 48 },
    { 82, 49 },
    { 82, 50 },
    { 82, 51 },
    { 82, 52 },
    { 82, 53 },
    { 82, 54 },
    { 82, 55 },
    { 82, 64 },
    { 82, 68 },
    { 82, 69 },
    { 82, 70 },
    { 82, 71 },
    { 82, 80 },
    { 82, 84 },
    { 82, 85 },
    { 82, 86 },
    { 82, 87 },
    { 82, 96 },
    { 82, 100 },
    { 82, 101 },
    { 82, 102 },
    { 82, 103 },
    { 82, 112 },
    { 82, 113 },
    { 82, 114 },
    { 82, 115 },
    { 82, 116 },
    { 82, 117 },
    { 82, 118 },
    { 82, 119 },
    { 83, 0 },
    { 83, 1 },
    { 83, 2 },
    { 83, 3 },
    { 83, 4 },
    { 83, 5 },
    { 83, 6 },
    { 83, 7 },
    { 83, 16 },
    { 83, 17 },
    { 83, 18 },
    { 83, 19 },
    { 83, 20 },
    { 83, 21 },
    { 83, 22 },
    { 83, 23 },
    { 83, 32 },
    { 83, 33 },
    { 83, 34 },
    { 83, 35 },
    { 83, 36 },
    { 83, 37 },
    { 83, 38 },
    { 83, 39 },
    { 83, 48 },
    { 83, 49 },
    { 83, 50 },
    { 83, 51 },
    { 83, 52 },
    { 83, 53 },
    { 83, 54 },
    { 83, 55 },
    { 83, 64 },
    { 83, 65 },
    { 83, 69 },
    { 83, 70 },
    { 83, 71 },
    { 83, 80 },
    { 83, 81 },
    { 83, 85 },
    { 83, 86 },
    { 83, 87 },
    { 83, 96 },
    { 83, 97 },
    { 83, 101 },
    { 83, 102 },
    { 83, 103 },
    { 83, 112 },
    { 83, 113 },
    { 83, 114 },
    { 83, 115 },
    { 83, 116 },
    { 83, 117 },
    { 83, 118 },
    { 83, 119 },
    { 96, 0 },
    { 96, 1 },
    { 96, 2 },
    { 96, 3 },
    { 96, 4 },
    { 96, 5 },
    { 96, 6 },
    { 96, 7 },
    { 96, 16 },
    { 96, 17 },
    { 96, 18 },
    { 96, 19 },
    { 96, 20 },
    { 96, 21 },
    { 96, 22 },
    { 96, 23 },
    { 96, 32 },
    { 96, 33 },
    { 96, 34 },
    { 96, 35 },
    { 96, 36 },
    { 96, 37 },
    { 96, 38 },
    { 96, 39 },
    { 96, 48 },
    { 96, 49 },
    { 96, 50 },
    { 96, 51 },
    { 96, 52 },
    { 96, 53 },
    { 96, 54 },
    { 96, 55 },
    { 96, 64 },
    { 96, 65 },
    { 96, 66 },
    { 96, 67 },
    { 96, 68 },
    { 96, 69 },
    { 96, 70 },
    { 96, 71 },
    { 96, 82 },
    { 96, 83 },
    { 96, 84 },
    { 96, 85 },
    { 96, 86 },
    { 96, 87 },
    { 96, 98 },
    { 96, 99 },
    { 96, 100 },
    { 96, 101 },
    { 96, 102 },
    { 96, 103 },
    { 96, 114 },
    { 96, 115 },
    { 96, 116 },
    { 96, 117 },
    { 96, 118 },
    { 96, 119 },
    { 97, 0 },
    { 97, 1 },
    { 97, 2 },
    { 97, 3 },
    { 97, 4 },
    { 97, 5 },
    { 97, 6 },
    { 97, 7 },
    { 97, 16 },
    { 97, 17 },
    { 97, 18 },
    { 97, 19 },
    { 97, 20 },
    { 97, 21 },
    { 97, 22 },
    { 97, 23 },
    { 97, 32 },
    { 97, 33 },
    { 97, 34 },
    { 97, 35 },
    { 97, 36 },
    { 97, 37 },
    { 97, 38 },
    { 97, 39 },
    { 97, 48 },
    { 97, 49 },
    { 97, 50 },
    { 97, 51 },
    { 97, 52 },
    { 97, 53 },
    { 97, 54 },
    { 97, 55 },
    { 97, 64 },
    { 97, 65 },
    { 97, 66 },
    { 97, 67 },
    { 97, 68 },
    { 97, 69 },
    { 97, 70 },
    { 97, 71 },
    { 97, 83 },
    { 97, 84 },
    { 97, 85 },
    { 97, 86 },
    { 97, 87 },
    { 97, 99 },
    { 97, 100 },
    { 97, 101 },
    { 97, 102 },
    { 97, 103 },
    { 97, 115 },
    { 97, 116 },
    { 97, 117 },
    { 97, 118 },
    { 97, 119 },
    { 98, 0 },
    { 98, 1 },
    { 98, 2 },
    { 98, 3 },
    { 98, 4 },
    { 98, 5 },
    { 98, 6 },
    { 98, 7 },
    { 98, 16 },
    { 98, 17 },
    { 98, 18 },
    { 98, 19 },
    { 98, 20 },
    { 98, 21 },
    { 98, 22 },
    { 98, 23 },
    { 98, 32 },
    { 98, 33 },
    { 98, 34 },
    { 98, 35 },
    { 98, 36 },
    { 98, 37 },
    { 98, 38 },
    { 98, 39 },
    { 98, 48 },
    { 98, 49 },
    { 98, 50 },
    { 98, 51 },
    { 98, 52 },
    { 98, 53 },
    { 98, 54 },
    { 98, 55 },
    { 98, 64 },
    { 98, 65 },
    { 98, 66 },
    { 98, 67 },
    { 98, 68 },
    { 98, 69 },
    { 98, 70 },
    { 98, 71 },
    { 98, 80 },
    { 98, 84 },
    { 98, 85 },
    { 98, 86 },
    { 98, 87 },
    { 98, 96 },
    { 98, 100 },
    { 98, 101 },
    { 98, 102 },
    { 98, 103 },
    { 98, 112 },
    { 98, 116 },
    { 98, 117 },
    { 98, 118 },
    { 98, 119 },
    { 99, 0 },
    { 99, 1 },
    { 99, 2 },
    { 99, 3 },
    { 99, 4 },
    { 99, 5 },
    { 99, 6 },
    { 99, 7 },
    { 99, 16 },
    { 99, 17 },
    { 99, 18 },
    { 99, 19 },
    { 99, 20 },
    { 99, 21 },
    { 99, 22 },
    { 99, 23 },
    { 99, 32 },
    { 99, 33 },
    { 99, 34 },
    { 99, 35 },
    { 99, 36 },
    { 99, 37 },
    { 99, 38 },
    { 99, 39 },
    { 99, 48 },
    { 99, 49 },
    { 99, 50 },
    { 99, 51 },
    { 99, 52 },
    { 99, 53 },
    { 99, 54 },
    { 99, 55 },
    { 99, 64 },
    { 99, 65 },
    { 99, 66 },
    { 99, 67 },
    { 99, 68 },
    { 99, 69 },
    { 99, 70 },
    { 99, 71 },
    { 99, 80 },
    { 99, 81 },
    { 99, 85 },
    { 99, 86 },
    { 99, 87 },
    { 99, 96 },
    { 99, 97 },
    { 99, 101 },
    { 99, 102 },
    { 99, 103 },
    { 99, 112 },
    { 99, 113 },
    { 99, 117 },
    { 99, 118 },
    { 99, 119 },
    { 112, 0 },
    { 112, 1 },
    { 112, 2 },
    { 112, 3 },
    { 112, 4 },
    { 112, 5 },
    { 112, 6 },
    { 112, 7 },
    { 112, 16 },
    { 112, 17 },
    { 112, 18 },
    { 112, 19 },
    { 112, 20 },
    { 112, 21 },
    { 112, 22 },
    { 112, 23 },
    { 112, 32 },
    { 112, 33 },
    { 112, 34 },
    { 112, 35 },
    { 112, 36 },
    { 112, 37 },
    { 112, 38 },
    { 112, 39 },
    { 112, 48 },
    { 112, 49 },
    { 112, 50 },
    { 112, 51 },
    { 112, 52 },
    { 112, 53 },
    { 112, 54 },
    { 112, 55 },
    { 112, 64 },
    { 112, 65 },
    { 112, 66 },
    { 112, 67 },
    { 112, 68 },
    { 112, 69 },
    { 112, 70 },
    { 112, 71 },
    { 112, 80 },
    { 112, 81 },
    { 112, 82 },
    { 112, 83 },
    { 112, 84 },
    { 112, 85 },
    { 112, 86 },
    { 112, 87 },
    { 112, 98 },
    { 112, 99 },
    { 112, 100 },
    { 112, 101 },
    { 112, 102 },
    { 112, 103 },
    { 112, 114 },
    { 112, 115 },
    { 112, 116 },
    { 112, 117 },
    { 112, 118 },
    { 112, 119 },
    { 113, 0 },
    { 113, 1 },
    { 113, 2 },
    { 113, 3 },
    { 113, 4 },
    { 113, 5 },
    { 113, 6 },
    { 113, 7 },
    { 113, 16 },
    { 113, 17 },
    { 113, 18 },
    { 113, 19 },
    { 113, 20 },
    { 113, 21 },
    { 113, 22 },
    { 113, 23 },
    { 113, 32 },
    { 113, 33 },
    { 113, 34 },
    { 113, 35 },
    { 113, 36 },
    { 113, 37 },
    { 113, 38 },
    { 113, 39 },
    { 113, 48 },
    { 113, 49 },
    { 113, 50 },
    { 113, 51 },
    { 113, 52 },
    { 113, 53 },
    { 113, 54 },
    { 113, 55 },
    { 113, 64 },
    { 113, 65 },
    { 113, 66 },
    { 113, 67 },
    { 113, 68 },
    { 113, 69 },
    { 113, 70 },
    { 113, 71 },
    { 113, 80 },
    { 113, 81 },
    { 113, 82 },
    { 113, 83 },
    { 113, 84 },
    { 113, 85 },
    { 113, 86 },
    { 113, 87 },
    { 113, 99 },
    { 113, 100 },
    { 113, 101 },
    { 113, 102 },
    { 113, 103 },
    { 113, 115 },
    { 113, 116 },
    { 113, 117 },
    { 113, 118 },
    { 113, 119 },
    { 114, 0 },
    { 114, 1 },
    { 114, 2 },
    { 114, 3 },
    { 114, 4 },
    { 114, 5 },
    { 114, 6 },
    { 114, 7 },
    { 114, 16 },
    { 114, 17 },
    { 114, 18 },
    { 114, 19 },
    { 114, 20 },
    { 114, 21 },
    { 114, 22 },
    { 114, 23 },
    { 114, 32 },
    { 114, 33 },
    { 114, 34 },
    { 114, 35 },
    { 114, 36 },
    { 114, 37 },
    { 114, 38 },
    { 114, 39 },
    { 114, 48 },
    { 114, 49 },
    { 114, 50 },
    { 114, 51 },
    { 114, 52 },
    { 114, 53 },
    { 114, 54 },
    { 114, 55 },
    { 114, 64 },
    { 114, 65 },
    { 114, 66 },
    { 114, 67 },
    { 114, 68 },
    { 114, 69 },
    { 114, 70 },
    { 114, 71 },
    { 114, 80 },
    { 114, 81 },
    { 114, 82 },
    { 114, 83 },
    { 114, 84 },
    { 114, 85 },
    { 114, 86 },
    { 114, 87 },
    { 114, 96 },
    { 114, 100 },
    { 114, 101 },
    { 114, 102 },
    { 114, 103 },
    { 114, 112 },
    { 114, 116 },
    { 114, 117 },
    { 114, 118 },
    { 114, 119 },
    { 115, 0 },
    { 115, 1 },
    { 115, 2 },
    { 115, 3 },
    { 115, 4 },
    { 115, 5 },
    { 115, 6 },
    { 115, 7 },
    { 115, 16 },
    { 115, 17 },
    { 115, 18 },
    { 115, 19 },
    { 115, 20 },
    { 115, 21 },
    { 115, 22 },
    { 115, 23 },
    { 115, 32 },
    { 115, 33 },
    { 115, 34 },
    { 115, 35 },
    { 115, 36 },
    { 115, 37 },
    { 115, 38 },
    { 115, 39 },
    { 115, 48 },
    { 115, 49 },
    { 115, 50 },
    { 115, 51 },
    { 115, 52 },
    { 115, 53 },
    { 115, 54 },
    { 115, 55 },
    { 115, 64 },
    { 115, 65 },
    { 115, 66 },
    { 115, 67 },
    { 115, 68 },
    { 115, 69 },
    { 115, 70 },
    { 115, 71 },
    { 115, 80 },
    { 115, 81 },
    { 115, 82 },
    { 115, 83 },
    { 115, 84 },
    { 115, 85 },
    { 115, 86 },
    { 115, 87 },
    { 115, 96 },
    { 115, 97 },
    { 115, 101 },
    { 115, 102 },
    { 115, 103 },
    { 115, 112 },
    { 115, 113 },
    { 115, 117 },
    { 115, 118 },
    { 115, 119 },
};

static void VIterateThroughPositions
    (
    color   side,   // IN | side to move
    PfnBool pfn     // IN | function to be called for each position
    )
    {
    CBoard  *pbd;
    square  sq;
    int     cpu;

    // Clear all boards
    for (cpu = 0; cpu < cCPUS; cpu ++)
        {
        pbd = & rgBoards[cpu];
        for (sq = A1; sq != XX; sq = rgsqNextSquare [sq])
            {
            pbd->m_rgColor[sq] = x_colorNeutral;
            pbd->m_rgBoard[sq] = x_pieceNone;
            }
        }

    // Now loop through all possible kings combinations
    if (pbd->m_rgfHasPawns[x_colorWhite] || pbd->m_rgfHasPawns[x_colorBlack])
        {
        prgsqKings = rgsqKings1806;
        cKings = 1806;
        }
    else if (pbd->m_fReverse)
        {
        prgsqKings = rgsqKings564;
        cKings = 564;
        }
    else
        {
        prgsqKings = rgsqKings462;
        cKings = 462;
        }
#if defined (SMP)
    if (1 == cCPUS)
        {
        if (x_colorWhite == side)
            {
            for (int i = 0; i < cKings; i ++)
                {
                VPrintDone (i, cKings);
                VIterateThroughPositions (prgsqKings[i][x_colorWhite], prgsqKings[i][x_colorBlack],
                                          rgBoards, side, pfn);
                }
            }
        else
            {
            for (int i = 0; i < cKings; i ++)
                {
                VPrintDone (i, cKings);
                VIterateThroughPositions (prgsqKings[i][x_colorBlack], prgsqKings[i][x_colorWhite],
                                          rgBoards, side, pfn);
                }
            }
        }
    else
        {
        // Start threads 1, ..., N-1
        iKings = 0;
        for (cpu = 0; cpu < cCPUS; cpu ++)
            {
            rgBoards[cpu].m_side = side;
            rgBoards[cpu].m_pfn  = pfn;
            Lock   (lockKings);
            UnLock (lockKings);
            rgBoards[cpu].m_fHasWork = true;
            }
        // Use this thread to perform calculations, too
        VDoWork (&rgBoards[cCPUS-1]);
        // Wait for other threads to finish
        for (cpu = 0; cpu < cCPUS-1; cpu ++)
            {
            while (rgBoards[cpu].m_fHasWork)
                usleep(100);
            }
        }
#else
    if (x_colorWhite == side)
        {
        for (int i = 0; i < cKings; i ++)
            {
            VPrintDone (i, cKings);
            VIterateThroughPositions (prgsqKings[i][x_colorWhite], prgsqKings[i][x_colorBlack],
                                      rgBoards, side, pfn);
            }
        }
    else
        {
        for (int i = 0; i < cKings; i ++)
            {
            VPrintDone (i, cKings);
            VIterateThroughPositions (prgsqKings[i][x_colorBlack], prgsqKings[i][x_colorWhite],
                                      rgBoards, side, pfn);
            }
        }
#endif
    }

//-----------------------------------------------------------------------------

void CBoard::VPrintFEN (color side)
    {
    char   rgszFEN[128];
    char   *pch;
    square sq, sqLast;

    pch = rgszFEN;
    sq = sqLast = A8;
    for (;;)
        {
        if (x_pieceNone != m_rgBoard[sq])
            {
            if (sq != sqLast)
                * pch ++ = (char) (sq - sqLast + '0');
            * pch ++ = (char) ("PNBRQK"[m_rgBoard[sq]-x_piecePawn]+(x_colorBlack==m_rgColor[sq]?0x20:0));
            sqLast = sq + 1;
            }
        if ((sq & 7) == 7)
            {
            * pch ++ = '/';
            if (H1 == sq)
                break;
            else
                {
                sq -= 0x17;
                sqLast = sq;
                }
            }
        else
            sq ++;
        }
    * pch ++ = ' ';
    * pch ++ = (x_colorWhite == side) ? 'w' : 'b';
    if (XX != m_sqEnPassant)
        {
        * pch ++ = ' ';
        * pch ++ = (char) ('a'+Column(m_sqEnPassant-rgiPawnStep[side]));
        * pch ++ = (char) ('1'+Row(m_sqEnPassant-rgiPawnStep[side]));
        }
    * pch = '\0';
    Lock (lockIO);
    puts (rgszFEN);
    fflush (stdout);
    UnLock (lockIO);
    }

//-----------------------------------------------------------------------------

void CBoard::VPrintPosition ()
    {
    int i;

    printf ("White pieces: %02X", m_rgsqPieceList[x_colorWhite][0]);
    for (i = m_rgPieceCount[x_colorWhite]-1; i > 0; i --)
        printf (" %02X", m_rgsqPieceList[x_colorWhite][i]);
    printf ("\nBlack pieces: %02X", m_rgsqPieceList[x_colorBlack][0]);
    for (i = m_rgPieceCount[x_colorBlack]-1; i > 0; i --)
        printf (" %02X", m_rgsqPieceList[x_colorBlack][i]);
    printf ("\nEn passant %02X\n", m_sqEnPassant);
    }

//-----------------------------------------------------------------------------
//
//  Calculate index of a position

INLINE INDEX CBoard::IndCalc
    (
    color side
    )
    {
    CSqInfo ssqInfo[2];

    ssqInfo[x_colorWhite].m_pSq    = m_rgsqPieceList[x_colorWhite];
    ssqInfo[x_colorWhite].m_pBoard = m_rgBoard;
    ssqInfo[x_colorBlack].m_pSq    = m_rgsqPieceList[x_colorBlack];
    ssqInfo[x_colorBlack].m_pBoard = m_rgBoard;
    return  m_rgpfnCalcIndex[side] ((square*) &ssqInfo[x_colorWhite],
                                    (square*) &ssqInfo[x_colorBlack],
                                    Linear(m_sqEnPassant), false);
    }

INLINE INDEX CBoard::IndCalc
    (
    int     iTb,
    color   side,
    square  *psqW,
    square  *psqB,
    square  sqEnP,
    int     fInverse
    )
    {
    CSqInfo ssqInfo[2];

    ssqInfo[x_colorWhite].m_pSq    = psqW;
    ssqInfo[x_colorWhite].m_pBoard = m_rgBoard;
    ssqInfo[x_colorBlack].m_pSq    = psqB;
    ssqInfo[x_colorBlack].m_pBoard = m_rgBoard;
    return PfnIndCalc(iTb, side) ((square*) &ssqInfo[x_colorWhite],
                                  (square*) &ssqInfo[x_colorBlack],
                                  sqEnP,
                                  fInverse);
    }

//-----------------------------------------------------------------------------
//
//  Calculate # of positions

static bool FCalcMaxIndex
    (
    CBoard  *pbd,
    color   side,
    color   xside
    )
    {
    INDEX   ind;

    // Valid position?
    if (pbd->FAtacked (pbd->m_rgsqPieceList[xside][0], side))
        return false;

    ind = pbd->IndCalc (side);

    if (ind > pbd->m_cb)
        pbd->m_cb = ind;

    // In theory, can print error message immediatelly,
    // but it's better to calculate how many memory necessary
    if (ind >= cbCalc)
        return true;

    // Mark position as valid
    rgpbFile[side][ind] = (BYTE) bev_draw;
    if (!pbd->FAtacked (pbd->m_rgsqPieceList[side][0], xside))
        return true;

    // Generate all pseudolegal moves
    // UNDONE: first generate only king moves, as usually king can escape
    CMove   *pmv;
    CMove   *pmvEnd;

    pmvEnd = pbd->PmvGenPseudoLegalMoves (pbd->m_rgmv, side);

    // Now iterate through all of them
    for (pmv = pbd->m_rgmv; pmv < pmvEnd; pmv ++)
        {
        pbd->VMakeMove (pmv);
        // Illegal move?
        if (! pbd->FAtacked (pbd->m_rgsqPieceList[side][0], xside))
            {
            pbd->VUnmakeMove (pmv);
            return true;
            }
        pbd->VUnmakeMove (pmv);
        }

    // No legal moves
    rgpbFile[side][ind] = (BYTE) bev_li0;
    return true;
    }

#if defined(VERIFY_INDEX)
//  Check whether several positions map to one index;

static INDEX iDuplicateIndex = (INDEX)(-1);
static bool fPrintDuplicates = false;
static INDEX    maxTbSize = 0;

static bool FCheckIndex ( CBoard *pbd, color side, color xside )
{
    INDEX   ind;

    // Valid position?
    if (pbd->FAtacked (pbd->m_rgsqPieceList[xside][0], side))
        return false;

    ind = pbd->IndCalc (side);

    if(ind >= maxTbSize) {
        printf("Index " DEC_INDEX_FORMAT "too large\n",ind);
        exit(1);
    }

    if (ind > pbd->m_cb)
        pbd->m_cb = ind;

    if(ind == iDuplicateIndex && fPrintDuplicates) {
        printf("Index = " DEC_INDEX_FORMAT "\n",ind);
        pbd->VPrintPosition();
        pbd->VPrintFEN(side);
    }
    else {
        if(bvArg.FGetBit(ind) && iDuplicateIndex == (INDEX)(-1))
            iDuplicateIndex = ind;
        bvArg.VSetBit(ind);
    }

    return true;
}
#endif

//-----------------------------------------------------------------------------
//
//  Calculate # of positions

static bool L_FCalcMaxIndex
    (
    CBoard  *pbd,
    color   side,
    color   xside
    )
    {
    INDEX   ind;

    // Valid position?
    if (pbd->FAtacked (pbd->m_rgsqPieceList[xside][0], side))
        return false;

    ind = pbd->IndCalc (side);
    ind <<= 1;
    if (ind > pbd->m_cb)
        pbd->m_cb = ind;

    // In theory, can print error message immediatelly,
    // but it's better to calculate how many memory necessary
    if (ind >= cbCalc)
        return true;

    // Mark position as valid
    rgpbFile[side][ind] = rgpbFile[side][ind+1] = 0;
    if (!pbd->FAtacked (pbd->m_rgsqPieceList[side][0], xside))
        return true;

    // Generate all pseudolegal moves
    // UNDONE: first generate only king moves, as usually king can escape
    CMove   *pmv;
    CMove   *pmvEnd;

    pmvEnd = pbd->PmvGenPseudoLegalMoves (pbd->m_rgmv, side);

    // Now iterate through all of them
    for (pmv = pbd->m_rgmv; pmv < pmvEnd; pmv ++)
        {
        pbd->VMakeMove (pmv);
        // Illegal move?
        if (! pbd->FAtacked (pbd->m_rgsqPieceList[side][0], xside))
            {
            pbd->VUnmakeMove (pmv);
            return true;
            }
        pbd->VUnmakeMove (pmv);
        }

    // No legal moves
    rgpbFile[side][ind]   = (BYTE) L_bev_li0;
    rgpbFile[side][ind+1] = (BYTE) (L_bev_li0 >> 8);
    return true;
    }

//-----------------------------------------------------------------------------
//
//  Write entire file to disk

static void VWriteFile
    (
    color   side,
    BYTE    *pb,
    int     iTb
    )
    {
    file fp;

    fp = rgFp[side];
    // Shrink kppkp
#if !defined (KPPKP_16BIT)
    if (tbid_kppkp == iTb)
        {
        int cbMax;

        cbMax = rgCb[side]/2;
        for (int cb = 0; cb < cbMax; cb ++)
            {
            int tbValue = L_Get (pb + 2*cb);

            if (0 == tbValue)
                pb [cb] = (BYTE) 0;
            else if (L_bev_broken == tbValue)
                pb [cb] = (BYTE) bev_broken;
            else if (-32640 == tbValue)
                pb [cb] = (BYTE) bev_limaxx;
            else if (32640 == tbValue)
                pb [cb] = (BYTE) bev_miminx;
            else if (tbValue > 32640 && tbValue <= L_bev_mi1)
                pb [cb] = (BYTE) (tbValue - 32640);
            else if (tbValue < -32640 && tbValue >= L_bev_li0)
                pb [cb] = (BYTE) (tbValue + 32640);
            else
                {
                printf ("Value out of range: side=%d index=" HEX_INDEX_FORMAT
                        " value=%d\n", side, cb, tbValue);
                pb [cb] = (BYTE) bev_broken;
                }
            }
        rgCb[side] /= 2;
        }
#endif
    // Now write the file
    f_write (pb, rgCb[side], fp, 0);
    }

//-----------------------------------------------------------------------------
//
//  Main loop - make move and probe table with opponent to move

// Print position if something is wrong

static void VInvalid
    (
    CBoard  *pbd,
    char    *pchText,
    color   side,
    INDEX   ind,
    int     tbtValue,
    int     tbtTable
    )
    {
    Lock (lockIO);
    printf ("\n*** %s: index = " HEX_INDEX_FORMAT
            " value = %02X table value = %02X side = %d\n",
            pchText, ind, tbtValue, tbtTable, side);
    pbd->VPrintPosition ();
    UnLock (lockIO);
    fflush (stdout);
    exit (1);
    }

// Convert 16-bytes score into string

static char *SzScore (char *rgBuffer, int bev)
{
    if (L_bev_draw == bev)
        sprintf (rgBuffer, "draw");
    else if (L_bev_broken == bev)
        sprintf (rgBuffer, "broken");
    else if (bev >= L_bev_mimin && bev <= L_bev_mi1)
        sprintf (rgBuffer, "mate in %d", L_bev_mi1 - bev + 1);
    else if (bev >= L_bev_li0 && bev <= L_bev_limax)
        sprintf (rgBuffer, "lost in %d", L_bev_mi1 + bev);
    else
        sprintf (rgBuffer, "?%d", bev);
    return rgBuffer;
}

// Print position if something is wrong

static void VInvalid16
    (
    CBoard  *pbd,
    char    *pchText,
    color   side,
    INDEX   ind,
    int     tbtValue,
    int     tbtTable
    )
    {
    char rgBuf1[80], rgBuf2[80];

    Lock (lockIO);
    printf ("\n*** %s: index = " HEX_INDEX_FORMAT "\n"
            "\tvalue = %02X '%s'\n\ttable value = %02X '%s'\n\tside = %d\n",
            pchText, ind,
            tbtValue, SzScore(rgBuf1, tbtValue),
            tbtTable, SzScore(rgBuf2, tbtTable),
            side);
    UnLock (lockIO);
    fflush (stdout);
    pbd->VPrintFEN (side);
    fflush (stdout);
    exit (1);
    }


//  Routine to make *minimal* quiet move,
//  do not bother with en promotion/capture.
//  Modify ONLY board structures that are necessary for index calculating.

INLINE void CBoard::VMinimalMakeQuietMove
    (
    CMove   *pmv,
    color   side
    )
    {
    square  sqFrom = pmv->m_sqFrom;
    square  sqTo = pmv->m_sqTo;

    pmv->m_sqSaveEnPassant = m_sqEnPassant;
    m_rgsqPieceList[side][pmv->m_iIndexFrom] = sqTo;
    // m_rgColor[sqTo] = side;
    m_rgBoard[sqTo] = m_rgBoard[sqFrom];
    // m_rgColor[sqFrom] = x_colorNeutral;
    // m_rgBoard[sqFrom] = x_pieceNone;

    // Set en passant
    VSetEnPassant (sqFrom, sqTo, side);
    }

INLINE void CBoard::VMinimalMakeQuietUnMove ( CMove *pmv, color side )
{
    square  sqFrom = pmv->m_sqFrom;
    square  sqTo = pmv->m_sqTo;

    m_rgsqPieceList[side][pmv->m_iIndexFrom] = sqTo;
    // m_rgColor[sqTo] = side;
    m_rgBoard[sqTo] = m_rgBoard[sqFrom];
    // m_rgColor[sqFrom] = x_colorNeutral;
    // m_rgBoard[sqFrom] = x_pieceNone;

    // Set en passant if the forward move would expose e.p.
    // Note that this creates an illegal position, so the
    // e.p should be cleared later...
    pmv->m_sqSaveEnPassant = m_sqEnPassant;
    VSetEnPassant (sqTo, sqFrom, side);
}

INLINE void CBoard::VUnmakeMinimalQuietMove
    (
    CMove   *pmv,
    color   side
    )
    {
    square  sqFrom = pmv->m_sqFrom;
    square  sqTo = pmv->m_sqTo;

    m_rgsqPieceList[side][pmv->m_iIndexFrom] = sqFrom;
    // m_rgColor[sqFrom] = side;
    // m_rgBoard[sqFrom] = m_rgBoard[sqTo];
    // m_rgColor[sqTo] = x_colorNeutral;
    m_rgBoard[sqTo] = x_pieceNone;
    m_sqEnPassant = pmv->m_sqSaveEnPassant;
    }

//-----------------------------------------------------------------------------
//
// Promote information from opposite table when everything is in memory

static bool FProbeOpposite
    (
    CBoard  *pbd,
    color   side,
    color   xside
    )
    {
    CMove   *pmv;
    CMove   *pmvEnd;
    tb_t    *ptbt;
    int     tbtBest;
    BYTE    *pbOpposite;
    INDEX   ind;
    INDEX   indMax;

    // Calculate index of current position
    ind = pbd->IndCalc (side);

    // Possible legal position?
    if (ind >= rgCb [side])
        {
        Assert (pbd->FAtacked (pbd->m_rgsqPieceList[xside][0], side));
        return false;
        }
    ptbt = (tb_t*) rgpbFile[side] + ind;
    tbtBest = *ptbt;

    // Legal position?
    if (bev_broken == tbtBest)
        {
        Assert (pbd->FAtacked (pbd->m_rgsqPieceList[xside][0], side));
        return false;
        }
    Assert (!pbd->FAtacked (pbd->m_rgsqPieceList[xside][0], side));

    // Is it necessary to iterate?
#if 1
    if (bev_limaxx != tbtBest && bev_miminx != tbtBest && (tbtBest <= tbtLo || tbtBest > tbtHi))
        return true;
#endif

    // Generate all pseudolegal moves
    pmvEnd = pbd->PmvGenPseudoLegalMoves (pbd->m_rgmv, side);

    // Now iterate through all of them
    pbOpposite = rgpbFile[xside];
    indMax = rgCb [xside];
    tbtBest = -65536;
    for (pmv = pbd->m_rgmv; pmv < pmvEnd; pmv ++)
        {
        // If capture/promotion, have to probe minor;
        // Otherwise, probe opposite
        if (0 == pmv->m_ulFlags)
            {
            int tbtScore;

            pbd->VMinimalMakeQuietMove (pmv, side);

            // Probe opposite table
            ind = pbd->IndCalc (xside);
            
            // Possible legal move?
            if (ind < indMax)
                {
                tbtScore = (tb_t) pbOpposite [ind];
                // Legal move?
                if (bev_broken != tbtScore)
                    {
                    tbtScore = TbtReplaceResult (tbtScore);
                    if (FBetter (tbtBest, tbtScore))
                        tbtBest = tbtScore;
                    }
                }
            pbd->VUnmakeMinimalQuietMove (pmv, side);
            }
        else
            {
            int iTb;

            pbd->VMakeMove (pmv);
            // Illegal move?
            if (pbd->FAtacked (pbd->m_rgsqPieceList[side][0], xside))
                {
                pbd->VUnmakeMove (pmv);
                continue;
                }
            // Probe minor table
            iTb = IDescFindFromCounters (pbd->m_rgiCount);
            if (0 != iTb)
                {
                INDEX indexChild;
                int tbtScore;

                if (iTb < 0)
                    {
                    iTb = -iTb;
                    indexChild = pbd->IndCalc (iTb, side,
                                               pbd->m_rgsqPieceList[x_colorBlack],
                                               pbd->m_rgsqPieceList[x_colorWhite],
                                               XX, true);
                    tbtScore = TbtProbeTable (iTb, side, indexChild);
                    }
                else
                    {
                    indexChild = pbd->IndCalc (iTb, xside,
                                               pbd->m_rgsqPieceList[x_colorWhite],
                                               pbd->m_rgsqPieceList[x_colorBlack],
                                               XX, false);
                    tbtScore = TbtProbeTable (iTb, xside, indexChild);
                    }
                if (bev_broken == tbtScore)
                    {
                    printf ("*** Invalid minor value [1] -- something wrong, index "HEX_INDEX_FORMAT"\n", indexChild);
                    pbd->VPrintPosition ();
                    fflush (stdout);
                    exit (1);
                    }
                pbd->VUnmakeMove (pmv);
                tbtScore = TbtReplaceResult (tbtScore);
                if (FBetter (tbtBest, tbtScore))
                    tbtBest = tbtScore;
                }
            else
                {
                if (2 != pbd->m_rgPieceCount[x_colorWhite]+pbd->m_rgPieceCount[x_colorBlack])
                    {
                    printf ("*** Unable to probe minor -- something wrong [TB=0]\n");
                    pbd->VPrintPosition ();
                    fflush (stdout);
                    exit (1);
                    }
                // KK, captured last enemy piece
                pbd->VUnmakeMove (pmv);
                *ptbt = bev_draw;
                return true;
                }
            }
        }
    if (-65536 == tbtBest)  // No legal moves
        *ptbt = (tb_t) (pbd->FAtacked (pbd->m_rgsqPieceList[side][0], xside) ? bev_li0 : bev_draw);
    else if (tbtBest != *ptbt)
        {
        *ptbt = (tb_t) tbtBest;
        fChanged = true;
        }
    return true;
    }

//-----------------------------------------------------------------------------
//
// Promote information from opposite table when everything is in memory

static bool L_FProbeOpposite
    (
    CBoard  *pbd,
    color   side,
    color   xside
    )
    {
    CMove   *pmv;
    CMove   *pmvEnd;
    BYTE    *ptbt;
    int     tbtBest;
    BYTE    *pbOpposite;
    INDEX   ind;
    INDEX   indMax;

    // Calculate index of current position
    ind = pbd->IndCalc (side);
    ind <<= 1;

    // Possible legal position?
    if (ind >= rgCb [side])
        {
        Assert (pbd->FAtacked (pbd->m_rgsqPieceList[xside][0], side));
        return false;
        }
    ptbt = rgpbFile[side] + ind;
    tbtBest = L_Get (ptbt);

    // Legal position?
    if (L_bev_broken == tbtBest)
        {
        Assert (pbd->FAtacked (pbd->m_rgsqPieceList[xside][0], side));
        return false;
        }
    Assert (!pbd->FAtacked (pbd->m_rgsqPieceList[xside][0], side));

    // Is it necessary to iterate?
#if 1
    if (tbtBest <= L_tbtLo || tbtBest > L_tbtHi)
        return true;
#endif

    // Generate all pseudolegal moves
    pmvEnd = pbd->PmvGenPseudoLegalMoves (pbd->m_rgmv, side);

    // Now iterate through all of them
    pbOpposite = rgpbFile[xside];
    indMax = rgCb [xside];
    tbtBest = -65536;
    for (pmv = pbd->m_rgmv; pmv < pmvEnd; pmv ++)
        {
        // If no capture/promotion, probe opposite
        // Otherwise, probe minor
        if (0 == pmv->m_ulFlags)
            {
            int tbtScore;

            pbd->VMinimalMakeQuietMove (pmv, side);

            // Probe opposite table
            ind = pbd->IndCalc (xside);
            ind <<= 1;
            
            // Possible legal move?
            if (ind < indMax)
                {
                tbtScore = L_Get (pbOpposite + ind);
                // Legal move?
                if (L_bev_broken != tbtScore)
                    {
                    tbtScore = L_TbtReplaceResult (tbtScore);
                    if (L_FBetter (tbtBest, tbtScore))
                        tbtBest = tbtScore;
                    }
                }
            pbd->VUnmakeMinimalQuietMove (pmv, side);
            }
        else
            {
            int iTb;

            pbd->VMakeMove (pmv);
            // Illegal move?
            if (pbd->FAtacked (pbd->m_rgsqPieceList[side][0], xside))
                {
                pbd->VUnmakeMove (pmv);
                continue;
                }
            // Probe minor table
            iTb = IDescFindFromCounters (pbd->m_rgiCount);
            if (0 != iTb)
                {
                INDEX indexChild;
                int tbtScore;

                if (iTb < 0)
                    {
                    iTb = -iTb;
                    indexChild = pbd->IndCalc (iTb, side,
                                               pbd->m_rgsqPieceList[x_colorBlack],
                                               pbd->m_rgsqPieceList[x_colorWhite],
                                               XX, true);
                    tbtScore = TbtProbeTable (iTb, side, indexChild);
                    }
                else
                    {
                    indexChild = pbd->IndCalc (iTb, xside,
                                               pbd->m_rgsqPieceList[x_colorWhite],
                                               pbd->m_rgsqPieceList[x_colorBlack],
                                               XX, false);
                    tbtScore = TbtProbeTable (iTb, xside, indexChild);
                    }
                if (bev_broken == tbtScore)
                    {
                    printf ("*** Invalid minor value [2] -- something wrong, index "HEX_INDEX_FORMAT"\n", indexChild);
                    pbd->VPrintPosition ();
                    fflush (stdout);
                    exit (1);
                    }
                pbd->VUnmakeMove (pmv);
                tbtScore = S_to_L (tbtScore);
                tbtScore = L_TbtReplaceResult (tbtScore);
                if (L_FBetter (tbtBest, tbtScore))
                    tbtBest = tbtScore;
                }
            else
                {
                if (2 != pbd->m_rgPieceCount[x_colorWhite]+pbd->m_rgPieceCount[x_colorBlack])
                    {
                    printf ("*** Unable to probe minor -- something wrong [%d]\n", iTb);
                    pbd->VPrintPosition ();
                    fflush (stdout);
                    exit (1);
                    }
                // KK, captured last enemy piece
                pbd->VUnmakeMove (pmv);
                ptbt[0] = ptbt[1] = 0;
                return true;
                }
            }
        }
    if (-65536 == tbtBest)  // No legal moves
        {
        tbtBest = pbd->FAtacked (pbd->m_rgsqPieceList[side][0], xside) ? L_bev_li0 : L_bev_draw;
        ptbt[0] = (BYTE) tbtBest;
        ptbt[1] = (BYTE) (tbtBest >> 8);
        }
    else if (tbtBest != L_Get (ptbt))
        {
        ptbt[0] = (BYTE) tbtBest;
        ptbt[1] = (BYTE) (tbtBest >> 8);
        fChanged = true;
        }
    return true;
    }

//-----------------------------------------------------------------------------
//
// Verify table

int iGlobalTb;

// Verify 8-bit table

static bool FVerify
    (
    CBoard  *pbd,
    color   side,
    color   xside
    )
    {
    CMove   *pmv;
    CMove   *pmvEnd;
    int     tbtTable;
    int     tbtBest;
    INDEX   indTable;
    INDEX   ind;
    bool    fWas_mimin;
    bool    fSymmetric = rgtbdDesc[iGlobalTb].m_fSymmetric;

    // Valid position?
    if (pbd->FAtacked (pbd->m_rgsqPieceList[xside][0], side))
        return false;

    // Calculate index of current position and probe table
    indTable = pbd->IndCalc (side);

    tbtTable = TbtProbeTable (iGlobalTb, fSymmetric ? x_colorWhite : side, indTable);

    // Generate all pseudolegal moves
    pmvEnd = pbd->PmvGenPseudoLegalMoves (pbd->m_rgmv, side);

    // Now iterate through all of them
    tbtBest = -65536;
    fWas_mimin = false;
    for (pmv = pbd->m_rgmv; pmv < pmvEnd; pmv ++)
        {
        pbd->VMakeMove (pmv);
        // Illegal move?
        if (pbd->FAtacked (pbd->m_rgsqPieceList[side][0], xside))
            {
            pbd->VUnmakeMove (pmv);
            continue;
            }

        tb_t tbtScore;

        // If capture/promotion, have to probe minor
        if (0 == pmv->m_ulFlags)
            {
            // Probe opposite table
            ind = pbd->IndCalc (xside);
            tbtScore = (tb_t) TbtProbeTable (iGlobalTb, fSymmetric ? x_colorWhite : xside, ind);
            if (bev_broken == tbtScore)
                VInvalid (pbd, "Broken opposite", side, indTable, bev_broken, bev_broken);
            }
        else
            {
            int iTb;

            // Probe minor table
            iTb = IDescFindFromCounters (pbd->m_rgiCount);
            if (0 != iTb)
                {
                if (iTb < 0)
                    {
                    INDEX indexChild;

                    iTb = -iTb;
                    indexChild = pbd->IndCalc (iTb, side,
                                               pbd->m_rgsqPieceList[x_colorBlack],
                                               pbd->m_rgsqPieceList[x_colorWhite],
                                               XX, true);
                    tbtScore = (tb_t) TbtProbeTable (iTb, side, indexChild);
                    if (bev_broken == tbtScore)
                        VInvalid (pbd, "Broken minor", side, indTable, bev_broken, bev_broken);
                    }
                else
                    {
                    INDEX indexChild;

                    indexChild = pbd->IndCalc (iTb, xside,
                                               pbd->m_rgsqPieceList[x_colorWhite],
                                               pbd->m_rgsqPieceList[x_colorBlack],
                                               XX, false);
                    tbtScore = (tb_t) TbtProbeTable (iTb, xside, indexChild);
                    if (bev_broken == tbtScore)
                        VInvalid (pbd, "Broken minor", side, indTable, bev_broken, bev_broken);
                    }
                }
            else
                {
                if (2 != pbd->m_rgPieceCount[x_colorWhite]+pbd->m_rgPieceCount[x_colorBlack])
                    {
                    printf ("*** XXX Unable to probe minor [%d]\n", iTb);
                    pbd->VPrintPosition ();
                    fflush (stdout);
                    exit (1);
                    }
                tbtScore = bev_draw;
                }
            }
        fWas_mimin |= (bev_mimin == tbtScore);
        tbtScore = TbtReplaceResult (tbtScore);
        if (FBetter (tbtBest, tbtScore))
            tbtBest = tbtScore;
        pbd->VUnmakeMove (pmv);
        }

    if (-65536 == tbtBest)  // No legal moves
        tbtBest = (tb_t) (pbd->FAtacked (pbd->m_rgsqPieceList[side][0], xside) ? bev_li0 : bev_draw);
    if (bev_limaxx == tbtBest && !fWas_mimin)
        VInvalid (pbd, "Score overflow", side, indTable, tbtBest, tbtTable);
    if (tbtBest != tbtTable)
        VInvalid (pbd, "Wrong result", side, indTable, tbtBest, tbtTable);
    return true;
    }

// Verify 16-bit table

static bool L_FVerify
    (
    CBoard  *pbd,
    color   side,
    color   xside
    )
    {
    CMove   *pmv;
    CMove   *pmvEnd;
    int     tbtTable;
    int     tbtBest;
    INDEX   indTable;
    INDEX   ind;

    // Valid position?
    if (pbd->FAtacked (pbd->m_rgsqPieceList[xside][0], side))
        return false;

    // Calculate index of current position and probe table
    indTable = pbd->IndCalc (side);
    tbtTable = L_TbtProbeTable (iGlobalTb, side, indTable);

    // Generate all pseudolegal moves
    pmvEnd = pbd->PmvGenPseudoLegalMoves (pbd->m_rgmv, side);

    // Now iterate through all of them
    tbtBest = -65536;
    for (pmv = pbd->m_rgmv; pmv < pmvEnd; pmv ++)
        {
        pbd->VMakeMove (pmv);
        // Illegal move?
        if (pbd->FAtacked (pbd->m_rgsqPieceList[side][0], xside))
            {
            pbd->VUnmakeMove (pmv);
            continue;
            }

        int tbtScore;

        // If capture/promotion, have to probe minor
        if (0 == pmv->m_ulFlags)
            {
            // Probe opposite table
            ind = pbd->IndCalc (xside);
            tbtScore = L_TbtProbeTable (iGlobalTb, xside, ind);
            if (L_bev_broken == tbtScore)
                VInvalid16 (pbd, "Broken opposite", side, indTable, L_bev_broken, L_bev_broken);
            }
        else
            {
            int iTb;

            // Probe minor table
            iTb = IDescFindFromCounters (pbd->m_rgiCount);
            if (0 != iTb)
                {
                if (iTb < 0)
                    {
                    INDEX indexChild;

                    iTb = -iTb;
                    indexChild = pbd->IndCalc (iTb, side,
                                               pbd->m_rgsqPieceList[x_colorBlack],
                                               pbd->m_rgsqPieceList[x_colorWhite],
                                               XX, true);
                    tbtScore = L_TbtProbeTable (iTb, side, indexChild);
                    if (L_bev_broken == tbtScore)
                        VInvalid16 (pbd, "Broken minor", side, indTable, L_bev_broken, L_bev_broken);
                    }
                else
                    {
                    INDEX indexChild;

                    indexChild = pbd->IndCalc (iTb, xside,
                                               pbd->m_rgsqPieceList[x_colorWhite],
                                               pbd->m_rgsqPieceList[x_colorBlack],
                                               XX, false);
                    tbtScore = L_TbtProbeTable (iTb, xside, indexChild);
                    if (L_bev_broken == tbtScore)
                        VInvalid16 (pbd, "Broken minor", side, indTable, L_bev_broken, L_bev_broken);
                    }
                }
            else
                {
                if (2 != pbd->m_rgPieceCount[x_colorWhite]+pbd->m_rgPieceCount[x_colorBlack])
                    {
                    printf ("*** YYY Unable to probe minor [%d]\n", iTb);
                    pbd->VPrintPosition ();
                    fflush (stdout);
                    exit (1);
                    }
                tbtScore = 0;
                }
            }
        tbtScore = L_TbtReplaceResult (tbtScore);
        if (L_FBetter (tbtBest, tbtScore))
            tbtBest = tbtScore;
        pbd->VUnmakeMove (pmv);
        }

    if (-65536 == tbtBest)  // No legal moves
        tbtBest = (pbd->FAtacked (pbd->m_rgsqPieceList[side][0], xside) ? L_bev_li0 : L_bev_draw);
    if (tbtBest != tbtTable)
        VInvalid16 (pbd, "Wrong result", side, indTable, tbtBest, tbtTable);
    return true;
    }

//-----------------------------------------------------------------------------

// Find position with the specified score in the table

static bool L_FFindScore
    (
    CBoard  *pbd,
    color   side,
    color   xside
    )
    {
    int     tbtTable;
    INDEX   indTable;

    // Valid position?
    if (pbd->FAtacked (pbd->m_rgsqPieceList[xside][0], side))
        return false;

    // Calculate index of current position and probe table
    indTable = pbd->IndCalc (side);
    tbtTable = L_TbtProbeTable (iGlobalTb, side, indTable);

    // Found score?
    if (tbtTable == tbtSearch)
        {
        pbd->VPrintFEN (side);
        pbd->m_cSearch ++;
        }
    return true;
    }

//-----------------------------------------------------------------------------

static void VCollectStatistics
    (
    BYTE    *pb,
    ULONG   cb
    )
    {
    tb_t *ptbt;

    ptbt = (tb_t*) pb;
    do
        {
        prgStatistics [128 + *ptbt] ++;
        ptbt ++;
        }
    while (-- cb);
    }

//-----------------------------------------------------------------------------

static void L_VCollectStatistics
    (
    BYTE    *pb,
    ULONG   cb
    )
    {
    do
        {
        prgStatistics [32768 + L_Get (pb)] ++;
        pb += 2;
        }
    while (cb -= 2);
    }

//-----------------------------------------------------------------------------
//
//      Service routines

static void VTab
    (
    int iLevel
    )
    {
    for (int i = 0; i < iLevel; i ++)
        printf ("  ");
    }

// Name looks reasonable?

bool FReasonableName
    (
    char    *pch
    )
    {
    bool    fWasKing;
    char    *pchTemp;

    if (strlen (pch) > MAX_TOTAL_PIECES)
        return false;
    for (pchTemp = pch; *pchTemp; pchTemp ++)
        *pchTemp = (char) tolower (*pchTemp);
    if ('k' != *pch)
        return false;
    fWasKing = false;
    for (pchTemp = pch+1; *pchTemp; pchTemp ++)
        {
        switch (*pchTemp)
            {
        case 'k':
            if (fWasKing)
                return false;
            else
                fWasKing = true;
            break;
        case 'p':
        case 'n':
        case 'b':
        case 'r':
        case 'q':
            break;
        default:
            return false;
            }
        }
    return fWasKing;
    }

// Print 8-bit statistics

static void VPrintStatistics
    (
    FILE    *fp,
    int     iLevel,
    const char    *szSide
    )
    {
    for (int i = 0; i < 256; i ++)
        {
        if (0 != prgStatistics [i])
            {
            tb_t    bev;

            if (stdout == fp)
                VTab (iLevel + 1);
            bev = (tb_t) (i - 128);
            if (bev_draw == bev)
                fprintf (fp, "%s: Draws:            " DEC_8_INDEX_FORMAT "\n",
                         szSide, prgStatistics [i]);
            else if (bev_broken == bev)
                fprintf (fp, "%s: Broken positions: " DEC_8_INDEX_FORMAT "\n",
                         szSide, prgStatistics [i]);
            else if (bev >= bev_mimin && bev <= bev_mi1)
                fprintf (fp, "%s: Mate in %3d:      " DEC_8_INDEX_FORMAT "\n",
                         szSide, bev_mi1 - bev + 1,
                                                            prgStatistics [i]);
            else if (bev >= bev_li0 && bev <= bev_limax)
                fprintf (fp, "%s: Lost in %3d:      " DEC_8_INDEX_FORMAT "\n",
                         szSide, bev_mi1 + bev, prgStatistics [i]);
            else if (bev_limaxx == bev)
                fprintf (fp, "%s: Lost in %3d:      " DEC_8_INDEX_FORMAT "\n",
                         szSide, 126, prgStatistics [i]);
            else if (bev_miminx == bev)
                fprintf (fp, "%s: Mate in %3d:      " DEC_8_INDEX_FORMAT "\n",
                         szSide, 127, prgStatistics [i]);
            else
                fprintf (fp, "%s: ?%3d:             " DEC_8_INDEX_FORMAT "\n",
                         szSide, bev, prgStatistics [i]);
            }
        }
    fflush (fp);
    }

// Print 16-bit statistics

static void L_VPrintStatistics
    (
    FILE    *fp,
    int     iLevel,
    const char    *szSide
    )
    {
    for (int i = 0; i < 65536; i ++)
        {
        if (0 != prgStatistics [i])
            {
            int bev;

            if (stdout == fp)
                VTab (iLevel + 1);
            bev = i - 32768;
            if (L_bev_draw == bev)
                fprintf (fp, "%s: Draws:            " DEC_8_INDEX_FORMAT "\n",
                         szSide, prgStatistics [i]);
            else if (L_bev_broken == bev)
                fprintf (fp, "%s: Broken positions: " DEC_8_INDEX_FORMAT "\n",
                         szSide, prgStatistics [i]);
            else if (bev >= L_bev_mimin && bev <= L_bev_mi1)
                fprintf (fp, "%s: Mate in %3d:      " DEC_8_INDEX_FORMAT "\n",
                         szSide, L_bev_mi1 - bev + 1, prgStatistics [i]);
            else if (bev >= L_bev_li0 && bev <= L_bev_limax)
                fprintf (fp, "%s: Lost in %3d:      " DEC_8_INDEX_FORMAT "\n",
                         szSide, L_bev_mi1 + bev, prgStatistics [i]);
            else
                fprintf (fp, "%s: ?%3d:             " DEC_8_INDEX_FORMAT "\n",
                         szSide, bev, prgStatistics [i]);
            }
        }
    fflush (fp);
    }

//-----------------------------------------------------------------------------
//
// Create pair of tablebases

static void VProcessTb (char *pch, int iLevel);

static void VBuildIfNecessary
    (
    char    *pch,
    int     iLevel,
    color   side
    )
    {
    int iTb;
    
    if (0 == strcmp (pch, "kk"))
        return;
    iTb = IDescFindByName (pch);
    if (iTb < 0)
        {
        iTb = -iTb;
        side = OtherSide (side);
        }
    if (0 == iTb ||
        NULL == rgtbdDesc[iTb].m_rgpchFileName[x_colorWhite] ||
        (!rgtbdDesc[iTb].m_fSymmetric && NULL == rgtbdDesc[iTb].m_rgpchFileName[x_colorBlack]))
        {
        VProcessTb (pch, iLevel+1);
        iTb = IDescFindByName (pch);
        if (iTb < 0)
            {
            iTb = -iTb;
            side = OtherSide (side);
            }
        }
    
    int fMapped;

    if (fMapFiles)
        {
        if (rgtbdDesc[iTb].m_fSymmetric)
            side = x_colorWhite;
#if defined (_WIN32) || defined(_WIN64)
        fMapped = FMapTableToMemory (iTb, side);
#else
        fMapped = FReadTableToMemory (iTb, side, NULL);
#endif
        if (!fMapped)
            {
            printf ("*** Unable to map %s", pch);
            fflush (stdout);
            exit (1);
            }
        }
    }

static void VRecurse
    (
    char    *pch,
    int     iLevel
    )
    {
    char    rgchRecursive[16];
    char    rgchCapPromote[16];
    char    *pchRecursive = rgchRecursive;
    char    *pchTemp;
    color   side = x_colorBlack;

    do
        {
        if ('k' == *pch)
            side = OtherSide (side);
        else
            {
            strcpy (pchRecursive, pch+1);
            VBuildIfNecessary (rgchRecursive, iLevel, side);
            if ('p' == *pch)
                {
                strcpy (pchRecursive+1, pch+1);
                for (int i = 0; i < 4; i ++)
                    {
                    int j;

                    // First, handle pawn promotion
                    *pchRecursive = "nbrq"[i];
                    VBuildIfNecessary (rgchRecursive, iLevel, OtherSide (side));
                    // Now, handle promotion with capture
                    if (NULL == strchr (pchRecursive, 'k'))
                        {
                        strcpy (rgchCapPromote, strrchr (rgchRecursive, 'k'));
                        j = strlen (rgchCapPromote);
                        pchTemp = & rgchRecursive [1];
                        }
                    else
                        {
                        rgchCapPromote[0] = 'k';
                        for (j = 1; 'k' != rgchRecursive[j]; j ++)
                            rgchCapPromote[j] = rgchRecursive[j];
                        pchTemp = & rgchRecursive[j+1];
                        }
                    rgchCapPromote[j] = 'k';
                    j ++;
                    while ('\0' != *pchTemp && 'k' != *pchTemp)
                        {
                        if ('p' != *pchTemp)
                            {
                            int k;

                            for (k = 0; 'k' != pchTemp[1+k] && '\0' != pchTemp[1+k]; k ++)
                                rgchCapPromote[j+k] = pchTemp[1+k];
                            rgchCapPromote[j+k] = '\0';
                            VBuildIfNecessary (rgchCapPromote, iLevel, x_colorBlack);
                            VBuildIfNecessary (rgchCapPromote, iLevel, x_colorWhite);
                            }
                        rgchCapPromote[j] = *pchTemp;
                        j ++;
                        pchTemp ++;
                        }
                    }
                }
            }
        * pchRecursive ++ = * pch ++;
        }
    while (*pch);
    }

static void VPrintElapsedTime
    (
    time_t tmStart
    )
    {
    time_t  tmFinish;
    ULONG   ulSeconds, ulMinutes, ulHours;

    tmFinish = time (NULL);
    ulSeconds = (ULONG) (tmFinish - tmStart);
    ulMinutes = ulSeconds / 60;
    ulSeconds %= 60;
    ulHours = ulMinutes / 60;
    ulMinutes %= 60; 
    if (0 != ulHours)
        printf (" %lu hour%s", ulHours, ulHours > 1 ? "s" : "");
    if (0 != ulMinutes)
        printf (" %lu minute%s", ulMinutes, ulMinutes > 1 ? "s" : "");
    if ((0 != ulSeconds) || (0 == ulHours + ulMinutes))
        printf (" %lu second%s", ulSeconds, ulSeconds > 1 ? "s" : "");
    }

// Set function pointers in each CBoard

static void VSetPfn
    (
    int     iTb,
    color   sd
    )
    {
    for (int cpu = 0; cpu < cCPUS; cpu ++)
        {
        rgBoards[cpu].m_rgpfnCalcIndex[sd] = PfnIndCalc(iTb, sd);
        rgBoards[cpu].m_rgpfnCalcIndex[OtherSide(sd)] = PfnIndCalc(iTb, OtherSide(sd));
        }
    }

// Initialize tables necessary for iterations through positions

void CBoard::VInitIterations
    (
    const char  *pchTable
    )
    {
    color   sd;

    VSetCounters (m_rgiCount, pchTable);
    m_rgPieceCount[x_colorWhite] = 1;
    m_rgPieceCount[x_colorBlack] = 1;
    m_rgPieces[x_colorWhite][0] =
    m_rgPieces[x_colorBlack][0] = x_pieceKing;
    for (sd = x_colorWhite; sd <= x_colorBlack; sd = (color) (sd + 1))
        {
        for (int iPieceType = 0; iPieceType < 5; iPieceType ++)
            {
            for (int i = 0; i < m_rgiCount[5 * sd + iPieceType]; i ++)
                m_rgPieces[sd][m_rgPieceCount[sd]++] = rgpiPiecesLinear[iPieceType];
            }
        }
    m_rgfHasPawns[x_colorWhite] = (0 != m_rgiCount[0]);
    m_rgfHasPawns[x_colorBlack] = (0 != m_rgiCount[5]);
    m_fBothSidesHavePawns = m_rgfHasPawns[x_colorWhite] && m_rgfHasPawns[x_colorBlack];
    for (sd = x_colorWhite; sd <= x_colorBlack; sd = (color) (sd + 1))
        {
        for (int i = 0; i < m_rgPieceCount[sd]; i ++)
            m_rgpsqNext[sd][i] = (x_piecePawn == m_rgPieces[sd][i]) ? rgsqNextSquarePawn : rgsqNextSquare;
        }
    // Now use symmetry
    m_rgpsqNext[x_colorWhite][0] = (m_rgfHasPawns[x_colorWhite] || m_rgfHasPawns[x_colorBlack]) ?
                                        rgsqNextSquareHalf : rgsqNextSquareTriangle;
    }

///////////////////////////////////////////////////////////////////////////////
//
// Bit vector routines for TB generation - made as template so that
// the same code could be used for both 8-bit and 16-bit TBs

#if defined (SMP)
    static lock_t   lockMR;
#endif

template <typename TB_T>
class BIT_algorithm
{
public:

    enum { Unknown = ((sizeof (TB_T) == 1) ? bev_unknown : L_bev_unknown) };
    enum { Li0 = ((sizeof (TB_T) == 1) ? bev_li0 : L_bev_li0) };
    enum { Mi1 = ((sizeof (TB_T) == 1) ? bev_mi1 : L_bev_mi1) };
    enum { Draw = 0 };
    enum { Broken = ((sizeof (TB_T) == 1) ? bev_broken : L_bev_broken) };

    static TB_T tbtLo;
    static TB_T tbtHi;

//-----------------------------------------------------------------------------
//  Unique functions that need specialization

    static INLINE TB_T TbtGet (BYTE *pb, size_t offset);
    static INLINE void VPutTbt (BYTE *pb, size_t offset, TB_T tbt);
    static void VFlushChunk (file fp, BYTE *pbBuf, INDEX indStart, size_t cPos);
    static size_t CposReadChunk (file fp, BYTE *pbBuf, INDEX indStart, size_t cPos);
    static TB_T TbtProbeMinor (int iTb, color side, INDEX indexChild);
    static TB_T TbtReplaceMinorResult (TB_T tbtScore);

//-----------------------------------------------------------------------------
//
//  Calculate # of positions - bit vector version

static bool BIT_FCalcMaxIndex
    (
    CBoard  *pbd,
    color   side,
    color   xside
    )
    {
    INDEX   ind;
    CMove   *pmv;
    CMove   *pmvEnd;

    // Valid position?
    if (pbd->FAtacked (pbd->m_rgsqPieceList[xside][0], side))
        return false;
    ind = pbd->IndCalc (side);

    if (ind > pbd->m_cb)
        pbd->m_cb = ind;
    // In theory, can print error message immediatelly,
    // but it's better to calculate how many memory necessary
    if (ind >= indBVSize)
        return true;

    // Now check for "lost in 0" and stalemates
    // Generate all pseudolegal moves
    // UNDONE: first generate only king moves, as usually king can escape
    pmvEnd = pbd->PmvGenPseudoLegalMoves (pbd->m_rgmv, side);
    // Now iterate through all of them
    for (pmv = pbd->m_rgmv; pmv < pmvEnd; pmv ++)
        {
        pbd->VMakeMove (pmv);
        // Illegal move?
        if (! pbd->FAtacked (pbd->m_rgsqPieceList[side][0], xside))
            {
            pbd->VUnmakeMove (pmv);
            // Mark position as valid
            bvNew.VSetBit (ind);
            return true;
            }
        pbd->VUnmakeMove (pmv);
        }
    // Mark position as "lost on 0" or "stalemate"
    bvArg.VSetBit (ind);
    if (!pbd->FAtacked (pbd->m_rgsqPieceList[side][0], xside))
        // Stalemate
        bvNew.VSetBit (ind);
    return true;
    }

//-----------------------------------------------------------------------------
//
//  Create file on disk - bit vector version

static void BIT_VWriteFile
    (
    file    fp,         // File to write
    BYTE    *pbBuf,     // What buffer to use
    size_t  cbBuf,      // Length of the buffer
    CBitVector bvArg,
    CBitVector bvNew,
    INDEX   cInd        // max index + 1
    )
    {
    INDEX   ind;
    INDEX   indLo;
    INDEX   indHi;
    size_t  cPos;

    ind = 0;
    do
        {
        cPos = cbBuf/(sizeof(TB_T) == 1 ? 1 : 2);
        if (cPos > (cInd - ind))
            cPos = (size_t) (cInd - ind);
        indLo = ind;
        indHi = ind + cPos;
        do
            {
            if (bvArg.FGetBit (ind))
    //            VPutTbt (pbBuf, (size_t) (ind-indLo), (bvNew.FGetBit (ind) ? Draw: Li0));
  VPutTbt (pbBuf, (size_t) (ind-indLo), (bvNew.FGetBit (ind) ? 0:((sizeof (TB_T) == 1) ? bev_li0 : L_bev_li0)));      
      else
   //          VPutTbt (pbBuf, (size_t) (ind-indLo), (bvNew.FGetBit (ind) ? Unknown : Broken));
          VPutTbt (pbBuf, (size_t) (ind-indLo), (bvNew.FGetBit (ind) ? ((sizeof (TB_T) == 1) ? bev_unknown : L_bev_unknown) : ((sizeof (TB_T) == 1) ? bev_broken : L_bev_broken) ));
         ind ++;
            }
        while (ind < indHi);
        VFlushChunk (fp, pbBuf, indLo, cPos);
        }
    while (ind < cInd);
    }

//-----------------------------------------------------------------------------
//
//  Write entire file to disk - bit vector version
//  Return true if something changed

static bool BIT_FWriteFileWin
    (
    file    fp,     // File to write
    BYTE    *pbBuf, // What buffer to use
    size_t  cbBuf,  // Length of the buffer
    TB_T    tbtVal, // Which value to write for each 1 in the bit vector
    CBitVector bv,  // Bit vector itself
    INDEX   cInd    // max index + 1
    )
    {
    INDEX   ind;
    INDEX   indLo;
    INDEX   indHi;
    bool    fChanged;
    size_t  cPos;
    TB_T    tbt;

    cPos = 0;
    indLo = 0;
    indHi = 0;
    fChanged = false;
    for (ind = 0; ind < cInd; ind ++)
        {
        if (bv.FGetBit (ind))
            {
            if (! (ind >= indLo && ind < indHi))
                {
                if (fChanged)
                    VFlushChunk (fp, pbBuf, indLo, cPos);
                cPos = cbBuf/(sizeof(TB_T) == 1 ? 1 : 2);
                if (cPos > cInd - ind)
                    cPos = (size_t) (cInd - ind);
                cPos = CposReadChunk (fp, pbBuf, ind, cPos);
                indLo = ind;
                indHi = ind + cPos;
                }
            tbt = TbtGet (pbBuf, (size_t) (ind - indLo));
            if (Unknown == tbt || tbt < tbtVal)
                {
                VPutTbt (pbBuf, (size_t) (ind - indLo), tbtVal);
                fChanged = true;
                }
            }
        }
    if (fChanged)
        VFlushChunk (fp, pbBuf, indLo, cPos);
    return fChanged;
    }

//-----------------------------------------------------------------------------
//
//  Write entire file to disk - bit vector version
//  Return true if something changed
//  Differs from the "standard" version, as it have to handle broken positions

static bool BIT_FWriteFileReverseWin
    (
    file    fp,     // File to write
    BYTE    *pbBuf, // What buffer to use
    size_t  cbBuf,  // Length of the buffer
    TB_T    tbtVal, // Which value to write for each 1 in the bit vector
    CBitVector bv,  // Bit vector itself
    INDEX   cInd    // max index + 1
    )
    {
    INDEX   ind;
    INDEX   indLo;
    INDEX   indHi;
    bool    fChanged;
    size_t  cPos;
    TB_T    tbt;

    cPos = 0;
    indLo = 0;
    indHi = 0;
    fChanged = false;
    for (ind = 0; ind < cInd; ind ++)
        {
        if (bv.FGetBit (ind))
            {
            if (! (ind >= indLo && ind < indHi))
                {
                if (fChanged)
                    VFlushChunk (fp, pbBuf, indLo, cPos);
                cPos = cbBuf/(sizeof(TB_T) == 1 ? 1 : 2);
                if (cPos > cInd - ind)
                    cPos = (size_t) (cInd - ind);
                cPos = CposReadChunk (fp, pbBuf, ind, cPos);
                indLo = ind;
                indHi = ind + cPos;
                }
            tbt = TbtGet (pbBuf, (size_t) (ind - indLo));
            if (Unknown == tbt || (Broken != tbt && tbt < tbtVal))
                {
                VPutTbt (pbBuf, (size_t) (ind - indLo), tbtVal);
                fChanged = true;
                }
            }
        }
    if (fChanged)
        VFlushChunk (fp, pbBuf, indLo, cPos);
    return fChanged;
    }

//-----------------------------------------------------------------------------
//
//  Write entire file to disk - bit vector version
//  Return true if something changed

static bool BIT_FWriteFileLost
    (
    file    fp, // File to write
    BYTE    *pbBuf, // What buffer to use
    size_t  cbBuf,  // Length of the buffer
    TB_T    tbtVal, // Which value to write for each 1 in the bit vector
    CBitVector bv,  // Bit vector itself
    INDEX   cInd    // max index + 1
    )
    {
    INDEX   ind;
    INDEX   indLo;
    INDEX   indHi;
    bool    fChanged;
    size_t  cPos;
    TB_T    tbt;

    indLo = 0;
    indHi = 0;
    fChanged = false;
    cPos = cbBuf/(sizeof(TB_T) == 1 ? 1 : 2);
    if (cPos > cInd)
        cPos = (size_t) cInd;
    cPos = CposReadChunk (fp, pbBuf, 0, cPos);
    indLo = 0;
    indHi = cPos;
    ind = 0;
    for (;;)
        {
        tbt = TbtGet (pbBuf, (size_t) (ind - indLo));
        if (bv.FGetBit (ind))
            {
            if (Unknown == tbt)
                {
                VPutTbt (pbBuf, (size_t) (ind - indLo), tbtVal);
                fChanged = true;
                }
            }
        else if (tbt == tbtVal)
            {
            VPutTbt (pbBuf, (size_t) (ind - indLo), Unknown);
            fChanged = true;
            }
        ind ++;
        if (ind == indHi)
            {
            if (fChanged)
                VFlushChunk (fp, pbBuf, indLo, cPos);
            if (ind == cInd)
                break;
            cPos = cbBuf/(sizeof(TB_T) == 1 ? 1 : 2);
            if (cPos > cInd - ind)
                cPos = (size_t) (cInd - ind);
            cPos = CposReadChunk (fp, pbBuf, ind, cPos);
            indLo = ind;
            indHi = ind + cPos;
            }
        }
    return fChanged;
    }

//-----------------------------------------------------------------------------
//
// Replace 'unknown' by 'draw' in the file

static void BIT_VReplaceUnknowns
    (
    file    fp, // File to write
    BYTE    *pbBuf, // What buffer to use
    size_t  cbBuf   // Length of the buffer
    )
    {
    INDEX   ind;
    size_t  cPos;

    ind = 0;
    for (;;)
        {
        cPos = CposReadChunk (fp, pbBuf, ind, cbBuf/(sizeof(TB_T) == 1 ? 1 : 2));
        if (0 == cPos)
            return;
        for (size_t i = 0; i < cPos; i ++)
            {
            if (Unknown == TbtGet (pbBuf, i))
                VPutTbt (pbBuf, i, Draw);
            }
        if (1 == sizeof (TB_T))
            VCollectStatistics (pbBuf, cPos);
        else
            L_VCollectStatistics (pbBuf, cPos*2);
        VFlushChunk (fp, pbBuf, ind, cPos);
        ind += cPos;
        }
    }

//-----------------------------------------------------------------------------
//
// Form bit vector in memory

// Predicates for different passes

typedef bool (* PfnPredicate) (TB_T tbt);

static INLINE bool FUnknownValue
    (
    TB_T    tbt
    )
    {
    return Unknown == tbt;
    }

static INLINE bool FLoValue
    (
    TB_T    tbt
    )
    {
    return (tbt == tbtLo);
    }

static INLINE bool FGeHiOrBrokenValue
    (
    TB_T    tbt
    )
    {
    return (Unknown != tbt) && (Broken == tbt || tbt >= tbtHi);
    }

static INLINE bool FUnstabilizedValueForWin
    (
    TB_T    tbt
    )
    {
    return (Unknown == tbt) || (tbt < 0 && tbt >= tbtLo) || (tbt >= 0 && tbt < tbtHi);
    }

static INLINE bool FUnstabilizedValueForLost
    (
    TB_T    tbt
    )
    {
    return (Unknown == tbt) || (tbt == tbtLo);
    }

// Function that will go through the file and form bit vector

class CFormBV
    {
public:
    static void BIT_VFormBV
        (
        PfnPredicate pfn,
        file      fp,       // File to read from
        BYTE      *pbBuf,   // What buffer to use
        size_t    cbBuf,    // Length of the buffer
        CBitVector bv,      // Bit vector itself
        INDEX     cInd      // max index + 1
        )
        {
        INDEX   ind;
        INDEX   indLo;
        INDEX   indHi;
        size_t  cPos;

        bv.VZero(cInd);
        cPos = CposReadChunk (fp, pbBuf, 0, cbBuf/(sizeof(TB_T) == 1 ? 1 : 2));
        indLo = 0;
        indHi = cPos;
        ind = 0;
        for (;;)
            {
            if (pfn (TbtGet (pbBuf, (size_t) (ind - indLo))))
                bv.VSetBitUnlocked (ind);
            ind ++;
            if (ind == indHi)
                {
                if (ind == cInd)
                    break;
                cPos = CposReadChunk (fp, pbBuf, ind, cbBuf/(sizeof(TB_T) == 1 ? 1 : 2));
                indLo = ind;
                indHi = ind + cPos;
                }
            }
        }
    };

//-----------------------------------------------------------------------------
//
// Promote information from opposite table - bit vector vesrion; search for win

static bool BIT_FSearchWin
    (
    CBoard  *pbd,
    color   side,
    color   xside
    )
    {
    CMove   *pmv;
    CMove   *pmvEnd;
    INDEX   ind;
    INDEX   indChild;
    INDEX   indMax;

    // Calculate index of current position
    ind = pbd->IndCalc (side);

    // Possible legal position?
    if (ind >= rgCb[side])
        return false;
    // Candidate position?
    if (!bvNew.FGetBit (ind))
        return true;
    Assert (!pbd->FAtacked (pbd->m_rgsqPieceList[xside][0], side));

    // Generate all pseudolegal moves
    pmvEnd = pbd->PmvGenPseudoLegalMoves (pbd->m_rgmv, side);

    // Now iterate through all of them
    indMax = rgCb [xside];
    for (pmv = pbd->m_rgmv; pmv < pmvEnd; pmv ++)
        {
        // Does not have to probe minors
        if (0 == pmv->m_ulFlags)
            {
            pbd->VMinimalMakeQuietMove (pmv, side);
            indChild = pbd->IndCalc (xside);
            pbd->VUnmakeMinimalQuietMove (pmv, side);
            
            // Legal move? Resulting position apply?
            if (indChild < indMax && bvArg.FGetBit (indChild))
                return true;
            }
        }
    bvNew.VClearBit (ind);
    return true;
    }

//-----------------------------------------------------------------------------
//
// Promote information from opposite table - bit vector vesrion; search for win
// "Reverse" version

static bool BIT_FReverseSearchWin
    (
    CBoard  *pbd,
    color   side,
    color   xside
    )
    {
    CMove   *pmv;
    CMove   *pmvEnd;
    INDEX   ind;
    INDEX   indChild;
    INDEX   indMax;
    int     cPieces;

    // Calculate index of current position
    ind = pbd->IndCalc (side);

    // Possible legal position?
    if (ind >= rgCb[side])
        return false;
    // Candidate position?
    if (!bvArg.FGetBit (ind))
        return true;
    Assert (!pbd->FAtacked (pbd->m_rgsqPieceList[xside][0], side));

    // Generate all pseudolegal moves and iterate through all of them
    // UNDONE: Implement unmoves for pawns!
    cPieces = pbd->m_rgPieceCount[xside];
    indMax = rgCb [xside];
    for (int iIndexFrom = 0; iIndexFrom < cPieces; iIndexFrom++)
        {
        pmvEnd = pbd->PmvGenPseudoLegalQuietUnMovesForPiece (pbd->m_rgmv, iIndexFrom, xside, side);
        for (pmv = pbd->m_rgmv; pmv < pmvEnd; pmv ++)
            {
            // Does not have to probe minors
            pbd->VMinimalMakeQuietMove (pmv, xside);
            indChild = pbd->IndCalc (xside);
            pbd->VUnmakeMinimalQuietMove (pmv, xside);
            
            // (Possible) legal move?
            if (indChild < indMax)
                bvNew.VSetBit (indChild);
            }
        }
    return true;
    }

//-----------------------------------------------------------------------------
//
// Promote information from opposite table - bit vector vesrion; search for win
// "Reverse" version when e.p. is possible

static bool BIT_FReverseSearchWinEP ( CBoard *pbd, color side, color xside )
{
    CMove   *pmv;
    CMove   *pmvEnd;
    INDEX   ind;
    INDEX   indChild;
    INDEX   indMax;
    int cPieces;

    // Calculate index of current position
    ind = pbd->IndCalc (side);

    // Possible legal position?
    if (ind >= rgCb[side])
        return false;
    // Candidate position?
    if (!bvArg.FGetBit (ind))
        return true;

    Assert (!pbd->FAtacked (pbd->m_rgsqPieceList[xside][0], side));

    // Generate all pseudolegal moves and iterate through all of them

    cPieces = pbd->m_rgPieceCount[xside];
    indMax = rgCb [xside];
    for (int iIndexFrom = 0; iIndexFrom < cPieces; iIndexFrom++) {
        // if e.p. possible, previous move must have been double pawn move
        if(pbd->m_sqEnPassant != XX) {
            square sqTo = pbd->m_rgsqPieceList[xside][iIndexFrom];
            if(sqTo != pbd->m_sqEnPassant + rgiPawnStep[xside])
                continue;
            assert(pbd->m_rgBoard[sqTo] == x_piecePawn);
            VAddQuietMove(pbd->m_rgmv, sqTo, sqTo - 2*rgiPawnStep[xside], iIndexFrom);
            pmvEnd = pbd->m_rgmv + 1;
        }
        else
            pmvEnd = pbd->PmvGenPseudoLegalQuietUnMovesForPiece (pbd->m_rgmv, iIndexFrom, xside, side);
        for (pmv = pbd->m_rgmv; pmv < pmvEnd; pmv ++) {
            // Does not have to probe minors
            pbd->VMinimalMakeQuietUnMove (pmv, xside);

            // Eliminate unmoves that lead to forward moves having
            // a different e.p. status than what we started with
            if(pbd->m_sqEnPassant != pmv->m_sqSaveEnPassant) {
                pbd->VUnmakeMinimalQuietMove (pmv, xside);
                continue;
            }

            // base position after unmove has no e.p.

            pbd->m_sqEnPassant = XX;

            indChild = pbd->IndCalc (xside);
            // (Possible) legal move?

            if (indChild < indMax) {
                bvNew.VSetBit (indChild);

                pbd->m_rgColor[pmv->m_sqTo] = xside;
                pbd->m_rgColor[pmv->m_sqFrom] = x_colorNeutral;
                pbd->m_rgBoard[pmv->m_sqFrom] = x_pieceNone;

                // if position is winning without e.p., it is winning with e.p.
                for (int i = pbd->m_rgPieceCount[side]-1; i > 0; i --) {
                    square sq = pbd->m_rgsqPieceList[side][i];
                    if ((rgiRowEnPassant[side] == Row(sq)) &&
                        (x_piecePawn == pbd->m_rgBoard[sq]) &&
                        (x_pieceNone == pbd->m_rgBoard[sq-rgiPawnStep[side]]) &&
                        (x_pieceNone == pbd->m_rgBoard[sq-rgiPawnStep[side]*2]) &&
                        ((0 != Column(sq) && x_piecePawn == pbd->m_rgBoard[sq-1] && xside == pbd->m_rgColor[sq-1]) ||
                         (7 != Column(sq) && x_piecePawn == pbd->m_rgBoard[sq+1] && xside == pbd->m_rgColor[sq+1])))
                    {
                        pbd->m_sqEnPassant = sq-rgiPawnStep[side];
                        indChild = pbd->IndCalc (xside);
                        if(indChild < indMax)
                            bvNew.VSetBit (indChild);
                        pbd->m_sqEnPassant = XX;
                    }
                }

                pbd->m_rgColor[pmv->m_sqTo] = x_colorNeutral;
                pbd->m_rgColor[pmv->m_sqFrom] = xside;
                pbd->m_rgBoard[pmv->m_sqFrom] = pbd->m_rgBoard[pmv->m_sqTo];
            }
            pbd->VUnmakeMinimalQuietMove (pmv, xside);
        }
    }
    return true;
}

//-----------------------------------------------------------------------------
//
// Promote information from opposite table - bit vector vesrion; search for loss

static bool BIT_FSearchLost
    (
    CBoard  *pbd,
    color   side,
    color   xside
    )
    {
    CMove   *pmv;
    CMove   *pmvEnd;
    INDEX   ind;
    INDEX   indChild;
    INDEX   indMax;
    int     cPieces;

    // Calculate index of current position
    ind = pbd->IndCalc (side);

    // Possible legal position?
    if (ind >= rgCb[side])
        return false;
    //  Candidate position?
    if (!bvNew.FGetBit (ind))
        return true;
    Assert (!pbd->FAtacked (pbd->m_rgsqPieceList[xside][0], side));

    // Generate all pseudolegal moves and iterate through all of them
    cPieces = pbd->m_rgPieceCount[side];
    indMax = rgCb [xside];
    for (int iIndexFrom = 0; iIndexFrom < cPieces; iIndexFrom++)
        {
        pmvEnd = pbd->PmvGenPseudoLegalQuietMovesForPiece (pbd->m_rgmv, iIndexFrom, side, xside);
        for (pmv = pbd->m_rgmv; pmv < pmvEnd; pmv ++)
            {
            pbd->VMinimalMakeQuietMove (pmv, side);
            indChild = pbd->IndCalc (xside);
            pbd->VUnmakeMinimalQuietMove (pmv, side);

            // Legal move? Resulting position apply?
            if (indChild < indMax && !bvArg.FGetBit (indChild))
                {
                bvNew.VClearBit (ind);
                return true;
                }
            }
        }
    // We found it!
    return true;
    }

//-----------------------------------------------------------------------------
//
// Write results of probing minor tables to disk

static int TB_CDECL ICompare
    (
    const void *arg1,
    const void *arg2
    )
    {
    const CMinorResult  *pmr1 = reinterpret_cast<const CMinorResult *> (arg1);
    const CMinorResult  *pmr2 = reinterpret_cast<const CMinorResult *> (arg2);

    if (pmr1->m_ind < pmr2->m_ind)
        return -1;
    else if (pmr1->m_ind == pmr2->m_ind)
        return 0;
    else
        return 1;
    }

static void BIT_VWriteMinorResults
    (
    CBoard  *pbd
    )
    {
    const CMinorResult  *pmr;
    INDEX               indMax;
    INDEX               indCur;
    INDEX               indChunk;
    size_t              cPos;
    BYTE                *pb;

#if defined (PRINT)
    printf ("DEBUG: writing %d results\n", pbd->m_cmr);
    fflush (stdout);
#endif
    if (0 == pbd->m_cmr)
        return;
    qsort (pbd->m_pmr, pbd->m_cmr, sizeof (CMinorResult), ICompare);
#if defined (PRINT)
    printf ("DEBUG: Sorted\n", pbd->m_cmr);
    fflush (stdout);
#endif
    pmr = pbd->m_pmr;
    indMax = pbd->m_pmr[pbd->m_cmr-1].m_ind;
    Lock (lockMR);
    pmr --;
    do
        {
        pmr ++;
        indCur = pmr->m_ind;
        cPos = pbd->m_cbBuf/(sizeof(TB_T) == 1 ? 1 : 2);
        if (cPos > indMax - indCur + 1)
            cPos = (size_t) (indMax - indCur + 1);
        cPos = CposReadChunk (pbd->m_fp, pbd->m_pbBuf, indCur, cPos);
        indChunk = indCur + (INDEX) cPos;
        pb = pbd->m_pbBuf;
        for (;;)
            {
            VPutTbt (pb, 0, (TB_T) pmr->m_tbt);
            if (pmr->m_ind == indMax || pmr[1].m_ind >= indChunk)   // UNDONE: guard record?
                break;
            pb += (pmr[1].m_ind - pmr->m_ind)*(sizeof(TB_T) == 1 ? 1 : 2);
            pmr ++;
            }
        VFlushChunk (pbd->m_fp, pbd->m_pbBuf, indCur, (size_t) (pmr->m_ind - indCur + 1));
        }
    while (pmr->m_ind != indMax);
    UnLock (lockMR);
    pbd->m_cmr = 0;
#if defined (PRINT)
    printf ("DEBUG: Wrote results\n", pbd->m_cmr);
    fflush (stdout);
#endif
    }

//-----------------------------------------------------------------------------
//
// Probe minor tables (bit vector approach)

static bool BIT_FProbeMinors
    (
    CBoard  *pbd,
    color   side,
    color   xside
    )
    {
    CMove   *pmv;
    CMove   *pmvEnd;
    int     tbtBest;
    INDEX   ind;

    // Calculate index of current position
    ind = pbd->IndCalc (side);

    // Possible legal position? Candidate position?
    if (ind >= rgCb[side] || !bvArg.FGetBit (ind))
        return true;    // EnPassant position cam be legal
    Assert (!pbd->FAtacked (pbd->m_rgsqPieceList[xside][0], side));

    // Generate all pseudolegal moves
    pmvEnd = pbd->PmvGenPseudoLegalMoves (pbd->m_rgmv, side);

    tbtBest = -65536;
    // Now iterate through all of them
    for (pmv = pbd->m_rgmv; pmv < pmvEnd; pmv ++)
        {
        // Bother only with minor probing
                if (0 != pmv->m_ulFlags)
            {
            int iTb;

            pbd->VMakeMove (pmv);
            // Illegal move?
            if (pbd->FAtacked (pbd->m_rgsqPieceList[side][0], xside))
                {
                pbd->VUnmakeMove (pmv);
                continue;
                }
            // Probe minor table
            iTb = IDescFindFromCounters (pbd->m_rgiCount);
            if (0 != iTb)
                {
                INDEX indexChild;
                int tbtScore;

                if (iTb < 0)
                    {
                    iTb = -iTb;
                    indexChild = pbd->IndCalc (iTb, side,
                                               pbd->m_rgsqPieceList[x_colorBlack],
                                               pbd->m_rgsqPieceList[x_colorWhite],
                                               XX, true);
                    tbtScore = TbtProbeMinor (iTb, side, indexChild);
                    }
                else
                    {
                    indexChild = pbd->IndCalc (iTb, xside,
                                               pbd->m_rgsqPieceList[x_colorWhite],
                                               pbd->m_rgsqPieceList[x_colorBlack],
                                               XX, false);
                    tbtScore = TbtProbeMinor (iTb, xside, indexChild);
                    }
                if (Broken == tbtScore)
                    {
                    printf ("*** Invalid minor value [3] -- something wrong, TB=%d index "HEX_INDEX_FORMAT"\n", iTb, indexChild);
                    pbd->VPrintPosition ();
                    fflush (stdout);
                    exit (1);
                    }
                pbd->VUnmakeMove (pmv);

                tbtScore = TbtReplaceMinorResult (tbtScore);
                if (tbtBest < tbtScore)
                    tbtBest = tbtScore;
                }
            else
                {
                if (2 != pbd->m_rgPieceCount[x_colorWhite]+pbd->m_rgPieceCount[x_colorBlack])
                    {
                    printf ("*** Unable to probe minor -- something wrong [%d]\n", iTb);
                    pbd->VPrintPosition ();
                    fflush (stdout);
                    exit (1);
                    }
                // KK, captured last enemy piece
                pbd->VUnmakeMove (pmv);
                tbtBest = Draw;
                break;
                }
            }
        }
    if (-65536 != tbtBest)  // There was promotion/capture
        {
        CMinorResult *pmr;

        if (pbd->m_cmr == pbd->m_cmrMax)
            {
            BIT_VWriteMinorResults (pbd);
            VCheckGuards();
            }
#if !defined (NDEBUG)
        if (0x12345678 != pbd->m_uGuard)
            {
            printf ("*** Illegal guard value: %X\n", pbd->m_uGuard);
            fflush(stdout);
            *(int*)0=0;
            }
#endif
        pmr = pbd->m_pmr + pbd->m_cmr;
        pmr->m_ind = ind;
        pmr->m_tbt = tbtBest;
        pbd->m_cmr ++;
        // Set longest minor result in CBoard, if necessary
        if (Draw != tbtBest)
            {
            int cMoves;

            if (1 == sizeof (TB_T)) // 8-bit table?
                {
                if (tbtBest >= bev_mimin && tbtBest <= bev_mi1)
                    {
                    cMoves = bev_mi1 - tbtBest + 1;
                    if (cMoves > pbd->m_cWinMax)
                        pbd->m_cWinMax = cMoves;
                    }
                else if (tbtBest >= bev_li0 && tbtBest <= bev_limax)
                    {
                    cMoves = bev_mi1 + tbtBest;
                    if (cMoves > pbd->m_cLostMax)
                        pbd->m_cLostMax = cMoves;
                    }
                else if (bev_limaxx == tbtBest)
                    {
                    cMoves = 126;
                    if (cMoves > pbd->m_cLostMax)
                        pbd->m_cLostMax = cMoves;
                    }
                else if (bev_miminx == tbtBest)
                    {
                    cMoves = 127;
                    if (cMoves > pbd->m_cWinMax)
                        pbd->m_cWinMax = cMoves;
                    }
                else
                    {
                    printf ("*** Invalid minor value [4] -- something wrong %04X\n", tbtBest);
                    pbd->VPrintPosition ();
                    fflush (stdout);
                    exit (1);
                    }
                }
            else    // 16-bit table
                {
                if (tbtBest >= L_bev_mimin && tbtBest <= L_bev_mi1)
                    {
                    cMoves = L_bev_mi1 - tbtBest + 1;
                    if (cMoves > 256)
                        {
                        printf ("*** Invalid cMoves value %d -- something wrong\n", cMoves);
                        pbd->VPrintPosition ();
                        fflush (stdout);
                        exit (1);
                        }
                    if (cMoves > pbd->m_cWinMax)
                        pbd->m_cWinMax = cMoves;
                    }
                else if (tbtBest >= L_bev_li0 && tbtBest <= L_bev_limax)
                    {
                    cMoves = L_bev_mi1 + tbtBest;
                    if (cMoves > 256)
                        {
                        printf ("*** Invalid cMoves value %d -- something wrong\n", cMoves);
                        pbd->VPrintPosition ();
                        fflush (stdout);
                        exit (1);
                        }
                    if (cMoves > pbd->m_cLostMax)
                        pbd->m_cLostMax = cMoves;
                    }
                else
                    {
                    printf ("*** Invalid minor value [5] -- something wrong %04X\n", tbtBest);
                    pbd->VPrintPosition ();
                    fflush (stdout);
                    exit (1);
                    }
                }
            }
        }
    return true;
    }

// Read Write information that is necessary for resume

static void VWriteStatus
    (
    char    *szFile,
    bool    *pfChanged,
    int     *pWinMax,
    int     *pLostMax,
    int     *pcIterations,
    int     *piIteration
    )
    {
    FILE    *fp;

    fp = fopen (szFile, "wt+");
    if (NULL == fp)
        {
        printf ("*** Unable to read status file\n");
        exit (1);
        }
    fprintf (fp,
             "%i %i %i %i %i %i %i %i \n",
             *pcIterations,
             *piIteration,
             pfChanged[0],
             pfChanged[1],
             pWinMax[0],
             pWinMax[1],
             pLostMax[0],
             pLostMax[1]);
    if (0 != fclose (fp))
        {
        printf ("*** Unable to close status file\n");
        exit (1);
        }
    }

// Read information that is necessary for resume

static void VReadStatus
    (
    char    *szFile,
    bool    *pfChanged,
    int     *pWinMax,
    int     *pLostMax,
    int     *pcIterations,
    int     *piIteration
    )
    {
    int rgiChanged[2];

    FILE    *fp;

    fp = fopen (szFile, "rt");
    if (NULL == fp)
        {
        printf ("*** Unable to read status file\n");
        exit (1);
        }
    fscanf (fp,
             "%i %i %i %i %i %i %i %i \n",
             pcIterations,
             piIteration,
             &rgiChanged[0],
             &rgiChanged[1],
             &pWinMax[0],
             &pWinMax[1],
             &pLostMax[0],
             &pLostMax[1]);
    pfChanged[0] = (bool) rgiChanged[0];
    pfChanged[1] = (bool) rgiChanged[1];
    if (0 != fclose (fp))
        {
        printf ("*** Unable to close status file\n");
        exit (1);
        }
    }

// Create TB - bit vector approach

static void BIT_VCreateTable
    (
    int iTb,
    int iLevel
    )
    {
    char    *pchTable;
    time_t  tmStart;
    color   sd;
    color   sdLimit;
    int     iIteration;
    int     cIterations;
    int     cpu;
    char    rgchTbName[2][16];
    char    rgchTempName[2][16];
    bool    fSymmetric = rgtbdDesc[iTb].m_fSymmetric;
    int     rgcWinMax[2];
    int     rgcLostMax[2];
    bool    rgfChanged[2];
    bool    rgfChangedNew[2];

    pchTable = rgtbdDesc[iTb].m_rgchName;
    VTab (iLevel);
    printf (!fResume ? "Starting %s\n" : "Resuming %s\n", pchTable);
    fflush (stdout);

    // Check minor tables
    if (!fResume)
        {
        VRecurse (pchTable, iLevel);
        VTab (iLevel);
        printf ("  All minor for %s exists, generating it\n", pchTable);
        if (rgtbdDesc[iTb].m_f16bit)
            {
            VTab (iLevel);
            printf ("  %s is 16-bit table\n", pchTable);
            }
        fflush (stdout);
        }
    
    // Initialization
    tmStart = time (NULL);
    for (cpu = 0; cpu < cCPUS; cpu ++)
        {
        rgBoards[cpu].m_fReverse = false;
        rgBoards[cpu].VInitIterations (pchTable);
        }
    sdLimit = x_colorBlack;
    if (fSymmetric)
        sdLimit = x_colorWhite;

    cbv = cbMemoryUsed/2 / sizeof (ULONG);
    indBVSize = ((INDEX)(cbMemoryUsed/2)) * 8;

    if (fResume)
        {
        rgCb[x_colorWhite] = rgtbdDesc[iTb].m_rgcbLength[x_colorWhite];
        rgCb[x_colorBlack] = rgtbdDesc[iTb].m_rgcbLength[x_colorBlack];
        for (sd = x_colorWhite; sd <= sdLimit; sd = (color) (sd + 1))
            {
            const char  *pchExt = PchExt (sd);
            char        rgszBuf[1024];

            VTab (iLevel);
            if (0 == strcmp ("-", pszResumePref))
                {
                printf ("  %s %s: Re-opening file\n", pchTable, PchTm (sd));
                fflush (stdout);
                strcpy (rgchTbName[sd], rgtbdDesc[iTb].m_rgchName);
                strcat (rgchTbName[sd], pchExt);
                strcpy (rgchTempName[sd], "temp");
                strcat (rgchTempName[sd], pchExt);
                }
            else
                {
                printf ("  %s %s: Copying and re-opening file\n", pchTable, PchTm (sd));
                fflush (stdout);
                strcpy (rgchTbName[sd], rgtbdDesc[iTb].m_rgchName);
                strcat (rgchTbName[sd], pchExt);
                strcpy (rgchTempName[sd], "temp");
                strcat (rgchTempName[sd], pchExt);

                strcpy (rgszBuf, COPY_COMMAND);
                strcat (rgszBuf, pszResumePref);
                strcat (rgszBuf, rgchTempName[sd]);
                strcat (rgszBuf, " ");
                strcat (rgszBuf, rgchTempName[sd]);
                if (0 != system (rgszBuf))
                    {
                    printf ("*** Unable to create copy\n");
                    exit (1);
                    }
                }
            rgFp[sd] = f_open (rgchTempName[sd], "rb+");
            if (rgCb[sd] > indBVSize)
                {
                printf ("*** Insufficient memory\n");
                fflush (stdout);
                exit (1);
                }
            }
        VReadStatus ("status.txt", rgfChanged, rgcWinMax, rgcLostMax, &cIterations, &iIteration);
        // Start major iterations
        if (iIteration < cIterations)
            {
            VTab (iLevel);
            printf ("  %s:     At least %d more iteration%s necessary\n",
                pchTable, cIterations-iIteration, (1 == cIterations-iIteration) ? "" : "s");
            fflush (stdout);
            }
        tbtLo = (TB_T) (Li0 + iIteration);
        tbtHi = (TB_T) (Mi1 - iIteration);
        }
    else
        {
        // Mark invalid positions and lost in 0; create files; probe minor tables
        for (sd = x_colorWhite; sd <= sdLimit; sd = (color) (sd + 1))
            {
            // Mark
            VTab (iLevel);
            printf ("  %s %s: Marking legal positions, stalemates, and lost in 0\n", pchTable, PchTm (sd));
            fflush (stdout);
            VSetPfn (iTb, sd);
            for (cpu = 0; cpu < cCPUS; cpu ++)
                {
                rgBoards[cpu].m_cb = 0;
                }
            bvArg.VZero (((INDEX) (cbv * sizeof (ULONG))) * 8);
            bvNew.VZero (((INDEX) (cbv * sizeof (ULONG))) * 8);
            VIterateThroughPositions (sd, &BIT_FCalcMaxIndex);
            VCheckGuards();

            // Find sizeof (file)
            rgCb[sd] = rgBoards[0].m_cb;
            for (cpu = 1; cpu < cCPUS; cpu ++)
                {
                if (rgBoards[cpu].m_cb > rgCb[sd])
                    rgCb[sd] = rgBoards[cpu].m_cb;
                }
            rgCb[sd] += 1;
            VTab (iLevel);
            printf ("  %s %s: File size: " DEC_INDEX_FORMAT
                    " positions\n", pchTable, PchTm (sd), rgCb[sd]);
            fflush (stdout);
            if (rgtbdDesc[iTb].m_rgcbLength[sd] != rgCb[sd])
                {
                printf ("*** Something wrong with table sizes\n");
                fflush (stdout);
                exit (1);
                }
            if (rgCb[sd] > indBVSize)
                {
                printf ("*** Insufficient memory\n");
                fflush (stdout);
                exit (1);
                }
            
            // Create file
            const char  *pchExt = PchExt (sd);

            VTab (iLevel);
            printf ("  %s %s: Creating file\n", pchTable, PchTm (sd));
            fflush (stdout);
            strcpy (rgchTbName[sd], rgtbdDesc[iTb].m_rgchName);
            strcat (rgchTbName[sd], pchExt);
            strcpy (rgchTempName[sd], "temp");
            strcat (rgchTempName[sd], pchExt);
            rgFp[sd] = f_open (rgchTempName[sd], "wb+");
            BIT_VWriteFile (rgFp[sd], pbBuf, cbBuf, bvArg, bvNew, rgCb[sd]);
            VCheckGuards();

            // Probe minor tables
            VTab (iLevel);
            printf ("  %s %s: Probing minor tables\n", pchTable, PchTm (sd));
            fflush (stdout);

            size_t  cbCPU;
            BYTE    *pbCPU;
            pbCPU = bvNew.PbData();
            cbCPU = ((bvNew.CbData() / cCPUS) >> 8) << 8;;
            for (cpu = 0; cpu < cCPUS; pbCPU += cbCPU, cpu ++)
                {
#if !defined(NDEBUG)
                rgBoards[cpu].m_uGuard = 0x12345678;
#endif
                rgBoards[cpu].m_pmr = (CMinorResult*) pbCPU;
                rgBoards[cpu].m_cmrMax = cbCPU / sizeof (CMinorResult);
                rgBoards[cpu].m_cmr = 0;
                rgBoards[cpu].m_pbBuf = pbBuf;  // Can do it as flush is non-multithreaded
                rgBoards[cpu].m_cbBuf = cbBuf;
                rgBoards[cpu].m_cWinMax = 0;
                rgBoards[cpu].m_cLostMax = 0;
                rgBoards[cpu].m_fp = rgFp[sd];
                }
            CFormBV::BIT_VFormBV (FUnknownValue, rgFp[sd], pbBuf, cbBuf, bvArg, rgCb[sd]);
#if defined (PRINT)
            printf ("DEBUG: BV formed\n");
            fflush (stdout);
#endif
            VIterateThroughPositions (sd, &BIT_FProbeMinors);
            VCheckGuards();
            rgcWinMax[sd] = 0;
            rgcLostMax[sd] = 0;
            for (cpu = 0; cpu < cCPUS; cpu ++)
                {
#if defined (PRINT)
                printf ("DEBUG: writing minor results for CPU %d\n", cpu);
                fflush (stdout);
#endif
                BIT_VWriteMinorResults (&rgBoards[cpu]);
#if defined (PRINT)
                printf ("DEBUG: wrote minor results for CPU %d\n", cpu);
                fflush (stdout);
#endif
                VCheckGuards();
                if (rgBoards[cpu].m_cWinMax > rgcWinMax[sd])
                    rgcWinMax[sd] = rgBoards[cpu].m_cWinMax;
                if (rgBoards[cpu].m_cLostMax > rgcLostMax[sd])
                    rgcLostMax[sd] = rgBoards[cpu].m_cLostMax;
                }
#if 0
                if (x_colorWhite == sd) {
                    char x[100];

                    VTbCloseFiles();
                    printf ("*** White probing done. Press <<enter>> ");
                    fflush(stdout);
                    gets(x);
                }
#endif
            }

#if defined (PRINT)
        printf ("DEBUG: Minor probing done\n");
        fflush (stdout);
#endif
        VTbCloseFiles();
        // Start major iterations
        cIterations = rgcWinMax[x_colorWhite];
        if (cIterations < rgcLostMax[x_colorWhite])
            cIterations = rgcLostMax[x_colorWhite];
        if (!fSymmetric)
            {
            if (cIterations < rgcWinMax[x_colorBlack])
                cIterations = rgcWinMax[x_colorBlack];
            if (cIterations < rgcLostMax[x_colorBlack])
                cIterations = rgcLostMax[x_colorBlack];
            }
        cIterations ++;
        VTab (iLevel);
        printf ("  %s:     At least %d iteration%s necessary\n",
                pchTable, cIterations, (1 == cIterations) ? "" : "s");
        fflush (stdout);
        iIteration = 0;
        tbtLo = Li0;
        tbtHi = Mi1;
        rgfChanged[x_colorWhite] = false;
        rgfChanged[x_colorBlack] = false;
        }

    // Start/continue major iterations
    if (fSymmetric)
        {
        rgCb[x_colorBlack] = rgCb[x_colorWhite];
        rgFp[x_colorBlack] = rgFp[x_colorWhite];
        rgcWinMax[x_colorBlack] = rgcWinMax[x_colorWhite];
        rgcLostMax[x_colorBlack] = rgcLostMax[x_colorWhite];
        }

    do
        {
        for (sd = x_colorWhite; sd <= sdLimit; sd = (color) (sd + 1))
            {
            if (iIteration <= rgcLostMax[OtherSide(sd)] || rgfChanged[OtherSide(sd)])
                {
                VTab (iLevel);
                printf ("  %s %s: Searching mate in %d\n", pchTable, PchTm (sd), iIteration+1);
                fflush (stdout);
                CFormBV::BIT_VFormBV (FLoValue, rgFp[OtherSide(sd)], pbBuf, cbBuf, bvArg, rgCb[OtherSide(sd)]);
                VCheckGuards();
                if (!fReverse)
                    {
                    VSetPfn (iTb, sd);
                    CFormBV::BIT_VFormBV (FUnstabilizedValueForWin, rgFp[sd], pbBuf, cbBuf, bvNew, rgCb[sd]);
                    VCheckGuards();
                    VIterateThroughPositions (sd, &BIT_FSearchWin);
                    VCheckGuards();
                    rgfChangedNew[sd] = BIT_FWriteFileWin (rgFp[sd], pbBuf, cbBuf, tbtHi, bvNew, rgCb[sd]);
                    }
                else
                    {
                    for (cpu = 0; cpu < cCPUS; cpu ++)
                        rgBoards[cpu].m_fReverse = true;
                    VSetPfn (iTb, OtherSide(sd));
                    bvNew.VZero (rgCb[sd]);
                    VCheckGuards();
            if(rgBoards[0].m_fBothSidesHavePawns)
                VIterateThroughPositions (OtherSide(sd), &BIT_FReverseSearchWinEP);
            else
                    VIterateThroughPositions (OtherSide(sd), &BIT_FReverseSearchWin);
                    for (cpu = 0; cpu < cCPUS; cpu ++)
                        rgBoards[cpu].m_fReverse = false;
                    VCheckGuards();
                    rgfChangedNew[sd] = BIT_FWriteFileReverseWin (rgFp[sd], pbBuf, cbBuf, tbtHi, bvNew, rgCb[sd]);
                    }
                }
            else
                rgfChangedNew[sd] = false;
            }
        if (fSymmetric)
            rgfChangedNew[x_colorBlack] = rgfChangedNew[x_colorWhite];
        rgfChanged[x_colorWhite] = rgfChangedNew[x_colorWhite];
        rgfChanged[x_colorBlack] = rgfChangedNew[x_colorBlack];
        tbtLo ++;
        for (sd = x_colorWhite; sd <= sdLimit; sd = (color) (sd + 1))
            {
            if (iIteration <= rgcWinMax[OtherSide(sd)] || iIteration <= rgcLostMax[sd] || rgfChanged[OtherSide(sd)])
                {
                VTab (iLevel);
                printf ("  %s %s: Searching lost in %d\n", pchTable, PchTm (sd), iIteration+1);
                fflush (stdout);
                VSetPfn (iTb, sd);
                CFormBV::BIT_VFormBV (FGeHiOrBrokenValue, rgFp[OtherSide(sd)], pbBuf, cbBuf, bvArg, rgCb[OtherSide(sd)]);
                VCheckGuards();
                CFormBV::BIT_VFormBV (FUnstabilizedValueForLost, rgFp[sd], pbBuf, cbBuf, bvNew, rgCb[sd]);
                VCheckGuards();
                VIterateThroughPositions (sd, &BIT_FSearchLost);
                VCheckGuards();
                rgfChangedNew[sd] = BIT_FWriteFileLost (rgFp[sd], pbBuf, cbBuf, tbtLo, bvNew, rgCb[sd]);
                }
            else
                rgfChangedNew[sd] = false;
            }
        tbtHi --;
        if (fSymmetric)
            rgfChangedNew[x_colorBlack] = rgfChangedNew[x_colorWhite];
        rgfChanged[x_colorWhite] = rgfChangedNew[x_colorWhite];
        rgfChanged[x_colorBlack] = rgfChangedNew[x_colorBlack];
        iIteration ++;
        if (fSaveStatus && (iIteration % 5) == 0)
            {
            VTab (iLevel);
            printf ("  %s:     Saving intermediate results\n", pchTable);
            fflush (stdout);
            VWriteStatus ("status.txt", rgfChanged, rgcWinMax, rgcLostMax, &cIterations, &iIteration);
            for (sd = x_colorWhite; sd <= sdLimit; sd = (color) (sd + 1))
                {
                char rgszBuf[1024];

                f_close (rgFp[sd]);
                strcpy (rgszBuf, COPY_COMMAND);
                strcat (rgszBuf, rgchTempName[sd]);
                strcat (rgszBuf, " ");
                strcat (rgszBuf, pszSavePref);
                strcat (rgszBuf, rgchTempName[sd]);
                if (0 != system (rgszBuf))
                    {
                    printf ("*** Unable to create copy\n");
                    exit (1);
                    }
                rgFp[sd] = f_open (rgchTempName[sd], "rb+");
                }
            if (fSymmetric)
                rgFp[x_colorBlack] = rgFp[x_colorWhite];
            }
        }
    while (iIteration < cIterations || rgfChanged[x_colorWhite] || rgfChanged[x_colorBlack]);

    // Collect statistics
    FILE    *fpStatistics;
    char    szStatistics[64];

    VTab (iLevel);
    printf ("  %s %s: Done\n", pchTable, PchTm (sd));
    fflush (stdout);

    strcpy (szStatistics, pchTable);
    strcat (szStatistics, ".tbs");
    fpStatistics = fopen (szStatistics, "wt+");
    if (NULL == fpStatistics)
        {
        printf ("*** Unable to create %s\n", szStatistics);
        fflush (stdout);
        exit (1);
        }

    // Replace 'unknown' by 'draw'
    for (sd = x_colorWhite; sd <= sdLimit; sd = (color) (sd + 1))
        {
        VTab (iLevel);
        printf ("  %s %s: Replacing 'unknown' by 'draw'\n", pchTable, PchTm (sd));
        fflush (stdout);
        memset (prgStatistics, 0, 65536 * sizeof (INDEX));
        BIT_VReplaceUnknowns (rgFp[sd], pbBuf, cbBuf);
        if (1 == sizeof (TB_T))
            {
            if (fVerbose)
                VPrintStatistics (stdout, iLevel, PchTm(sd));
            VPrintStatistics (fpStatistics, iLevel, PchTm(sd));
            }
        else
            {
            if (fVerbose)
                L_VPrintStatistics (stdout, iLevel, PchTm(sd));
            L_VPrintStatistics (fpStatistics, iLevel, PchTm(sd));
            }
        }
    // All done!
    for (sd = x_colorWhite; sd <= sdLimit; sd = (color) (sd + 1))
        {
        f_close (rgFp[sd]);
        remove (rgchTbName[sd]);
        if (rename (rgchTempName[sd], rgchTbName[sd]))
            {
            printf ("*** Unable to rename file\n");
            fflush (stdout);
            exit (1);
            }
        }
    fclose (fpStatistics);

    // Register both files
    FCheckExistance (".", iTb, x_colorWhite);
    if (!fSymmetric)
        FCheckExistance (".", iTb, x_colorBlack);

    // Print statistics
    VTab (iLevel);
    printf ("%s done in", pchTable);
    VPrintElapsedTime (tmStart);
    printf ("\n\n");
    fflush (stdout);
    }
};

///////////////////////////////////////////////////////////////////////////////
//
// Specialization of template

// First, 8-bit versions

template<typename TB_T> TB_T BIT_algorithm<TB_T>::tbtLo;
template<typename TB_T> TB_T BIT_algorithm<TB_T>::tbtHi;

template <> INLINE tb_t BIT_algorithm<tb_t>::TbtGet
    (
    BYTE    *pb,
    size_t  offset
    )
    {
    return (tb_t) pb[offset];
    }

template <> INLINE void BIT_algorithm<tb_t>::VPutTbt
    (
    BYTE    *pb,
    size_t  offset,
    tb_t    tbt
    )
    {
    pb[offset] = (BYTE) tbt;
    }

template <> void BIT_algorithm<tb_t>::VFlushChunk
    (
    file    fp,
    BYTE    *pbBuf,
    INDEX   indStart,
    size_t  cPos
    )
    {
    f_write (pbBuf, cPos, fp, indStart);
    }

template <> size_t BIT_algorithm<tb_t>::CposReadChunk
    (
    file    fp,
    BYTE    *pbBuf,
    INDEX   indStart,
    size_t  cPos
    )
    {
    return f_read (pbBuf, cPos, fp, indStart);
    }

template <> tb_t BIT_algorithm<tb_t>::TbtProbeMinor (int iTb, color side, INDEX indexChild)
    {
    return (tb_t) TbtProbeTable (iTb, side, indexChild);
    }

template <> tb_t BIT_algorithm<tb_t>::TbtReplaceMinorResult (tb_t tbtScore)
    {
    return TbtReplaceResult (tbtScore);
    }

// 16-bit versions

template <> INLINE int BIT_algorithm<int>::TbtGet
    (
    BYTE    *pb,
    size_t  offset
    )
    {
#if defined (_MSC_VER)
    return *(short*) (pb+offset*2);
#else
    return (((signed char) pb[offset*2+1]) << 8) + pb[offset*2];
#endif
    }

template <> INLINE void BIT_algorithm<int>::VPutTbt
    (
    BYTE    *pb,
    size_t  offset,
    int     tbt
    )
    {
#if defined (_MSC_VER)
    *(short*) (pb+offset*2) = (short) tbt;
#else
    pb[offset*2]   = (BYTE) tbt;
    pb[offset*2+1] = (BYTE) (tbt >> 8);
#endif
    }

template <> void BIT_algorithm<int>::VFlushChunk
    (
    file    fp,
    BYTE    *pbBuf,
    INDEX   indStart,
    size_t  cPos
    )
    {
#if defined (_MSC_VER)
    LONG    lHiPos;
    DWORD   dwPtr;
    BOOL    bResult;
    DWORD   cbWritten;

#if defined (PRINT)
    printf ("DEBUG: writing %u at position " HEX_INDEX_FORMAT
            "\n", cPos, indStart);
    fflush (stdout);
#endif
    lHiPos = (LONG) (indStart >> 31);
    indStart <<= 1;
    dwPtr = SetFilePointer (fp, (LONG) indStart, &lHiPos, FILE_BEGIN);
    if (0xFFFFFFFF == dwPtr && NO_ERROR != GetLastError())
        {
        printf ("*** Seek failed\n");
        fflush (stdout);
        exit (1);
        return;
        }
    bResult = WriteFile (fp, pbBuf, cPos*2, &cbWritten, NULL);
    if (0 == bResult || cbWritten != cPos*2)
        {
        printf ("*** Write failed\n");
        fflush (stdout);
        exit (1);
        return;
        }
#if defined (PRINT)
    printf ("DEBUG: Wrote\n");
    fflush (stdout);
#endif
#else
    f_write (pbBuf, 2*cPos, fp, 2*indStart);
#endif
    }

template <> size_t BIT_algorithm<int>::CposReadChunk
    (
    file    fp,
    BYTE    *pbBuf,
    INDEX   indStart,
    size_t  cPos
    )
    {
#if defined (_MSC_VER)
    LONG    lHiPos;
    DWORD   dwPtr;
    BOOL    bResult;
    DWORD   cbRead;

#if defined (PRINT)
    printf ("DEBUG: reading %u at position " HEX_INDEX_FORMAT
            "\n", cPos, indStart);
    fflush (stdout);
#endif
    lHiPos = (LONG) (indStart >> 31);
    indStart <<= 1;
    dwPtr = SetFilePointer (fp, (LONG) indStart, &lHiPos, FILE_BEGIN);
    if (0xFFFFFFFF == dwPtr && NO_ERROR != GetLastError())
        {
        printf ("*** Seek failed\n");
        fflush (stdout);
        exit (1);
        return 0;
        }
    bResult = ReadFile (fp, pbBuf, cPos*2, &cbRead, NULL);
    if (0 == bResult)
        {
        printf ("*** Read failed\n");
        fflush (stdout);
        exit (1);
        return 0;
        }
#if defined (PRINT)
    printf ("DEBUG: read %u\n", cbRead/2);
    fflush (stdout);
#endif
    return cbRead/2;
#else
    return f_read (pbBuf, 2*cPos, fp, 2*indStart) / 2;
#endif
    }

template <> int BIT_algorithm<int>::TbtProbeMinor (int iTb, color side, INDEX indexChild)
    {
    return L_TbtProbeTable (iTb, side, indexChild);
    }

template <> int BIT_algorithm<int>::TbtReplaceMinorResult (int tbtScore)
    {
    return L_TbtReplaceResult (tbtScore);
    }

///////////////////////////////////////////////////////////////////////////////
//
// Create TB - "classic" approach, not bit vector one

static void VCreateTable
    (
    int iTb,
    int iLevel
    )
    {
    char    *pchTable;
    time_t  tmStart;
    color   sd;
    color   sdLimit;
    int     iIteration;
    int     cpu;
    char    rgchTbName[2][16];
    char    rgchTempName[2][16];
    bool    fSymmetric = rgtbdDesc[iTb].m_fSymmetric;

    pchTable = rgtbdDesc[iTb].m_rgchName;
    VTab (iLevel);
    printf ("Starting %s\n", pchTable);
    fflush (stdout);

    // Check minor tables
    VRecurse (pchTable, iLevel);
    VTab (iLevel);
    printf ("  All minor for %s exists, generating it\n", pchTable);
    fflush (stdout);
    
    // Initialization
    tmStart = time (NULL);
    for (cpu = 0; cpu < cCPUS; cpu ++)
        {
        rgBoards[cpu].m_fReverse = false;
        rgBoards[cpu].VInitIterations (pchTable);
        }

    // Build tables
    ULONG   cbTables;
    ULONG   cb;

    cbTables = cbMemoryUsed;
    // cbTables = min (available memory, necessary memory)
    cb = rgtbdDesc[iTb].m_rgcbLength[x_colorWhite];
    if (!fSymmetric)
        cb += rgtbdDesc[iTb].m_rgcbLength[x_colorBlack];
    if (cbTables > cb)
        cbTables = cb;
    if (tbid_kppkp == iTb)
        {
        cbTables *= 2;
        if (cbTables > cbMemoryUsed)
            cbTables = cbMemoryUsed;
        for (ULONG cb = 0; cb < cbTables; cb += 2)
            {
            pbMemory[cb]   = (BYTE) L_bev_broken;
            pbMemory[cb+1] = (BYTE) (L_bev_broken >> 8);
            }
        }
    else
        memset (pbMemory, bev_broken, cbTables);
    rgpbFile[x_colorWhite] = pbMemory;
    cbCalc = cbMemoryUsed;

    sdLimit = x_colorBlack;
    if (fSymmetric)
        sdLimit = x_colorWhite;
    for (sd = x_colorWhite; sd <= sdLimit; sd = (color) (sd + 1))
        {
        VSetPfn (iTb, sd);
        for (cpu = 0; cpu < cCPUS; cpu ++)
            {
            rgBoards[cpu].m_cb = 0;
            }
        // Find sizeof(file)
        VTab (iLevel);
        printf ("  %s %s: Determining file size\n", pchTable, PchTm (sd));
        fflush (stdout);
        VIterateThroughPositions (sd, tbid_kppkp == iTb ? &L_FCalcMaxIndex: &FCalcMaxIndex);
        rgCb[sd] = rgBoards[0].m_cb;
        for (cpu = 1; cpu < cCPUS; cpu ++)
            {
            if (rgBoards[cpu].m_cb > rgCb[sd])
                rgCb[sd] = rgBoards[cpu].m_cb;
            }
        rgCb[sd] += 1 + (tbid_kppkp == iTb);
        VTab (iLevel);
        printf ("  %s %s: File size: " DEC_INDEX_FORMAT
                " positions\n", pchTable, PchTm (sd), rgCb[sd]);
        fflush (stdout);
        if (x_colorWhite == sd)
            {
            cbCalc = (rgCb[x_colorWhite] > cbMemoryUsed) ? 0 : cbMemoryUsed-rgCb[x_colorWhite];
            rgpbFile[x_colorBlack] = rgpbFile[x_colorWhite] + rgCb[x_colorWhite];
            }
        }
    if (fSymmetric)
        {
        rgCb[x_colorBlack] = rgCb[x_colorWhite];
        rgpbFile[x_colorBlack] = rgpbFile[x_colorWhite];
        if (rgCb[x_colorWhite] > cbMemoryUsed)
            {
            printf ("*** Insufficient memory\n");
            fflush (stdout);
            exit (1);
            }
        }
    else if (rgCb[x_colorWhite] + rgCb[x_colorBlack] > cbMemoryUsed)
        {
        printf ("*** Insufficient memory\n");
        fflush (stdout);
        exit (1);
        }
    if (iTb != tbid_kppkp &&
        rgtbdDesc[iTb].m_rgcbLength[x_colorWhite]+rgtbdDesc[iTb].m_rgcbLength[x_colorBlack] != rgCb[x_colorWhite]+rgCb[x_colorBlack])
        {
        printf ("*** Something wrong with table sizes\n");
        fflush (stdout);
        exit (1);
        }

    // Create files
    for (sd = x_colorWhite; sd <= sdLimit; sd = (color) (sd + 1))
        {
        const char  *pchExt = PchExt (sd);

        VTab (iLevel);
        printf ("  %s %s: Creating file\n", pchTable, PchTm (sd));
        fflush (stdout);
        strcpy (rgchTbName[sd], rgtbdDesc[iTb].m_rgchName);
        strcat (rgchTbName[sd], pchExt);
        strcpy (rgchTempName[sd], "temp");
        strcat (rgchTempName[sd], pchExt);

        rgFp[sd] = f_open (rgchTempName[sd], "wb+");
        f_write (rgpbFile[sd], rgCb[sd], rgFp[sd], 0);
        // Close and re-open file, to be sure it's flushed
        f_close (rgFp[sd]);
#if !defined (KPPKP_16BIT)
        rgFp[sd] = (tbid_kppkp == iTb) ? f_open (rgchTempName[sd], "wb+"):
                                         f_open (rgchTempName[sd], "rb+");
#else
        rgFp[sd] = f_open (rgchTempName[sd], "rb+");
#endif
        }

    // Now start iterations
    iIteration = 1;
    tbtLo = bev_li0;
    tbtHi = bev_mi1;
    L_tbtLo = L_bev_li0;
    L_tbtHi = L_bev_mi1;
    do
        {
        for (sd = x_colorWhite; sd <= sdLimit; sd = (color) (sd + 1))
            {
            VTab (iLevel);
            printf ("  %s %s: Starting major iteration %d\n", pchTable, PchTm (sd), iIteration);
            fflush (stdout);
            VSetPfn (iTb, sd);
            fChanged = false;
            VIterateThroughPositions (sd, tbid_kppkp == iTb ? &L_FProbeOpposite : &FProbeOpposite);
            }
        iIteration ++;
        tbtLo = (tb_t) (tbtLo + ((1 == rgBoards[0].m_rgPieceCount[x_colorBlack]) || (1 != (iIteration & 1))));
        tbtHi = (tb_t) (tbtHi - ((1 == rgBoards[0].m_rgPieceCount[x_colorBlack]) || (1 != (iIteration & 1))));
        L_tbtLo = (L_tbtLo + ((1 == rgBoards[0].m_rgPieceCount[x_colorBlack]) || (1 != (iIteration & 1))));
        L_tbtHi = (L_tbtHi - ((1 == rgBoards[0].m_rgPieceCount[x_colorBlack]) || (1 != (iIteration & 1))));
        }
    while (fChanged);

    // Collect statistics
    FILE    *fpStatistics;
    char    szStatistics[64];

    VTab (iLevel);
    printf ("  %s %s: Done\n", pchTable, PchTm (sd));
    fflush (stdout);

    strcpy (szStatistics, pchTable);
    strcat (szStatistics, ".tbs");
    fpStatistics = fopen (szStatistics, "wt+");
    if (NULL == fpStatistics)
        {
        printf ("*** Unable to create %s\n", szStatistics);
        fflush (stdout);
        exit (1);
        }
    memset (prgStatistics, 0, 65536 * sizeof (INDEX));
    if (tbid_kppkp == iTb)
        {
        L_VCollectStatistics (rgpbFile[x_colorWhite], rgCb[x_colorWhite]);
        if (fVerbose)
            L_VPrintStatistics (stdout, iLevel, "wtm");
        L_VPrintStatistics (fpStatistics, iLevel, "wtm");
        }
    else
        {
        VCollectStatistics (rgpbFile[x_colorWhite], rgCb[x_colorWhite]);
        if (fVerbose)
            VPrintStatistics (stdout, iLevel, "wtm");
        VPrintStatistics (fpStatistics, iLevel, "wtm");
        }
    memset (prgStatistics, 0, 65536 * sizeof (INDEX));
    if (tbid_kppkp == iTb)
        {
        L_VCollectStatistics (rgpbFile[x_colorBlack], rgCb[x_colorBlack]);
        if (fVerbose)
            L_VPrintStatistics (stdout, iLevel, "btm");
        L_VPrintStatistics (fpStatistics, iLevel, "btm");
        }
    else
        {
        VCollectStatistics (rgpbFile[x_colorBlack], rgCb[x_colorBlack]);
        if (fVerbose)
            VPrintStatistics (stdout, iLevel, "btm");
        VPrintStatistics (fpStatistics, iLevel, "btm");
        }
    fclose (fpStatistics);

    VWriteFile (x_colorWhite, rgpbFile[x_colorWhite], iTb);
    if (!fSymmetric)
        VWriteFile (x_colorBlack, rgpbFile[x_colorBlack], iTb);

    // All done!
    for (sd = x_colorWhite; sd <= sdLimit; sd = (color) (sd + 1))
        {
        f_close (rgFp[sd]);
        remove (rgchTbName[sd]);
        if (rename (rgchTempName[sd], rgchTbName[sd]))
            {
            printf ("*** Unable to rename file\n");
            fflush (stdout);
            exit (1);
            }
        }

    // Register both files
    FCheckExistance (".", iTb, x_colorWhite);
    if (!fSymmetric)
        FCheckExistance (".", iTb, x_colorBlack);

    // Print statistics
    VTab (iLevel);
    printf ("%s done in", pchTable);
    VPrintElapsedTime (tmStart);
    printf ("\n\n");
    fflush (stdout);
    }

// Convert TB name into TB #

static int  IFindTb
    (
    char    *pch
    )
    {
    int rgiCount1[10], rgiCount2[10];
    
    if (FReasonableName (pch))
        {
        VSetCounters (rgiCount1, pch);
        for (int iTb = 1; iTb < cTb; iTb ++)
            {
            VSetCounters (rgiCount2, rgtbdDesc[iTb].m_rgchName);
            if (0 == memcmp (rgiCount1, rgiCount2, 10 * sizeof (int)) ||
                (0 == memcmp (rgiCount1, rgiCount2+5, 5 * sizeof (int)) &&
                 0 == memcmp (rgiCount1+5, rgiCount2, 5 * sizeof (int))))
                return iTb;
            }
        }
    printf ("*** Illegal tablebase name: %s\n", pch);
    fflush (stdout);
    exit (1);
    return 0;   // To make compiler happy
    }

static void VProcessTb
    (
    char    *pch,
    int     iLevel
    )
    {
    int iTb;

    iTb = IFindTb (pch);
    if (fBitVectors)
        {
        if (!rgtbdDesc[iTb].m_f16bit)
            BIT_algorithm<tb_t>::BIT_VCreateTable (iTb, iLevel);
        else
#if defined (T33_INCLUDE) || defined (T42_INCLUDE)
            BIT_algorithm<int>::BIT_VCreateTable (iTb, iLevel);
#else
            VCreateTable (iTb, iLevel);
#endif
        }
    else
        VCreateTable (iTb, iLevel);
    }

#if defined (CALCULATE_SIZEOF)

static void VCalcTbSize
    (
    int iTb
    )
    {
    char    *pchTable;
    color   sd;
    color   sdLimit;
    int     iIteration;
    int     cpu;

    pchTable = rgtbdDesc[iTb].m_rgchName;

    // Initialization
    for (cpu = 0; cpu < cCPUS; cpu ++)
        rgBoards[cpu].VInitIterations (pchTable);
    
    // Calculate sizeof(tables)

    sdLimit = x_colorBlack;
    if (rgtbdDesc[iTb].m_fSymmetric)
        sdLimit = x_colorWhite;
    cbCalc = 0;
    for (sd = x_colorWhite; sd <= sdLimit; sd = (color) (sd + 1))
        {
        VSetPfn (iTb, sd);
        // Find sizeof(file)
        rgCb[sd] = 0;
        for (cpu = 0; cpu < cCPUS; cpu ++)
            rgBoards[cpu].m_cb = 0;
        VIterateThroughPositions (sd, &FCalcMaxIndex);
        rgCb[sd] = rgBoards[0].m_cb;
        for (cpu = 1; cpu < cCPUS; cpu ++)
            {
            if (rgBoards[cpu].m_cb > rgCb[sd])
                rgCb[sd] = rgBoards[cpu].m_cb;
            }
        rgCb[sd] ++;
        }
    if (rgtbdDesc[iTb].m_fSymmetric)
        rgCb[x_colorBlack] = rgCb[x_colorWhite];
    }

#endif

//-----------------------------------------------------------------------------
//
//  Verify existing table

static void VVerifyTb
    (
    char *pch
    )
    {
    char    *pchTable;
    time_t  tmStart;
    color   sd;
    int     iTb;
    int     cpu;
#if defined (_WIN32) || defined(_WIN64)
    HANDLE  rghFile[2];
    HANDLE  rghFileMapping[2];
#endif
    color   sdLimit;
    
    iTb = IFindTb (pch);
    pchTable = rgtbdDesc[iTb].m_rgchName;
    printf ("Verifying %s\n", pchTable);
    fflush (stdout);

    // Initialization
    tmStart = time (NULL);
    for (cpu = 0; cpu < cCPUS; cpu ++)
        rgBoards[cpu].VInitIterations (pchTable);
    sdLimit = x_colorBlack;
    if (rgtbdDesc[iTb].m_fSymmetric)
        sdLimit = x_colorWhite;

    // Map/load tables, if necessary
    if (fMapFiles)
        {
        for (sd = x_colorWhite; sd <= sdLimit; sd = (color) (sd + 1))
            {
            int fMapped;

#if defined (_WIN32) || defined(_WIN64)
            fMapped = FMapTableToMemory (iTb, sd, &rghFile[sd], &rghFileMapping[sd]);
#else
            fMapped = FReadTableToMemory (iTb, sd, NULL);
#endif
            if (!fMapped)
                {
                printf ("*** Unable to map %s", pch);
                fflush (stdout);
                exit (1);
                }
            }
        }

    // Now check both tables
    iGlobalTb = iTb;
    for (sd = colVerify; sd <= sdLimit; sd = (color) (sd + 1))
        {
        printf ("  Verifying %s %s\n", pchTable, PchTm (sd));
        fflush (stdout);
        VSetPfn (iTb, sd);
        VIterateThroughPositions (sd, &L_FVerify);
        }
    
    // Unmap files, if necessary
#if defined (_WIN32) || defined(_WIN64)
    if (fMapFiles)
        {
        for (sd = x_colorWhite; sd <= sdLimit; sd = (color) (sd + 1))
            FUnMapTableFromMemory (iTb, sd, rghFile[sd], rghFileMapping[sd]);
        }
#endif

    // Print statistics
    printf ("%s verified in", pchTable);
    VPrintElapsedTime (tmStart);
    printf ("\n\n");
    fflush (stdout);
    }

#if defined(VERIFY_INDEX)
static void VVerifyIndex(int iTb, color side)
{
    char    *pchTable;
    int cpu;

    pchTable = rgtbdDesc[iTb].m_rgchName;

    for(cpu = 0; cpu < cCPUS; cpu++) {
        rgBoards[cpu].VInitIterations(pchTable);
        rgBoards[cpu].m_cb = 0;
    }

    VSetPfn(iTb, side);
    VIterateThroughPositions(side, &FCheckIndex);

    rgCb[side] = rgBoards[0].m_cb;
    for(cpu = 1; cpu < cCPUS; cpu ++) {
        if (rgBoards[cpu].m_cb > rgCb[side])
            rgCb[side] = rgBoards[cpu].m_cb;
    }
    rgCb[side] += 1;
}
#endif

//-----------------------------------------------------------------------------
//
//  Search value in the table

static  int     cMoves;
static  int     iWinLossDraw;

#define SEARCH_WIN  0
#define SEARCH_LOSS 1
#define SEARCH_DRAW 2

static void VSearchTb
    (
    char *pch
    )
    {
    char    *pchTable;
    time_t  tmStart;
    int     iTb;
    int     cpu;
#if defined (_WIN32) || defined(_WIN64)
    HANDLE  hFile;
    HANDLE  hFileMapping;
#endif
    unsigned cSearch;
    
    iTb = IFindTb (pch);
    pchTable = rgtbdDesc[iTb].m_rgchName;
    if (SEARCH_DRAW == iWinLossDraw)
        printf ("Scanning %s %s for draw\n\n", pchTable, PchTm(colSearch));
    else
        printf ("Scanning %s %s for %s in %d\n\n",
                pchTable, PchTm(colSearch),
                SEARCH_WIN == iWinLossDraw ? "win" : "loss", cMoves);
    fflush (stdout);

    // Initialization
    tmStart = time (NULL);
    for (cpu = 0; cpu < cCPUS; cpu ++)
        {
        rgBoards[cpu].VInitIterations (pchTable);
        rgBoards[cpu].m_cSearch = 0;
        }

    // Map/load tables, if necessary
    if (fMapFiles)
        {
        int fMapped;

#if defined (_WIN32) || defined(_WIN64)
        fMapped = FMapTableToMemory (iTb, colSearch, &hFile, &hFileMapping);
#else
        fMapped = FReadTableToMemory (iTb, colSearch, NULL);
#endif
        if (!fMapped)
            {
            printf ("*** Unable to map %s", pch);
            fflush (stdout);
            exit (1);
            }
        }

    // Now search the value
    iGlobalTb = iTb;
    VSetPfn (iTb, colSearch);
    VIterateThroughPositions (colSearch, &L_FFindScore);
    
    // Unmap files, if necessary
#if defined (_WIN32) || defined(_WIN64)
    if (fMapFiles)
        FUnMapTableFromMemory (iTb, colSearch, hFile, hFileMapping);
#endif

    // Print statistics
    cSearch = 0;
    for (cpu = 0; cpu < cCPUS; cpu ++)
        cSearch += rgBoards[cpu].m_cSearch;
    if (0 == cSearch)
        printf ("No positions found\n");
    else
        printf ("\n%u position%s found\n", cSearch, 1 == cSearch ? "" : "s");
    printf ("%s scanned in", pchTable);
    VPrintElapsedTime (tmStart);
    printf ("\n\n");
    fflush (stdout);
    }

//-----------------------------------------------------------------------------

void CBoard::VQuery (void)
    {
    char    rgchPieces[128];
    int     rgiCounters[10];
    int     i;
    square  sq;
    piece   pi;
    color   side;
    color   probe;
    int     fInvert;
    int     iTb;
    INDEX   ind;
    piece   *psqW;
    piece   *psqB;
    int     tbValue;

again:
    for (i = 0; i < 120; i ++)
        m_rgColor[i] = x_colorNeutral;
    for (i = 0; i < 120; i ++)
        m_rgBoard[i] = x_pieceNone;
    m_rgPieceCount[x_colorWhite] = 
    m_rgPieceCount[x_colorBlack] = 1;
    memset (rgiCounters, 0, sizeof (rgiCounters));

    printf ("Enter piece locations (e.g. 'wke3 wpb4 bka6 bph6') or 'quit':\n");
    fflush (stdout);
    gets (rgchPieces);
    fflush (stdout);
    if (0 == memcmp (rgchPieces, "quit", 4))
        return;
    i = 0;
    while ('\0' != rgchPieces[i])
        {
        side = x_colorWhite;
        switch (rgchPieces[i])
            {
        case ' ':
        case '\t':
        case '\n':
        case '\r':
            i ++;
            break;
        case 'b':
            side = x_colorBlack;
            // Fall through
        case 'w':
            if ('\0' == rgchPieces[i+1] ||
                rgchPieces[i+2] < 'a' || rgchPieces[i+2] > 'h' ||
                rgchPieces[i+3] < '1' || rgchPieces[i+3] > '8')
                goto error;
            sq = SquareMake (rgchPieces[i+3]-'1', rgchPieces[i+2] - 'a');
            switch (rgchPieces[i+1])
                {
            case 'p':
                pi = x_piecePawn;
                break;
            case 'n':
                pi = x_pieceKnight;
                break;
            case 'b':
                pi = x_pieceBishop;
                break;
            case 'r':
                pi = x_pieceRook;
                break;
            case 'q':
                pi = x_pieceQueen;
                break;
            case 'k':
                pi = x_pieceKing;
                break;
            default:
                goto error;
                }
            i += 4;
            m_rgColor[sq] = side;
            m_rgBoard[sq] = pi;
            if (x_pieceKing == pi)
                m_rgsqPieceList[side][0] = sq;
            else
                {
                rgiCounters[side*5+pi-x_piecePawn] ++;
                m_rgsqPieceList[side][m_rgPieceCount[side]] = sq;
                m_rgPieceCount[side] ++;
                }
            break;
        default:
            goto error;
            }
        }
        if (m_rgPieceCount[x_colorWhite] + m_rgPieceCount[x_colorBlack] < 3 ||
            m_rgPieceCount[x_colorWhite] + m_rgPieceCount[x_colorBlack] > 6)
            goto error;
        m_sqEnPassant = XX;
        for (side = x_colorWhite; side <= x_colorBlack; side ++)
            {
            if (FAtacked (m_rgsqPieceList[OtherSide(side)][0], side))
                continue;
            iTb = IDescFindFromCounters (rgiCounters);
            if (0 == iTb)
                {
                printf ("%s tablebase not found\n", PchTm (side));
                goto again;
                }
            if (iTb > 0)
                {
                probe = side;
                fInvert = false;
                psqW = m_rgsqPieceList[x_colorWhite];
                psqB = m_rgsqPieceList[x_colorBlack];
                }
            else
                {
                probe = OtherSide (side);
                fInvert = true;
                psqW = m_rgsqPieceList[x_colorBlack];
                psqB = m_rgsqPieceList[x_colorWhite];
                iTb = -iTb;
                }
            if (!FRegistered(iTb, probe))
                {
                printf ("%s tablebase not found\n", PchTm (side));
                goto again;
                }
            ind = IndCalc (iTb, probe, psqW, psqB, Linear(m_sqEnPassant), fInvert);
            tbValue = L_TbtProbeTable (iTb, probe, ind);
            if (tbValue >= L_bev_mimin && tbValue <= L_bev_mi1)
                printf ("%s: Mate in %3d\n", PchTm (side), L_bev_mi1 - tbValue + 1);
            else if (tbValue >= L_bev_li0 && tbValue <= L_bev_limax)
                printf ("%s: Lost in %3d\n", PchTm (side), L_bev_mi1 + tbValue);
            else if (0 == tbValue)
                printf ("%s: Draw\n", PchTm (side));
            else
                printf ("%s: Invalid table value %02X\n", PchTm (side), tbValue & 0xFF);
            }
        goto again;
error:
        printf ("Illegal input - reenter position or press ctrl-C\n");
        goto again;
    }

//-----------------------------------------------------------------------------

static void VShowHelp
    (
    bool fShowHeader
    )
    {
    if (fShowHeader)
        {
        printf
            (
            "New tablebase generator version 4.00\n"
            "(C) 1998, 1999, 2000 Eugene Nalimov, eugenen@microsoft.com\n"
            "\n"
            );
        }
    printf
        (
        "Usage: tbgen options tablebases\n"
        "\n"
        "Valid options:\n"
                );
    if (CPUS >= 2)
        printf
        (
        "  -2               - use 2 processors\n"
        );
    if (CPUS >= 4)
        printf
        (
        "  -4               - use 4 processors\n"
        );
    if (CPUS >= 8)
        printf
        (
        "  -8               - use 8 processors\n"
        );

    printf
        (
        "  -b{+|-}number    - search all win/lost positions of the specified depth, btm\n"
        "  -b number        - use bit vector algorithm with specified amount of memory\n"
        "                     for buffer (Mb)\n"
        "  -c number        - specify TB cache size (in Mb)\n"
        "  -d dirs          - search TBs in the specified directories\n"
        "  -e               - verify existing tablebases\n"
        "                     (you must to specify -c when verifying tablebase)\n"
#if defined (_WIN32) || defined(_WIN64)
        "  -i               - generator will run with low priority\n"
#endif
        "  -m {number|auto} - specify memory size (in Mb), or let generator determine\n"
        "                     memory size automatically (for specified TB only, all\n"
        "                     larger promotion TBs must be present)\n"
        "  -p               - use file mapping (Win32/64 only) or virtual memory\n"
        "                     (do not use -p together with -c when creating tablebase)\n"
        "  -q               - query tablebases\n"
        "  -s prefix        - regularly save status, so generation can be resumed\n"
        "                     (requires bit vector mode)\n"
#if defined(VERIFY_INDEX)
    "  -ti              - check index uniqueness for tablebase\n"
#endif
        "  -u prefix        - resume interrupted generation (requires bit vector mode)\n"
        "  -v               - verbose mode\n"
        "  -w{+|-}number    - search all win/lost positions of the specified depth, wtm\n"
        );
    printf
        (
        "\n"
        "Examples: tbgen -m 570 -p -d c:\\tb;d:\\tb kbpkb\n"
        );
    if (CPUS > 1)
        printf
        (
        "          tbgen -2 -m auto -b 16 -p -d c:\\tb;d:\\tb -s sav_ -u sav_ kbpkb\n"
        );
    else
        printf
        (
        "          tbgen -m auto -b 16 -p -d c:\\tb;d:\\tb -s sav_ -u sav_ kbpkb\n"
        );
    printf
        (
        "          tbgen -w+106 -c 128 kbbknn\n"
        );
    }

//-----------------------------------------------------------------------------
//
//      Main

static  bool    fSearch = false;
static  bool    fQuery  = false;
static  char    *pszDirs = "";

int TB_CDECL main
    (
    int     argc,
    char    *argv[]
    )
    {
    time_t  tmStart;
    int     iArg;
    bool    fAuto = false;
#if defined(VERIFY_INDEX)
    bool    fVerifyIndex = false;
#endif

#if defined (CALC_TOTAL_SIZE)
    unsigned uKbSize = 0;

    for (int iTb = 0; iTb < cTb; iTb ++)
        uKbSize +=  (rgtbdDesc[iTb].m_rgcbLength[0] + 1023)/1024 +
                    (rgtbdDesc[iTb].m_rgcbLength[1] + 1023)/1024;
    printf ("Total TB size is %dKb\n", uKbSize);
    return 0;
#endif

#if defined (SMP)
    LockInit (lockKings);
    LockInit (lockMR);
    LockInit (lockIO);
    for (int i = 0; i < sizeof (rgBitLocks) / sizeof (lock_t); i ++)
        LockInit (rgBitLocks[i]);
#endif
    prgStatistics = (INDEX *) malloc (65536 * sizeof (INDEX));
    if (NULL == prgStatistics)
        {
        printf ("Unable to allocate %luKb of memory", 65536 * sizeof (INDEX) / 1024);
        fflush (stdout);
        exit (1);
        }
    tmStart = time (NULL);
    if (1 == argc)
        {
        VShowHelp (false);
        fflush (stdout);
        exit (0);
        }
    cCPUS = 1;
    // Process non-TB arguments
    for (iArg = 1; iArg < argc;)
        {
        if ('-' == argv[iArg][0] || '/' == argv[iArg][0])
            {
            switch (argv[iArg][1])
                {
            case 'u':
            case 'U':
                if (iArg+1 < argc)
                    {
                    fResume = true;
                    pszResumePref = argv[iArg+1];
                    memmove (argv+iArg, argv+iArg+1, (argc-iArg) * sizeof(char*));
                    argc --;
                    break;
                    }
                else
                    goto error;
            case 's':
            case 'S':
                if (iArg+1 < argc)
                    {
                    fSaveStatus = true;
                    pszSavePref = argv[iArg+1];
                    memmove (argv+iArg, argv+iArg+1, (argc-iArg) * sizeof(char*));
                    argc --;
                    break;
                    }
                else
                    goto error;
#if defined (SMP)
            case '1':
            case '2':
            case '3':
            case '4':
            case '5':
            case '6':
            case '7':
            case '8':
            case '9':
                cCPUS = atoi(argv[iArg]+1);
                if (cCPUS > CPUS)
                    cCPUS = CPUS;
                break;
#endif
            case 'b':
            case 'B':
                if ('+' == argv[iArg][2] || '-' == argv[iArg][2] || 'd' == argv[iArg][2])
                    {
                    colSearch = x_colorBlack;
                    goto search;
                    }
                fReverse = (argv[iArg][1] == 'b');
                fBitVectors = true;
                if (iArg+1 < argc)
                    {
                    char *pchEnd;

                    cbBuf = strtoul (argv[iArg+1], &pchEnd, 10);
                    if (!isdigit (argv[iArg+1][0]) || '\0' != *pchEnd || 0 == cbBuf)
                        goto error;
                    cbBuf *= 1024*1024;
                    memmove (argv+iArg, argv+iArg+1, (argc-iArg) * sizeof(char*));
                    argc --;
                    }
                else
                    goto error;
                break;
#if defined(VERIFY_INDEX)
        case 't':
        case 'T':
            if(argv[iArg][2] == 'i' || argv[iArg][2] == 'I')
                fVerifyIndex = true;
            else
                goto error;
            break;
#endif
            case 'w':
            case 'W':
                colSearch = x_colorWhite;
search:
                fSearch = true;
                if ('+' == argv[iArg][2])
                    {
                    tbtSearch = L_bev_mi1 - atoi (&argv[iArg][3]) + 1;
                    iWinLossDraw = SEARCH_WIN;
                    }
                else if ('-' == argv[iArg][2])
                    {
                    tbtSearch = L_bev_li0 + atoi (&argv[iArg][3]);
                    iWinLossDraw = SEARCH_LOSS;
                    }
                else if ('d' == argv[iArg][2])
                    {
                    tbtSearch = 0;
                    iWinLossDraw = SEARCH_DRAW;
                    break;
                    }
                else
                    goto error;
                cMoves = atoi (&argv[iArg][3]);
                break;
            case '?':
                VShowHelp (true);
                fflush (stdout);
                exit (0);
            case 'v':
            case 'V':
                fVerbose = true;
                break;
            case 'e':
            case 'E':
                fVerify = true;
                colVerify = x_colorWhite;
                if ('b' == argv[iArg][2])
                    colVerify = x_colorBlack;
                break;
            case 'r':
            case 'R':
                fPrint = true;
                break;
            case 'p':
            case 'P':
                fMapFiles = true;
                break;
#if defined (_WIN32) || defined(_WIN64)
            case 'i':
            case 'I':
                SetPriorityClass (GetCurrentProcess, IDLE_PRIORITY_CLASS);
                break;
#endif
            case 'd':
            case 'D':
                if (iArg+1 < argc)
                    {
                    pszDirs = argv[iArg+1];
                    memmove (argv+iArg, argv+iArg+1, (argc-iArg) * sizeof(char*));
                    argc --;
                    break;
                    }
                else
                    goto error;
            case 'c':
            case 'C':
                if (iArg+1 < argc)
                    {
                    char *pchEnd;

                    cbMemoryCache = strtoul (argv[iArg+1], &pchEnd, 10);
                    if (!isdigit (argv[iArg+1][0]) || '\0' != *pchEnd || 0 == cbMemoryCache)
                        goto error;
                    cbMemoryCache *= 1024*1024;
                    memmove (argv+iArg, argv+iArg+1, (argc-iArg) * sizeof(char*));
                    argc --;
                    break;
                    }
                else
                    goto error;
            case 'q':
            case 'Q':
                fQuery = true;
                break;
            case 'm':
            case 'M':
                if (iArg+1 < argc)
                    {
                    if (!strcmp (argv[iArg+1], "auto"))
                        fAuto = true;
                    else
                        {
                        char *pchEnd;

                        cbMemoryUsed = strtoul (argv[iArg+1], &pchEnd, 10);
                        if (!isdigit (argv[iArg+1][0]) || '\0' != *pchEnd || 0 == cbMemoryUsed)
                            goto error;
                        cbMemoryUsed *= 1024*1024;
                    }
                    memmove (argv+iArg, argv+iArg+1, (argc-iArg) * sizeof(char*));
                    argc --;
                    break;
                    }
                // Fall through to error case
            default:
error:          printf ("Illegal option '%s'\n", argv[iArg]);
                VShowHelp (false);
                fflush (stdout);
                exit (1);
                }
            memmove (argv+iArg, argv+iArg+1, (argc-iArg) * sizeof(char*));
            argc --;
            }
        else
            iArg ++;
        }

    if ((fResume | fSaveStatus) & !fBitVectors)
        {
        printf ("Saving/restoring requires bit vector mode");
        fflush (stdout);
        exit (1);
        }

#if defined (SMP)
    printf ("Using %d processor%s\n", cCPUS, cCPUS > 1 ? "s" : "");
    fflush (stdout);
#endif
#if defined (T_INDEX64)
    printf ("64-bit index\n");
    fflush (stdout);
#endif
#if defined (T41_INCLUDE)
    printf ("4+1 TBs enabled\n");
    fflush (stdout);
#endif
#if defined (T33_INCLUDE)
    printf ("3+3 TBs enabled\n");
    fflush (stdout);
#endif
#if defined (T42_INCLUDE)
    printf ("4+2 TBs enabled\n");
    fflush (stdout);
#endif

#if defined(VERIFY_INDEX)
    if(fVerifyIndex) {
        IInitializeTb (NULL);
        VInitializeAttacks();

        int iTb, nTb = -1;

        printf("Checking uniqueness of position indices\n");

        if(argc >= 2) nTb = IFindTb(argv[1]);

        maxTbSize = 0;
        // first find maximum size
        for(iTb = tbid_kpk; iTb < cTb; iTb++) {
            if(nTb > 0 && iTb != nTb)
                continue;
            if(rgtbdDesc[iTb].m_rgcbLength[x_colorWhite] > maxTbSize)
                maxTbSize = rgtbdDesc[iTb].m_rgcbLength[x_colorWhite];
            if(rgtbdDesc[iTb].m_rgcbLength[x_colorBlack] > maxTbSize)
                maxTbSize = rgtbdDesc[iTb].m_rgcbLength[x_colorBlack];
        }

        cbv = (maxTbSize + 8*sizeof(ULONG)-1) / (8*sizeof(ULONG));
        indBVSize = 8*sizeof(ULONG)*(INDEX)cbv;

        printf("Allocating bit vector with %lu bytes\n",cbv*sizeof(ULONG));

            bvArg.VInit (cbv * sizeof(ULONG));

        for (iTb = tbid_kpk; iTb < cTb; iTb ++) {
            if(nTb > 0 && iTb != nTb)
                continue;
            color sd, sdLimit;

            sdLimit = x_colorBlack;
            if(rgtbdDesc[iTb].m_fSymmetric)
                sdLimit = x_colorWhite;

            for(sd = x_colorWhite; sd <= sdLimit; sd = (color)(sd+1)) {
                bvArg.VZero(indBVSize);

                iDuplicateIndex = (INDEX)(-1);
                fPrintDuplicates = false;

                maxTbSize = rgtbdDesc[iTb].m_rgcbLength[sd];

                printf("Checking %s, %s...",
                   rgtbdDesc[iTb].m_rgchName, PchTm(sd));
                VVerifyIndex(iTb, sd);

                if(iDuplicateIndex != (INDEX)(-1)) {
                    printf("\n%s %s contains non-unique index " DEC_INDEX_FORMAT "\n",
                      rgtbdDesc[iTb].m_rgchName, PchTm(sd),
                      iDuplicateIndex);
                    fPrintDuplicates = true;
                    printf("Printing out duplicate positions:\n");
                    fflush(stdout);
                    VVerifyIndex(iTb, sd);
                }
                else if(maxTbSize != rgCb[sd]) {
                    printf("\n%s %s size= "
                           DEC_INDEX_FORMAT " expected= " DEC_INDEX_FORMAT "\n",
                           rgtbdDesc[iTb].m_rgchName, PchTm(sd),
                           rgCb[sd], maxTbSize);
                }
                else
                    printf("ok\n");
                fflush (stdout);
            }
        }
        exit(0);
    }
#endif  // VERIFY_INDEX

    IInitializeTb (pszDirs);
    VInitializeAttacks();

    if (fQuery)
        {
        pbCache = (BYTE *) malloc (1024*1024);
        if (NULL == pbCache)
            {
            printf ("Unable to allocate %luMb of memory", (1024*1024)/(1024*1024));
            fflush (stdout);
            exit (1);
            }
        FTbSetCacheSize (pbCache, 1024*1024);
        rgBoards[0].VQuery ();
        return 0;
        }

    if (fAuto)
        {
        for (iArg = 1; iArg < argc; iArg ++)
            {
            int     iTb;
            INDEX   size;

            iTb = IFindTb (argv[iArg]);
            if (fBitVectors) {
                size = rgtbdDesc[iTb].m_rgcbLength[x_colorWhite];
                if (size < rgtbdDesc[iTb].m_rgcbLength[x_colorBlack])
                    size = rgtbdDesc[iTb].m_rgcbLength[x_colorBlack];
                size = (size + 7) / 8;
                // Unfortunately, have to use 2 times more memory even for
                // symmetric tables, as we need 2 bits per position...
                size <<= 1;
            }
            else
                size = rgtbdDesc[iTb].m_rgcbLength[x_colorWhite] +
                       rgtbdDesc[iTb].m_fSymmetric ? 0 :
                                                     rgtbdDesc[iTb].m_rgcbLength[x_colorBlack];
            size = ((size + 1024*1024 - 1) >> 20) << 20;
            if (size > cbMemoryUsed)
                cbMemoryUsed = size;
            }
        printf ("Using %luMb of memory\n", (ULONG) (cbMemoryUsed/(1024*1024)));
        }

    if (0 == cbMemoryUsed && !(fVerify | fSearch) || 0 == cbMemoryCache && !fMapFiles)
        {
        VShowHelp (false);
        fflush (stdout);
        exit (1);
        }
    if (fBitVectors)
        {
        bvArg.VInit (cbMemoryUsed/2);
        bvNew.VInit (cbMemoryUsed/2);
        }
    else
        {
        pbMemory = (BYTE *) malloc (cbMemoryUsed);
        if (NULL == pbMemory)
            {
            printf ("Unable to allocate %luMb of memory", cbMemoryUsed/(1024*1024));
            fflush (stdout);
            exit (1);
            }
        }
    if (0 != cbBuf)
        {
        pbBuf = (BYTE *) malloc (cbBuf+sizeof (ULONG));
        if (NULL == pbBuf)
            {
            printf ("Unable to allocate %luMb of memory", cbBuf/(1024*1024));
            fflush (stdout);
            exit (1);
            }
        pulGuardBuf = (ULONG*) (pbBuf + cbBuf);
        *pulGuardBuf = 0x12345678;
        }
    if (!fMapFiles || fVerify || fSearch)
        {
        pbCache = (BYTE *) malloc (cbMemoryCache);
        if (NULL == pbCache)
            {
            printf ("Unable to allocate %luMb of memory", cbMemoryCache/(1024*1024));
            fflush (stdout);
            exit (1);
            }
        }

#if defined (CALCULATE_TOTAL_SIZE)
    long long s;
    int iTb;

    s = 0;
    for (iTb = 1; iTb <= tbid_kqqpkq; iTb ++)
        {
        s += rgtbdDesc[iTb].m_rgcbLength[0];
        s += rgtbdDesc[iTb].m_rgcbLength[1];
        }
    printf ("Size = " DEC_INDEX_FORMAT "\n", s);
    exit (0);
#endif

#if defined (CALCULATE_SIZEOF)
    int iTb;

    for (iTb = tbid_kpppkp; iTb <= tbid_kpppkp; iTb ++)
        {
        VCalcTbSize (iTb);
        if (rgCb[0] != rgtbdDesc[iTb].m_rgcbLength[0] || rgCb[1] != rgtbdDesc[iTb].m_rgcbLength[1])
            {
            printf ("%s: size = " DEC_INDEX_FORMAT
                    ", " DEC_INDEX_FORMAT "\n",
                    rgtbdDesc[iTb].m_rgchName, rgCb[0], rgCb[1]);
            fflush (stdout);
            }
        }
    exit (0);
#endif

#if defined (SMP)
    if (cCPUS > 1)
        {
        for (int cpu = 0; cpu < cCPUS; cpu ++)
            {
            rgBoards[cpu].m_fHasWork = false;
            if (cpu != cCPUS-1)
                tfork (threads[cpu], VWorker, &rgBoards[cpu]);
            }
        }
#endif

    FTbSetCacheSize (pbCache, cbMemoryCache);
    for (iArg = 1; iArg < argc; iArg ++)
        {
        if (fSearch)
            VSearchTb (argv[iArg]);
        else if (fVerify)
            VVerifyTb (argv[iArg]);
        else
            VProcessTb (argv[iArg], 0);
        }
    VTbCloseFiles ();

    // Print statistics
    printf ("All done in");
    VPrintElapsedTime (tmStart);
    printf ("\n");

    return 0;
    }
